#!/usr/bin/perl
#
# Simple expression evaluator for reproducing MineSet's 'add_col' functionality
# 1/1/1999 by William H. Hsu
#

$ARGC = $#ARGV + 2;
if (($ARGC < 2) || ($ARGC > 2))
{
#  print "Usage: ADDCOLS.perl dataset-name {new-dataset-name}\n";
  print "Usage: ADDCOLS.perl dataset-name\n";
  exit;
}

$newColFile = $ARGV[0] . ".newcols";
$dataFile = $ARGV[0] . ".data";
$namesFile = $ARGV[0] . ".names";

# Read each column definition from the .newcols file
open (fpNewCol, $newColFile);

while ($line = <fpNewCol>)
{
  $colName = &getColName ($line);
  $colType = &getColType ($line);
  $colExpr = &getColExpr ($line);
  print "Adding column " . $colName . ", type " . $colType . ": " . $colExpr .
	"\n";
  &addDataCol ($dataFile, $namesFile, $colExpr);
  &appendName ($namesFile, $colName, $colType);
}

close (fpNewCol);

# Extract the name of the new column
sub getColName
{
  local ($line) = @_;
  local (@lineContents) = split (" ", $line);
  return $lineContents[0];
}

# Extract the type of the new column
# ### LATER: hook this up with the MLC++/MineSet schema compiler
sub getColType
{
  local ($line) = @_;
  local (@lineContents) = split (" ", $line);
  return $lineContents[1];
}

# Extract the definition of the new column
sub getColExpr
{
  local ($line) = @_;
  local (@lineContents) = split (" ", $line);
  local ($N) = $#lineContents + 1;

  for ($i = 0; $i < $N; $i++)
  {
    $expr[$i] = $lineContents[$i];
  }
  shift @expr;
  shift @expr;

  $result = "";
  for ($i = 0; $i < $N; $i++)
  {
    $result = $result . $expr[$i];
  }

  return $result;
}

# Append each new column to the .names (schema) file
sub appendName
{
  local ($namesFile, $colName, $colType) = @_;
  system "echo \"$colName: $colType.\" >> $namesFile";
}

# Compute the values for the new column and splice them into the .data file
sub addDataCol
{
  local ($dataFile, $namesFile, $colExpr) = @_;
  local (@colNames) = &getNames ($colExpr);

  # get the list of column numbers for attributes referenced in an expression
  #   by searching the names file
  local (%colIndices) = &getIndices ($namesFile, @colNames);

  &applySubstitution ($dataFile, $colExpr, %colIndices);
}

sub getNames
{
  local ($colExpr) = @_;
  local (@exprNoOps) = split (/-|\+|\*|\/|\(|\)/, $colExpr);
  local ($N) = $#exprNoOps + 1;
  local ($newExpr) = "";

  for ($i = 0; $i < $N; $i++)
  {
    $newExpr = $newExpr . " " . $exprNoOps[$i];
  }

  @attributeNames = split (" ", $newExpr);

  return @attributeNames;
}

# given a list of names and a names file, find the position of each name
sub getIndices
{
  local ($namesFile, @colNames) = @_;
  local (%filteredNames) = &filterNamesFile ($namesFile);
  local ($N) = $#colNames + 1;
  local (%colIndices);

  for ($i = 0; $i < $N; $i++)
  {
    # analogue of grep -n: find the column number corresponding to each name
    # store as an associative array (hash)
    $colIndices{$colNames[$i]} = $filteredNames{$colNames[$i]};
  }

  return %colIndices;
}

# scan the list of names and insert them into an associative array (hash)
sub filterNamesFile
{
  local ($namesFile) = @_;
  local ($i) = 0;
  local ($isDeclaration);
  local (@tokens, @extractedNames);

  open (fpNames, $namesFile);

  while ($line = <fpNames>)
  {
    @tokens = split (" ", $line);
    $isDeclaration = $tokens[0] =~ /^[a-z]+[a-z0-9]*\:/i;
    if ($isDeclaration)
    {
      @extractedNames = split (":", $tokens[0]);
      $key = $extractedNames[0];
      $filteredNames{$key} = $i;
      $i++;
    }
  }

  close (fpNames);

  return %filteredNames;
}

# read in the data in a stream and apply a substitution to each row
sub applySubstitution
{
  local ($dataFile, $colExpr, %colIndices) = @_;
  local ($tempDataFile) = $dataFile . "\.temp";

  open (fpData, $dataFile);
  open (fpNewData, "\>" . $tempDataFile);

  while ($dataPoint = <fpData>)
  {
    # pass the column indices to a function that takes a hash of subscripts, an
    #   expression, and a row of data and substitutes the appropriate values
    $newDataPoint =
	&applySubstitutionDataPoint ($colExpr, $dataPoint, %colIndices);
    print fpNewData $newDataPoint;
  }

  close ($fpData);
  close ($fpNewData);

  system "mv $tempDataFile $dataFile";
}

sub applySubstitutionDataPoint
{
  local ($colExpr, $dataPoint, %colIndices) = @_;
  local ($scannedExpr) = "";
  local (@dataValues) = split (/\,|\./, $dataPoint);
  local ($N) = $#dataValues - 1;
  local ($token, $prefix);
  local ($newValue, $newLine);

  # scan the expression: substitute for identifiers and copy all other tokens
  while (!($colExpr eq ""))
  {
    if ($colExpr =~ /^(-|\+|\*|\/|\(|\))/)
    {
      $prefix = "\\" . $1;
      $token = $1;
      $colExpr =~ s/$prefix//;
    }
    elsif ($colExpr =~ /(^[a-zA-Z]+[a-zA-Z0-9]*)/)
    {
      # substitute the values from the data point
      $token = $dataValues[$colIndices{$1}];
      $prefix = $1;
      $colExpr =~ s/$prefix//;
    }

    $scannedExpr = $scannedExpr . $token;
  }

  $newValue = eval $scannedExpr;
  $newLine = "";

  # print out the old row of data (comma-separated), WITHOUT the target value
  for ($i = 0; $i < $N; $i++)
  {
    $newLine = $newLine . $dataValues[$i] . "\,";
  }

  # print the new data value
  $newLine = $newLine . $newValue . "\,";

  # print the target value and the terminating .
  $newLine = $newLine . $dataValues[$N] . "\.\n";

  return $newLine;
}
