// MLC++ - Machine Learning Library in -*- C++ -*-
// See Descrip.txt for terms and conditions relating to use and distribution.
/***************************************************************************
  Description  : 2d Array template with error checking.  Can be initialized
                   with either with simply the number of rows and columns
		   or the number of rows and columns given a starting
		   row and column.  In either case, there also exists an
		   optional final initialization argument for initializing
		   the array to an initial value.
		 Access to elements is either by operator(), or by
		   index() which ranges form 0 to size-1.
		 Provides operator= for copying an Array2, if the
		   Array2's are the same size.
		 Derived Class PtrArray2 is an Array2 of pointers, which,
		   when its destructor is called, calls delete on
		   each of the pointers.
		 The default initialization is to NULL, but via the 3-argument
		   constructor, the user can turn off this O(numRows*numCols)
		   operation.
  Assumptions  : 
  Comments     : Index is not used in some routines because profiling shows
                   that they took huge amounts of time.  We access elements
		   directly.
		 For PtrArray2, because of the destructor, all of the pointers
		   should be pointing to either 1) a piece of memory that
		   can be freed, or 2) NULL.  Furthermore, to help prevent
		   errors, the user is not allowed to initialize the
		   Array2 with anything except for NULL.  
  Complexity   : Initializing an Array2 takes O(num-elements), as does the
                   destructor and free() routines.
  Enhancements : Display() should show a 2-D display instead of the string
                   it currently displays.
  History      : Jim Kelly                                        11/08/96
                   Added 4 min/max_in_row/col methods.
               : Dave Manley                                       9/26/93
                   Initial revision
***************************************************************************/
#include <basics.h>

// Template files shouldn't have RCSID
// RCSID("MLC++, $RCSfile: Array2.c,v $ $Revision: 1.32 $")


/***************************************************************************
    Description :  creates space for a newly constructed Array2
    Comments    :  
***************************************************************************/
template <class Element>   
void Array2<Element>::alloc_array(int startR, int startC,
				 int numR,int numC) 
{
   if (numR < 0 || numC < 0)
      err << "Array2<>::Illegal bounds requested, starting row: " << startR <<
	 " starting column: " << startC << fatal_error;
   startRow = startR;
   startCol = startC;
   numRows = numR;
   numCols = numC;
   elements = new Element[numR * numC];
}


/***************************************************************************
    Description :  creates space for new Array2 and assigns its values
    Comments    :  O(num-elements) time, assumes no need to free Array2
***************************************************************************/
template <class Element>   
void Array2<Element>::copy(const Array2<Element>& source) 
{  
   alloc_array(source.startRow, source.startCol, source.numRows,
	       source.numCols);
   for (int i = 0; i < size(); i++)
      elements[i] = source.elements[i];
}

/***************************************************************************
  Description : Compares two arrays.  Must have the same size and starting
                  rows and columns to be equal.
  Comments    :
***************************************************************************/

template <class Element>
Bool Array2<Element>::equal_given_that_bounds_are_equal
   (const Array2<Element>& rhs) const
{
   DBG(ASSERT(bounds_are_equal(rhs)));
   for (int x = 0; x < size(); x++)
      if (!(elements[x] == rhs.elements[x]))
	 return FALSE;

   return TRUE;
}

/***************************************************************************
    Description :  Constructor which takes number of Rows and columns.
    Comments    :  Creates Array2 where rows and columns start from 0.
***************************************************************************/
template <class Element> 
Array2<Element>::Array2(int numR, int numC)
{
   alloc_array(0, 0, numR, numC);
}


/***************************************************************************
    Description :  Constructor which takes number of Rows and columns.
                   Initializes elements to given initial value.
    Comments    :  Creates Array2 where rows and columns start from 0.
                   Takes O(num-elements) time.
***************************************************************************/
template <class Element> 
Array2<Element>::Array2(int numR, int numC,
			const Element& initialValue)
{
   alloc_array(0, 0, numR, numC);
   init_values(initialValue);
}

/***************************************************************************
    Description :  Constructor which takes starting Row, starting Column,
                     and number of following rows and columns.
    Comments    :  
***************************************************************************/
template <class Element> 
Array2<Element>::Array2(int startR, int startC,
			int numR, int numC)
{
   alloc_array(startR, startC, numR, numC);
}


/***************************************************************************
    Description :  Constructor which takes starting Row, starting Column,
                     and number of following rows and columns.
		   Initializes elements to given initial value.
    Comments    :  Takes O(num-elements) time.
***************************************************************************/
template <class Element> 
Array2<Element>::Array2(int startR, int startC,
			int numR, int numC,
			const Element& initialValue)
{
   alloc_array(startR, startC, numR, numC);
   init_values(initialValue);
}


/***************************************************************************
    Description :  Copy constructor which takes an extra argument.
    Comments    :  We don't use the standard issue copy constructor because
                     we want to know exactly when we call the copy constructor.
***************************************************************************/
template <class Element> 
Array2<Element>::Array2(const Array2<Element>& source,
			CtorDummy /*dummyArg*/)
{
   copy(source);
}


/***************************************************************************
    Description :  copies values from elem.
    Comments    :  Runs in O(num-elements) time.
                   Arrays must be the same size.  Their starting rows and
		     columns must also match.
***************************************************************************/
template <class Element> 
Array2<Element>& Array2<Element>::operator=(const Array2<Element>& elem) 
{
   if (this != &elem) {
      DBG(if (elem.size() != size())
	  err << "Array2<>::operator=: Cannot assign array sized: " <<
	  elem.size() << " to an array sized: " << size() << fatal_error);
      DBG(if (elem.start_row() != start_row() ||
	      elem.start_col() != start_col())
	  err << "Array2<>::operator=: Cannot assign array["
	  << elem.start_row() << ":" << elem.high_row() << ", "
	  << elem.start_col() << ":" << elem.high_col() << "] to an array["
	  << start_row() << ":" << high_row() << ", "
	  << start_col() << ":" << high_col() << "]" << fatal_error);
      for (int i = 0; i < size(); i++)
	 elements[i] = elem.elements[i];
   }
   return *this;
}

/***************************************************************************
    Description :  Initializes all elements of an Array2 to the initialValue
    Comments    :  Runs in O(num-elements) time
***************************************************************************/
template <class Element> 
void Array2<Element>::init_values(const Element& initialValue) 
{
  // walk through the Array2 and initialize all of the values
  int arraySize = size(); // Profiling shows init is critical
  for (int i = 0; i < arraySize; i++)
     elements[i] = initialValue;
}

/***************************************************************************
    Description :  The next two functions return a reference and a const
                     reference, respectively, to the correct element
		     referenced from a zero base. 
                   
    Comments    :  Out of bounds reference is a fatal error.
***************************************************************************/
template <class Element> 
Element& Array2<Element>::index(int i)
{
   DBG(if (i < 0 || i >= size())
       err<<"Array2<>::index: reference " << i <<
       " is out of bounds (" << 0 << ".." << size() - 1 << ')'
       << fatal_error); 
  
   return elements[i];
}

template <class Element> 
const Element& Array2<Element>::index(int i) const
{
   DBG(if (i < 0 || i >= size())
       err<<"Array2<>::index() const:  reference " << i <<
       " is out of bounds (" << 0 << ".." << size() - 1 << ')'
       << fatal_error); 
  
   return elements[i];
}


/***************************************************************************
    Description :  Return a reference to the element referenced.
    Comments    :  Out of bounds reference is a fatal error.
***************************************************************************/

template <class Element> 
Element& Array2<Element>::operator()(int row, int col)
{
   DBG(if (row < startRow || row >= numRows + startRow)
       err<<"Array2<>::operator(): row reference " << row <<
       " is out of bounds (" << startRow << ".." << startRow + numRows - 1
       << ')' << fatal_error;

      if (col < startCol || col >= numCols +startCol)
       err<<"Array2<>::operator(): column reference " << col <<
       " is out of bounds (" << startCol << ".." << startCol + numCols - 1
       << ')' << fatal_error); 
  
   return elements[(row-startRow)*numCols + (col-startCol)];
}

/***************************************************************************
  Description : Find min (or max) in given row (or column.)  Return a
                reference to the element in the array and set index value of
		element.
  Comments    : Array must not be empty.
                Tie breaker during equality prefer earlier elements in
                  the array.
***************************************************************************/

template <class Element> 
const Element& Array2<Element>::max_in_row(int row, int& idx) const
{
   if (numRows == 0 || numCols == 0)
      err << "Array2<Element>::max_in_row() - empty array" << fatal_error;

   // it's possible start_row is not 0 so do subtraction since what we want is
   // to pass number-of-rows * number-of-cols per row to get to the row we're
   // intereseted in. 
   int start = (row - start_row()) * numCols;
   const Element* max = &index(start);
   idx = start;
   for (int i = start+1; i < start + numCols; i++)
      if (index(i) > *max) {
	 max = &index(i);
	 idx = i;
      }
   return(*max);
}

template <class Element> 
const Element& Array2<Element>::min_in_row(int row, int& idx) const
{
   if (numRows == 0 || numCols == 0)
      err << "Array2<Element>::min_in_row() - empty array" << fatal_error;

   // It's possible start_row is not 0 so do subtraction since what we want is
   // to pass number-of-rows * number-of-cols per row to get to the row we're
   // intereseted in.
   int start = (row - start_row()) * numCols;
   const Element* min = &index(start);
   idx = start;
   for (int i = start+1; i < start + numCols; i++)
      if (index(i) < *min) {
	 min = &index(i);
	 idx = i;
      }
   return(*min);
}

template <class Element> 
const Element& Array2<Element>::max_in_col(int col, int& idx) const
{
   if (numRows == 0 || numCols == 0)
      err << "Array2<Element>::max_in_col() - empty array" << fatal_error;

   // It's possible start_col is not 0 so do subtraction so we start at
   // correct position relative to the starting col.
   int start = col - start_col();
   const Element* max = &index(start);
   idx = start;
   for (int i = start+numCols; i < numRows * numCols; i+=numCols)
      if (index(i) > *max) {
	 max = &index(i);
	 idx = i;
      }
   return(*max);
}

template <class Element> 
const Element& Array2<Element>::min_in_col(int col, int& idx) const
{
   if (numRows == 0 || numCols == 0)
      err << "Array2<Element>::min_in_col() - empty array" << fatal_error;

   // It's possible start_col is not 0 so do subtraction so we start at
   // correct position relative to the starting col.
   int start = col - start_col();
   const Element* min = &index(start);
   idx = start;
   for (int i = start+numCols; i < numRows * numCols; i+=numCols)
      if (index(i) < *min) {
	 min = &index(i);
	 idx = i;
      }
   return(*min);
}

/***************************************************************************
    Description :  Sums of array.  sum_rows returns an array with an entry
                     (equal to the summation of the entries for that row)
		     for each row.  sum_cols likewise returns an array, but
		     with entries for the columns.  total_sum adds up all
		     the rows and columns.
    Comments    :  Array must not be empty (size 0)--as 'zero' is not known
                     for an arbitrary parameterized class.
		   For sum_rows() and sum_cols() the output array must be
		     the same size as, and start at the same offset as, the
		     appropriate dimension of the Array2.
		   fast_index() is used for speed.
***************************************************************************/
template <class Element>   
void Array2<Element>::sum_rows(Array<Element>& returnColumn) const
{
   if (numRows == 0 || numCols == 0)
      err << "Array2<Element>::sum_rows(): empty array" << fatal_error;

   if (numRows != returnColumn.size())
      err << "Array2<Element>::sum_rows(): output array of wrong size--"
	  << returnColumn.size() << " instead of " << numRows << fatal_error;

   if (startRow != returnColumn.low())
      err << "Array2<Element>::sum_rows(): output array indexed incorrectly--"
	 "starts at " << returnColumn.low() << " instead of "
	  << startRow << fatal_error;
   
   int i = 0;
   for (int row = 0; row < numRows; row++) {
      Element total = fast_index(i++);
      for (int col = 1; col < numCols; col++) {
	 total += fast_index(i++);
      }
      returnColumn.fast_index(row) = total;
   }
   DBG(ASSERT(i == numRows*numCols));
}

template <class Element>   
void Array2<Element>::sum_cols(Array<Element>& returnRow) const
{
   if (numRows == 0 || numCols == 0)
      err << "Array2<Element>::sum_cols(): empty array" << fatal_error;

   if (numCols != returnRow.size())
      err << "Array2<Element>::sum_cols(): output array of wrong size--"
	  << returnRow.size() << " instead of " << numCols << fatal_error;

   if (startCol != returnRow.low())
      err << "Array2<Element>::sum_cols(): output array indexed incorrectly--"
	 "starts at " << returnRow.low() << " instead of "
	  << startCol << fatal_error;

   int i = 0;
   for (int col = 0; col < numCols; col++)
      returnRow.fast_index(col) = fast_index(i++);

   for (int row = 1; row < numRows; row++) {
      for (int col = 0; col < numCols; col++)
	 returnRow.fast_index(col) += fast_index(i++);
   }
   DBG(ASSERT(i == numRows*numCols));
}

template <class Element> 
Element Array2<Element>::total_sum() const
{
   if (numRows == 0 || numCols == 0)
      err << "Array2<Element>::total_sum(): empty array" << fatal_error;

   int arraySize = size();
   
   Element total = fast_index(0);
   for (int i = 1; i < arraySize; i++)
      total += fast_index(i);

   return total;
}

/***************************************************************************
  Description : Output array to an MLCStream
  Comments    :
***************************************************************************/
template<class Element>
void Array2<Element>::display(MLCOStream& stream) const
{
   stream << "Matrix [" << start_row() << ':' << high_row()
          << ',' << start_col()  << ':' << high_col() << "] is:" << endl;

   for (int i = start_row(); i <= high_row(); i++) {
      for (int j = start_col(); j < high_col(); j++)
         stream <<  (*this)(i,j) << ", ";

      // If we have columns (not 0 columns), we print the last element
      if (j > start_col())
	 stream << (*this)(i,j);  // no trailing comma.
      if (i < high_row()) 
	 stream << endl;
   }
}

/***************************************************************************
    Description :  Two constructors for PtrArray2.
                   Initializes all the pointers in the specified Array2 by
                     default or if the user passes in true for init.  (Leaves
		     unitialized if the user specifies false).
    Comments    :  Takes O(num-elements) time if the user wishes to
                     initialize the Array2.
		   Cannot use init_vals because it takes a reference.
***************************************************************************/
template <class Element> 
PtrArray2<Element>::PtrArray2(int numR, int numC,
			      Bool init)
                  :Array2<Element>(numR, numC)
{
   if (init == TRUE)
      for (int i = 0; i < size(); i++)
	 elements[i] = NULL;
}

template <class Element> 
PtrArray2<Element>::PtrArray2(int startR, int startC,
			      int numR, int numC,
			      Bool init)
                  :Array2<Element>(startR, startC, numR, numC)
{
   if (init == TRUE)
      for (int i = 0; i < size(); i++)
	 elements[i] = NULL;
}


/***************************************************************************
    Description :  Calls delete for all of the Pointers in the Array2.
    Comments    :  Takes O(num-elements) time.
***************************************************************************/
template <class Element> 
PtrArray2<Element>::~PtrArray2()
{
   for (int i = 0; i < size(); i++)
      delete elements[i];
}

/***************************************************************************
    Description :  Assigning to a PtrArray2 causes a fatal error.
    Comments    :  The reasoning for this: Assignment will cause two arrays
                     to point to the same element.  When the
		     destructor is called, elements are freed twice.
***************************************************************************/
template <class Element> 
PtrArray2<Element>& PtrArray2<Element>::operator=(const PtrArray2<Element>&
						  /*elem*/) 
{
   err<<"PtrArray2<>::operator=: Cannot assign to a PtrArray2" << fatal_error;
   return (*this);
}

template <class Element>
DEF_DISPLAY(Array2<Element>)


