// MLC++ - Machine Learning Library in -*- C++ -*-
// See Descrip.txt for terms and conditions relating to use and distribution.
/***************************************************************************
  Description  : BFSearch implements a best-first search State Space Search.
                   See SSSearch.c and the function comments below
		   for more information.
  Assumptions  :
  Comments     :
  Complexity   : Depends on space being searched.
  Enhancements :
  History      : Clay Kunz                                          6/22/97
                   Added Beam Search into BFSearch
                 Brian Frasca                                       5/21/94
                   Initial revision (.h,.c)
***************************************************************************/
#include <basics.h>
#include <BFSearch.h>
#include <DblLinkList.h>
#include <GetOption.h>

// RCSID("MLC++, $RCSfile: BFSearch.c,v $ $Revision: 1.26 $")


// Option information.  Must be #define's instead of const declarations
// because this file is templated.
#define MAX_EXPANSIONS_HELP "This option ..."
#define EPSILON_HELP "This option..."
#define MAX_FITNESS_HELP "What the optimal node's fitness should be. "\
"0 if unbounded"
#define BEAM_WIDTH_HELP "The number of nodes to expand at each step"
#define PRUNE_HELP "Whether to prune the unexpanded nodes at each step"
#define DEFAULT_MAX_STALE 5
#define DEFAULT_EPSILON 0.001
#define DEFAULT_MAX_FITNESS 1.0
#define DEFAULT_BEAM_WIDTH 1
#define DEFAULT_PRUNE FALSE

/***************************************************************************
  Description : Returns TRUE iff the termination condition has been reached.
                The termination condition is when the number of consecutive
		  non-improving node expansions is greater than some maximum
		  value (set through an option).
		Empty open list and exceeding the limit number of
		  evaluations will also terminate the search.
  Comments    : This is a protected method.
***************************************************************************/
template <class LocalInfo, class GlobalInfo>
Bool BFSearch<LocalInfo, GlobalInfo>::terminate()
{
   if (openList.empty()) {
      LOG(2, "Terminating due to empty open list" << endl);
      return TRUE;
   }
   
   ASSERT(get_eval_limit() >= 0);
   ASSERT(graph->get_num_states() >= 0);
   ASSERT(numNonImprovingExpansions >= 0);
   ASSERT(maxNonImprovingExpansions > 0);
   if (numNonImprovingExpansions >= maxNonImprovingExpansions) {
      LOG(2, "Terminating because number of non-improving expansions "
	  "hit the maximum (" << maxNonImprovingExpansions <<
	  ")" << endl);
      return TRUE;
   }

   if (get_eval_limit() && graph->get_num_states() > get_eval_limit()) {
      LOG(2, "Terminating because more nodes were evaluated than "
	  "EVAL_LIMIT (" << graph->get_num_states() << ">"
	  << get_eval_limit() << ")" << endl);
      return TRUE;
   }
   
   return FALSE;
}

template <class LocalInfo, class GlobalInfo>
void BFSearch<LocalInfo, GlobalInfo>::set_defaults()
{
   StateSpaceSearch<LocalInfo, GlobalInfo>::set_defaults();
   set_max_non_improving_expansions(DEFAULT_MAX_STALE);
   set_min_expansion_improvement(DEFAULT_EPSILON);
   set_max_fitness(DEFAULT_MAX_FITNESS);
   set_beam_width(DEFAULT_BEAM_WIDTH);
   set_prune(DEFAULT_PRUNE);
}

template <class LocalInfo, class GlobalInfo>
void BFSearch<LocalInfo, GlobalInfo>::set_beam_width(int width)
{
   if (width < 1)
      err << "BFSearch::set_beam_width: the given width " << width
	  << " is less than 1, which is not allowed" << fatal_error;
   beamWidth = width;
}

template <class LocalInfo, class GlobalInfo>
void BFSearch<LocalInfo, GlobalInfo>::set_user_options(const MString& stem)
{
   StateSpaceSearch<LocalInfo, GlobalInfo>::set_user_options(stem);
   set_max_non_improving_expansions(
      get_option_int(stem + "MAX_STALE", maxNonImprovingExpansions,
		     MAX_EXPANSIONS_HELP, FALSE));
   set_min_expansion_improvement(
      get_option_real(stem + "EPSILON", minExpansionImprovement,
		      EPSILON_HELP, FALSE));
   set_max_fitness(get_option_real(stem + "MAX_FITNESS", maxFitness,
				   MAX_FITNESS_HELP, TRUE));
   set_beam_width(get_option_int_range(stem + "BEAM_WIDTH", beamWidth, 1,
				       INT_MAX, BEAM_WIDTH_HELP, TRUE));
   set_prune(get_option_bool(stem + "PRUNE", prune, PRUNE_HELP, TRUE));

}



/***************************************************************************
  Description : This function does a best-first search beginning with the
                  given state.  It stops when the termination condition is
		  reached (see above).
  Comments    : The given starting node is deleted in the destructor of
                  StateSpaceSearch (the base class) or when search() is
		  called again.
***************************************************************************/
template <class LocalInfo, class GlobalInfo>
const State<LocalInfo, GlobalInfo>& BFSearch<LocalInfo, GlobalInfo>::
search(State<LocalInfo, GlobalInfo>* initialState, GlobalInfo *gInfo,
       const MString& displayFileName)
{
   // Setup variables.
   numExpansions = 0;
   numNonImprovingExpansions = 0;
   bestNode = NULL;
   Real bestEval = MIN_EVALUATION;
   delete graph;
   graph = new StateSpace< State<LocalInfo, GlobalInfo> >;
   graph->set_log_level(get_log_level()-1);

   // Start with empty open/closed lists.
   openList.free();
   closedList.free();
   
   // Evalute initial node and put in on the open list.
   initialState->eval(gInfo, get_show_test_set_perf() == showAlways);
   initialState->set_eval_num(graph->get_next_eval_num());
   NodePtr root = graph->create_node(initialState);
   graph->set_initial_node(root);
   openList.append(root);

   int numChildrenGenerated = 1;

   // Declare these here so they don't get allocated & deleted with each loop
   Array<NodePtr> bestNodes(beamWidth);
   Array<Real> fitnesses(beamWidth);
   
   while (!terminate()) {

      // Find the best nodes on the open list.  Note that there must be some
      // node on the open list or else we wouldn't pass the termination
      // condition.

      int numGoodNodes = MLC::min(beamWidth, openList.length());

      for (int i = 0; i < numGoodNodes; i++) {
	 DBG(ASSERT(openList.length() > 0));
	 Real bestOpenNodeFitness = MIN_EVALUATION;
	 DLLPix<NodePtr> bestOpenNodePix(openList,0);
	 for (DLLPix<NodePtr> nodePix(openList,1); nodePix; ++nodePix) {
	    const State<LocalInfo, GlobalInfo>& tmpState =
	       graph->get_state(openList(nodePix));
	    if (tmpState.get_fitness() > bestOpenNodeFitness) {
	       bestOpenNodeFitness = tmpState.get_fitness();
	       bestOpenNodePix = nodePix;
	    }
	 }
	 DBG(ASSERT(bestOpenNodePix));
	 bestNodes[i] = openList(bestOpenNodePix);
	 fitnesses[i] = bestOpenNodeFitness;
	 DBG(ASSERT(fitnesses[i] ==
		    graph->get_state(bestNodes[i]).get_fitness()));
	 LOG(3, "Added " << graph->get_state(bestNodes[i])
	     << " to list of parents to expand" << endl);
	 // Remove best node from open list and put in on the closed list.
	 closedList.append(bestNodes[i]);
	 openList.del(bestOpenNodePix);
      }


      // We know that the arrays of bestNodes & fitnesses are sorted from best
      // to worst (kth best).  So the very best is in bestNodes[0]
      NodePtr bestChild = bestNodes[0];
      Real bestChildFitness = fitnesses[0];
      
      if (maxFitness != 0.0 &&
	  (mlc.approx_greater(bestChildFitness + minExpansionImprovement,
			     maxFitness) ||
	  mlc.approx_equal(bestChildFitness + minExpansionImprovement,
			   maxFitness))) {

	 bestNode = bestChild;

	 // terminating because got max (or near) fitness
	 if(get_show_test_set_perf() == showBestOnly)
	    graph->get_state(bestNode).eval(gInfo, TRUE, FALSE);
	 LOG(2, "Terminating because we found a node with near maximum "
	     "fitness:" << maxFitness << endl);
	 break; 
      }
      
      // Prune everything we didn't expand, if appropriate
      if (prune)
	 while (!openList.empty())
	    closedList.append(openList.remove_front());
      
      // if we're showing real error on best nodes only, get
      // the real error of the best node now
      if(get_show_test_set_perf() == showBestOnly)
	 graph->get_state(bestChild).eval(gInfo, TRUE, FALSE);
      
      // Update number of non-improving expansions.
      // In this case, an expansion is non-improving if all of the children
      // are no good (so if the best one is no good)
      // @@ changing the next if statement to:
      // @@   if (bestChildFitness - bestEval > minExpansionImprovement)
      // @@ causes t_FSSInducer to fail on the diff
      if (mlc.approx_greater(bestChildFitness,
			     bestEval + minExpansionImprovement))
	 numNonImprovingExpansions = 0;
      else if (numChildrenGenerated > 0)
	 numNonImprovingExpansions++;
      if (numChildrenGenerated > 0)
	 numExpansions++;
      
      // Remember best node added to the closed list.
      if (mlc.approx_greater(bestChildFitness , bestEval)) {
	 bestEval = bestChildFitness;
	 bestNode = bestChild;
	 DRIBBLE("\nNew best node (" << graph->get_num_states() << " evals) "
             << graph->get_state(bestNode) << endl);

      }

      for (i = 0; i < numGoodNodes; i++)
	 LOG(2, "expanding " << graph->get_state(bestNodes[i]) << endl);
      LOG(2, numNonImprovingExpansions << " non-improving expansions."
	  << endl);

      // If termination condition is reached, exit without evaluating nodes
      // from final expansion.
      if (numNonImprovingExpansions >= maxNonImprovingExpansions) {
	 LOG(2, "Terminating because number of non-improving expansions "
	     "hit the maximum (" << maxNonImprovingExpansions <<
	     ")" << endl);
	 break;
      }
      
      // For all successors which are not in the graph (on open or closed
      // lists), evaluate them and put them on the open list.
      Bool computeRealNow = (get_show_test_set_perf() == showAlways);
      numChildrenGenerated = 0;
      for (i = 0; i < numGoodNodes; i++) {
	 DblLinkList<State<LocalInfo, GlobalInfo>*>* successors =
	    graph->get_state(bestNodes[i]).gen_succ(gInfo, graph,
						    computeRealNow);
	 numChildrenGenerated += successors->length();
	 while (!successors->empty()) {
	    NodePtr childNode = successors->remove_front()->get_node(graph);
	    ASSERT(childNode);
	    openList.append(childNode);
	 }
	 delete successors;
      }
   }
   DRIBBLE("\nFinal best node "
       << graph->get_state(bestNode) << endl);
   DRIBBLE("Expanded " << numExpansions << " nodes" << endl);
   LOG(2, "evaluated " << graph->get_num_states() << " nodes" << endl);
  
   graph->set_final_node(bestNode);

   // if we're showing real error on the final state only, compute
   // real error here
   if(get_show_test_set_perf() == showFinalOnly)
      graph->get_state(bestNode).eval(gInfo, TRUE, FALSE);
   
   // display the state space, if called for
   if(displayFileName != EMPTY_STRING) {
      MLCOStream out(displayFileName);
      DotGraphPref pref;
      graph->display(out, pref);
   }
   return graph->get_state(bestNode);
}
