// MLC++ - Machine Learning Library in -*- C++ -*-
// See Descrip.txt for terms and conditions relating to use and distribution.

// This is an include file.  For a full description of the class and
// functions, see the file name with a "c" suffix.

#ifndef _BagCounters_h
#define _BagCounters_h 1

#include <Array.h>
#include <Array2.h>
#include <InstanceRC.h>

class MInStore;
class MOutStore;

// The PerBagCounters_ class helps BagCounters become persistent.
// it is a slightly flatter storage mechanism for BagCounters.
class PerBagCounters_ {
public:
   NO_DEFAULT_OPS(PerBagCounters_);
   // store a single, flat array of real-valued counts for the whole
   // set of counters
   Array<Real> *counts;

   PerBagCounters_()
      : counts(NULL) {}
   virtual void read(MInStore& out);
   virtual void write(MOutStore& out);
};


class BagCounters {  // contains no virtual functions
   // See BagCounters.c for a description of this complicated structure.
   NO_COPY_CTOR(BagCounters);

   // Member data
   PtrArray<Array2<Real>*>* valueCounts;
   Array<Real>* labelCounts;
   PtrArray<Array<Real>*>* attrCounts;
   DBG_DECLARE(SchemaRC schema;)

   // Private methods
   DBG_DECLARE(void check_nominal(int attrNum) const;) 

   void allocate_all(const SchemaRC aSchema);
   
protected:
   // Protected methods
   // Increment and Decrement are prefix (operation, then return)
   void update_counters(const InstanceRC& instance, int delta);
public:
   Real OK(int level = 0) const; // returns sum of label counters.
   // The constructor zeroes all counts.
   BagCounters(const SchemaRC& aSchema);
   // copy constructor
   BagCounters(const BagCounters& source, CtorDummy);
   // persistent constructor
   BagCounters(const PerBagCounters_& data, const SchemaRC& aSchema);
   ~BagCounters();

   // comparison methods.
   Bool operator==(const BagCounters& rhs) const;
   Bool operator!=(const BagCounters& rhs) const
      {return !operator==(rhs);}

   // count of label=labelVal, attributeNum = attrVal
   Real val_count(NominalVal labelVal, int attrNum,
                 NominalVal attrVal) const;
   // count of attrNum = attrVal (over all labels)
   Real attr_count(int attrNum, NominalVal attrVal) const;
   Real label_count(NominalVal labelVal) const;
   // count number of existing different values.
   int attr_num_vals(int attrNum) const;
   // count number of actual different labels appearing.
   int label_num_vals() const;
   void add_instance(const InstanceRC& instance);
   void del_instance(const InstanceRC& instance);
   void display(MLCOStream& stream = Mcout) const;
   // accessor functions
   const Array<Array2<Real>*>& value_counts() const; 
   const Array<Real>& label_counts() const; 
   const Array<Array<Real>*>& attr_counts() const; 

   // construction of persistent data
   void make_persistent(PerBagCounters_& data) const;

   BagCounters& operator=(const BagCounters& source);

   // @@ a hack to change ownership
   PTHR_DECL(void change_ownership() {
      DBG_DECLARE(schema.change_ownership();) })

};

DECLARE_DISPLAY(BagCounters)


#endif
