// MLC++ - Machine Learning Library in -*- C++ -*-
// See Descrip.txt for terms and conditions relating to use and distribution.

// This is an include file.  For a full description of the class and
// functions, see the file name with a "c" suffix.

#ifndef _CGraph_h
#define _CGraph_h 1

// WARNING: Clear must be defined BEFORE #include <LEDA/graph.h>
// Let LEDA use our clear procedure to avoid OC warning when
// deleting the GRAPH.  Note the functions have precedence over 
// templates (Stroustrup page 597).
class AugCategory;
class ostream;
class Categorizer;
class NodeCategorizer;
class NodeInfo;
class BoolArray;
class CGraph;
inline void Clear(NodeInfo*& x) { x = NULL; } // already deallocated.
inline void Clear(AugCategory*& x) { x = NULL; } // already deallocated.


// used to print edge labels in LEDA's print routine
ostream& operator<<(ostream& stream, const AugCategory *ac);
ostream& operator<<(ostream& stream, const NodeInfo *nodeInfo);


#include <LEDA/graph.h>
#include <DblLinkList.h>
#include <LogOptions.h>

extern const int DEFAULT_LEVEL; 

typedef node NodePtr;  // LEDA's node is really a pointer to a node
typedef edge EdgePtr;  // LEDA's edge is really a pointer to an edge

class NodeInfo {
   NO_DEFAULT_OPS(NodeInfo);

   // Member data
   NodeCategorizer* cat;
   int lev;
   DblLinkList<NodePtr> parents;
public:
   void OK(int okLevel = 0) const;
   void OK_expect(NodePtr expectedNode, CGraph *expectedGraph, int level)
      const; 

   // Sets categorizer to badCategorizer
   NodeInfo(int levl);
   NodeInfo(NodeCategorizer*& categorizer, int levl);
   NodeInfo(NodeCategorizer*& categorizer, int levl,
	    const DblLinkList<NodePtr>& par);
   NodeInfo(const NodeInfo& info, CtorDummy);
   ~NodeInfo();
   int level() const {return lev;}
   void set_level(int newLev) { lev = newLev; }
   // Gets ownership of categorizer
   void set_categorizer(NodeCategorizer*& categorizer);
   const NodeCategorizer& get_categorizer() const {return *cat;}
   // non-const version of get_categorizer()
   NodeCategorizer& get_categorizer() { return *cat; }
   // Add and delete par to the parents list.
   void add_parent(NodePtr par);
   Bool del_parent(NodePtr par);
   void assign_categorizer(NodeInfo* newInfo);
   // Returns the parents list.
   const DblLinkList<NodePtr>& get_parents() const { return parents; }

   virtual NodeInfo *create_my_type(int level) const
       { return new NodeInfo(level); }
   virtual NodeInfo *clone() const { return new NodeInfo(*this, ctorDummy); }
   virtual int class_id() const { return -1; }
};


typedef GRAPH<NodeInfo*, AugCategory*> CGRAPH;


class CGraph : public CGRAPH {
   NO_DEFAULT_OPS(CGraph);
   LOG_OPTIONS;

   // Member data
   NodeInfo *prototype;   // used to build correct NodeInfo types when
                          // building graph, by using NodeInfo's
                          // create_my_type member function.

   // Private methods
   virtual Bool is_child(NodePtr c, NodePtr p) const ;
   virtual Bool is_parent(NodePtr p, NodePtr c) const;
public:
   virtual void OK(int level=0);
   CGraph();
   CGraph(NodeInfo *& proto);
   CGraph(const CGraph& source, CtorDummy);
   virtual ~CGraph();
   virtual int num_leaves() const;
   // Update and access function for parent nodes.
   virtual void add_parent(NodePtr n, NodePtr par);
   virtual void del_parent(NodePtr n, NodePtr par);
   virtual NodePtr get_child(const NodePtr parent, 
			     const AugCategory& edgeLabel) const;
   void assign_categorizer(NodePtr oldNode, NodePtr newNode);
   void set_categorizer(NodePtr nodePtr, NodeCategorizer*& categorizer);
   virtual const DblLinkList<NodePtr>& get_parents(NodePtr n) const;
   virtual NodePtr new_node(NodeInfo *nodeInfo);
   EdgePtr new_edge(NodePtr v, NodePtr w, AugCategory* x);
   void del_edge(EdgePtr e);
   virtual NodeInfo *get_prototype() const { return prototype; }
   virtual void change_threshold(NodePtr n, Real threshold);

   // "index" is a friend function usable only within this class.
   // node_index does the same thing but is a member function.
   int node_index(NodePtr whichNode) const { return index(whichNode); }
   virtual void set_used_attr(BoolArray& usedAttr) const;
   virtual void project(const BoolArray& mask);
   virtual int num_attr(const LogOptions& logOptions, int maxAttr) const;
   virtual int longest_path(NodePtr source) const;
   NodeInfo*& node_info(NodePtr node) {return operator[](node);}
   AugCategory*& edge_info(EdgePtr edge) {return operator[](edge);}
   
   // zero class id means a histogram of all nodes.
   virtual Array<int>* level_histogram(int classID = 0) const;
   virtual Array<int>* edge_level_histogram(int classID = 0) const;
};

#endif
