// MLC++ - Machine Learning Library in -*- C++ -*-
// See Descrip.txt for terms and conditions relating to use and distribution.

// This is an include file.  For a full description of the class and
// functions, see the file name with a "c" suffix.

#ifndef _CatDist_h
#define _CatDist_h 1

#include <Array.h>
#include <SchemaRC.h>
#include <AugCategory.h>
#include <LogOptions.h>

// helper class for CatDist.  This is used to sort values for
// set_tiebreaking_by_values().  It wants bigger values first,
// so operator< is reversed.
class RealAndCategory {
public:  
   Real p;
   Category c;

   RealAndCategory() : p(0.0), c(UNKNOWN_CATEGORY_VAL) { }
   RealAndCategory& operator=(const RealAndCategory& rhs) {
      p = rhs.p; c = rhs.c; return *this; }
   Bool operator<(const RealAndCategory& rhs) const {
      return ((p > rhs.p) || ((p == rhs.p) && (c < rhs.c))); }
   Bool operator==(const RealAndCategory& rhs) const {
      return ((p == rhs.p) && (c == rhs.c)); }
   void display(MLCOStream& stream) const {
      stream << c << ": " << p; }
};

// don't use DEF_DISPLAY so we can keep all code in the header
inline MLCOStream& operator<<(MLCOStream& stream, const RealAndCategory& rc);
inline MLCOStream& operator<<(MLCOStream& stream, const RealAndCategory& rc)
{ rc.display(stream); return stream; }


// Helper function to apply evidence projection to a distribution
Real single_evidence_projection(Real count, Real total, Real maxEvidence);

void apply_evidence_projection(Array<Real>& counts, Real eviFactor,
			       Bool firstIsUnknown);


class CatDist {
   NO_COPY_CTOR(CatDist);
   NO_OPERATOR_EQ(CatDist);
   LOG_OPTIONS;
   
   Array<Real> dist;
   Array<int> tiebreakingOrder;
   SchemaRC schema;

public:
   static void check_tiebreaking_order(const Array<int>& order);
   
   static Array<int>* get_default_tiebreaking_order(int numLabelVals);

   static Array<int>*
   merge_tiebreaking_order(const Array<int>& tieBreakingOrder,
			   const Array<Real>& weightDistribution);
   static Array<int>*
   tiebreaking_order(const Array<Real>& weightDistribution);
   static Category majority_category(const Array<Real>& weightDistribution,
				     const Array<int>& tieBreakingOrder);
   
   enum CorrectionType { none, laplace, evidence };
   
   virtual void OK(int level=0);

   // Build an empty CatDist--this will always predict unknown
   CatDist(const SchemaRC& aSchema);
   
   // build a CatDist from frequency counts + optional correction
   CatDist(const SchemaRC& aSchema, const Array<Real>& frequencyCounts,
	   CorrectionType cType = none, Real cParam = 1.0);

   // build a CatDist if we know the distribution already
   CatDist(const SchemaRC& aSchema, Real unknownProb,
	   const Array<Real>& knownDist);

   // build an all-or-nothing CatDist given the majority category
   CatDist(const SchemaRC& aSchema, Category singleCat);

   CatDist(const CatDist& cDist, CtorDummy);
   virtual ~CatDist() { }

   // set the scores for the CatDist.  These functions behave exactly
   // like the corresponding constructors
   void set_scores(const Array<Real>& frequencyCounts,
		   CorrectionType cType = none, Real cParam = 1.0);
   void set_scores(Real unknownProb,
		   const Array<Real>& knownDist);
   void set_scores(Category singleCat);
   
   void smooth_toward(const Array<Real>& distribution, Real smoothFactor);

   virtual Bool operator==(const CatDist& rhs) const;
   
   // read results
   virtual const Array<Real>& get_scores() const { return dist; }
   virtual const SchemaRC& get_schema() const { return schema; }

   // produce a category
   virtual AugCategory best_category() const;

   // set the schema for this categorizer
   virtual void set_schema(const SchemaRC& newSchema);
   
   // display.
   virtual void display(MLCOStream& stream) const;

   // get/set tiebreaker information
   virtual void set_preferred_category(Category cat);
   virtual void set_default_tiebreaking();
   virtual void set_tiebreaking_order(const Array<int>& order);
   virtual const Array<int>& get_tiebreaking_order() const
   { return tiebreakingOrder; }
   virtual void set_tiebreaking_by_values(const Array<Real>& values);
};

DECLARE_DISPLAY(CatDist)

#endif

