// MLC++ - Machine Learning Library in -*- C++ -*-
// See Descrip.txt for terms and conditions relating to use and distribution.

// This is an include file.  For a full description of the class and
// functions, see the file name with a "c" suffix.

#ifndef _CatGraph_h
#define _CatGraph_h 1

#include <MLCStream.h>
#include <DisplayPref.h>
#include <CGraph.h>
#include <Attribute.h>
#include <LogOptions.h>
#include <Persistence.h>

class AugCategory;

// persistent helpers for CatGraph; nodes are stored as categorizer/level
// pairs in a large array.  edges are an array of startIndex, endIndex,
// cat, and descr.

class PerEdge_ {
public:
   int fromIndex;
   int toIndex;
   int cat;
   MString descr;

   void read(MInStore& in);
   void write(MOutStore& out);
};

class PerCatGraph_ {
public:
   Array<PerEdge_> *edges;

   void read(MInStore& in);
   void write(MOutStore& out);
};


class CatGraph {
   NO_COPY_CTOR(CatGraph);
   LOG_OPTIONS;

// must be visible to subclasses for them to uses LEDA methods
protected:
   CGraph& cGraph;
   Bool graphAlloc;
   Bool isSparse;

   virtual void process_XStream_display(const DisplayPref& dp,
					Bool hasNodeLosses = FALSE,
					Bool hasLossMatrix = FALSE) const;
   virtual void process_DotPostscript_display(MLCOStream& stream,
					      const DisplayPref& dp,
					      Bool hasNodeLosses = FALSE,
					      Bool hasLossMatrix = FALSE)
      const;
   virtual void process_DotGraph_display(MLCOStream& stream,
					 const DisplayPref& dp,
					 Bool hasNodeLosses = FALSE,
					 Bool hasLossMatrix = FALSE) const;
   virtual void convertToDotFormat(MLCOStream& stream,
				   const DisplayPref& pref,
				   Bool hasNodeLosses = FALSE,
				   Bool hasLossMatrix = FALSE) const;
public:
   virtual void OK(int level = 1) const;
   CatGraph(Bool isGraphSparse = FALSE);
   CatGraph(NodeInfo*& prototype, Bool isGraphSparse = FALSE);
   CatGraph(CGraph&, Bool isGraphSparse = FALSE); // using caller's CGraph (no
                                                  // copy is made)
   CatGraph(const CatGraph& source, CtorDummy);
   CatGraph(const PerCatGraph_& data,
	    NodeInfo*& prototype,
	    const Array<NodeInfo *>& nodes,
	    Bool isGraphSparse = FALSE);

   virtual ~CatGraph();

   virtual CatGraph *clone() const { return new CatGraph(*this, ctorDummy); }

   CGraph& get_graph() const;
   NodePtr create_node(int level = DEFAULT_LEVEL);
   // gets ownership of Categorizer
   NodePtr create_node(NodeCategorizer*&, int level = DEFAULT_LEVEL);
  // gets ownership of Categorizer
   virtual void set_categorizer(const NodePtr, NodeCategorizer*&);
   NodeCategorizer& get_categorizer(const NodePtr);  // @@?should be protected
   const NodeCategorizer& get_categorizer(const NodePtr) const;
   int level(const NodePtr) const;
   // gets ownership of AugCategory
   virtual void connect(const NodePtr from, const NodePtr to, AugCategory*&);
   int num_children(const NodePtr parent) const;
   virtual NodePtr get_child(const NodePtr parent,
				   const AugCategory& edgeLabel) const;
   virtual NodePtr get_child_if_exists(const NodePtr parent, 
				       const AugCategory& edgeLabel) const;

   // hasLossMatrix is relevant only if you have nodeAcc
   virtual void display(Bool hasNodeLosses, Bool hasLossMatrix,
			MLCOStream& stream = Mcout,
			const DisplayPref& dp = defaultDisplayPref) const;
   virtual void display(MLCOStream& stream = Mcout,
			const DisplayPref& dp = defaultDisplayPref) const
      { display(FALSE, FALSE, stream, dp);}

   Bool is_sparse() const { return isSparse; }
   Bool check_reachable(const NodePtr source, 
			Bool fatalOnFalse=FALSE)const;
   Bool check_tree(const NodePtr root, Bool fatalOnFalse = FALSE)const;
   Bool check_node_in_graph(const NodePtr nodePtr, 
			    Bool fatalOnFalse = FALSE) const;
   virtual int num_nodes() const;
   virtual int num_leaves() const;
   virtual const AugCategory* get_edge(const NodePtr nodePtr,
				       Category edgeCategory) const;
   virtual int num_attr(int maxAttr) const
      { return cGraph.num_attr(get_log_options(), maxAttr);}
   virtual int longest_path(NodePtr source) const
   { return cGraph.longest_path(source);}

   virtual void set_used_attr(BoolArray& usedAttr)
      { cGraph.set_used_attr(usedAttr); }
   virtual void project(const BoolArray& mask)
      { cGraph.project(mask); }

   virtual void make_persistent(PerCatGraph_& data) const;
   virtual Array<NodeInfo *> *get_node_array() const;
   virtual const NodeInfo& get_prototype() const;
   Array<int>* level_histogram(int classID = 0) const
      {return cGraph.level_histogram(classID);}
   Array<int>* edge_level_histogram(int classID = 0) const
      {return cGraph.edge_level_histogram(classID);}

};

DECLARE_DISPLAY(CatGraph)

#endif
