// MLC++ - Machine Learning Library in -*- C++ -*-
// See Descrip.txt for terms and conditions relating to use and distribution.

// This is an include file.  For a full description of the class and
// functions, see the file name with a "c" suffix.

#ifndef _CatTestResult_h
#define _CatTestResult_h 1

#include <Array.h>
#include <Array2.h>
#include <AugCategory.h>
#include <LogOptions.h>
#include <MEnum.h>
#include <CatDist.h>
#include <StatData.h>
#include <InstanceList.h>

class Categorizer;

extern const MString DEFAULT_NAMES_EXT;
extern const MString DEFAULT_TEST_EXT;

//@@ this is used by LearnCurve in addition to
//   ConfMat rename to DisplayOptionType??
enum DispConfMatType {no, ascii, scatterviz, both};
static const MEnum dispConfMatEnum = MEnum("no", no)
   << MEnum("ascii", ascii)
   << MEnum("scatterviz", scatterviz)
   << MEnum("both", both);

// CatOneTestResult and CatCounters should probably be nested classes,
//   but the compiler complains that templates on nested classes is
//   not implemented.  
struct CatOneTestResult {
private:
   NO_COPY_CTOR(CatOneTestResult);
public:
   // Public member data
   InstanceRC* instance;       // Pointer because InstanceRC
			       // constructor requires an argument, so then
			       // this could not be in an Array.
   AugCategory* augCat;        // Pointer in case AugCategory is subclassed.
   Category correctCat;        // What we should have guessed. 
   CatDist *predDist;          // Predicted category distribution
   CatDist *correctDist;       // Correct category distribution
   Bool inTrainIL;             // TRUE iff instance appeared in training
                               // InstanceList.

   // Methods
   CatOneTestResult();
   Bool operator==(const CatOneTestResult& rhs) const;	 
   Bool operator!=(const CatOneTestResult& rhs) const	 
      {return ! ((*this)== rhs);}
};

struct ScoringMetrics {
private:
   NO_DEFAULT_OPS(ScoringMetrics);
public:
   // Public member data
   int numCorrect;
   int numIncorrect;
   Real totalLoss;
   Real minimumLoss;
   Real maximumLoss;
   Real weightCorrect;
   Real weightIncorrect;
   Real meanSquaredError;
   Real meanAbsoluteError;
   Real totalLogLoss;

   // Methods
   ScoringMetrics();
};

// helper functions which accumulate the various types of metrics
// we use 
void compute_scoring_metrics(ScoringMetrics& metrics,
			     const AugCategory& predictedCat,
			     const AugCategory& correctCat,
			     const CatDist& predictedDist,
			     const CatDist& correctDist,
			     Real weight,
			     const SchemaRC& testSchema,
			     Bool computeLogLoss = FALSE);
void accumulate_scoring_metrics(ScoringMetrics& dest,
				const ScoringMetrics& src);

struct CatCounters {
private:
   NO_DEFAULT_OPS(CatCounters);
public:
   // Public member data
   Real numTestSet;      // Total weight of instances in test-set
                         // in this category.
   ScoringMetrics metrics;   // Evaluation metrics for this category

   // Methods
   CatCounters();
};


class LiftElement {
   NO_DEFAULT_OPS(LiftElement);

   // Member data
   Real score;
   Real weight;
   Real lift;   //if no loss function 0 (incorrect) 1 (correct)
public:
   LiftElement() : score(0), weight(0), lift(0) {}
   //Used for sort, want scores largest first in descending order
   Bool operator< (const LiftElement& rhs) const
      {return (score > rhs.get_score()); }
   Bool operator== (const LiftElement& rhs) const
      {return (score == rhs.get_score()); }
   Bool operator!= (const LiftElement& rhs) const
      {return ! ((*this)== rhs);}
   void set(Real s, Real w, Real l) { score = s; weight = w; lift = l; }
   Real get_score() const { return score; }
   Real get_weight() const { return weight; }
   Real get_lift () const { return lift; }
};
      

// if the InstanceList pointers are changed to references,
// there is a problem in the initialization that leads to run-time errors
class CatTestResult {
 public:
   // enumerated type used by error() to indicate set of test instances to
   //   report on.  Normal = all, Memorized = test instance in training set,
   //   Generalized = test instance not in training set.
   enum ErrorType { Normal, Generalized, Memorized };
private:  
   NO_DEFAULT_OPS(CatTestResult);
   LOG_OPTIONS;

   // Member data
   const InstanceList* trainIL;
   const InstanceList* testIL;
   int numOnTrain;
   Real weightOnTrain;
   ScoringMetrics metrics;
   StatData lossStats;
   Array<CatOneTestResult> results;
   
   Array<CatCounters> catDistrib;
   Array2<Real> confusionMatrix;
   Bool ownsTestIL;
   Bool ownsTrainIL;
   Bool inTrainingSetInitialized;

   static Bool computeLogLoss;

   // Private methods
   void check_for_unknown_classes (int &step, int &start) const;
protected:
   // Protected methods
   void initialize(const Categorizer& cat);
   void initialize();
   void initializeTrainTable() const;
   
public:
   static Real log_loss(const CatDist& predDist,
			const AugCategory& correctCat);
   static void check_for_unknown_classes(int &step, int &start,
					 const Array2<Real>& confusionMatrix);
   static void display_ascii_confusion_matrix(MLCOStream& stream,
					 const Array2<Real>& confusionMatrix,
					      const SchemaRC schema);
   
   void OK(int level = 1) const;
   CatTestResult(const Categorizer& cat,
                 const InstanceList& trainILSource,
                 const MString& testFile,
		 const MString& namesExtension = DEFAULT_NAMES_EXT,
	         const MString& testExtension = DEFAULT_TEST_EXT);
   CatTestResult(const Categorizer& cat,
                 const InstanceList& trainILSource,
                 const InstanceList& testILSource);
   CatTestResult(const Array<CatOneTestResult>& resultsArray,
		 const InstanceList& trainILSource,
                 const InstanceList& testILSource);
   virtual ~CatTestResult();

   void assign_instance_lists(InstanceList *& train,
			      InstanceList *& test);

   const InstanceList& get_training_instance_list() const {return *trainIL;}
   const InstanceList& get_testing_instance_list() const {return *testIL;}
   const Array<CatOneTestResult>& get_results() const {return results;}

   // metrics
   const ScoringMetrics& get_metrics() const { return metrics; }
   int num_test_instances() const;
   int num_train_instances() const;
   Real total_test_weight() const;
   Real total_train_weight() const;
   int num_on_train() const;
   int num_off_train() const;
   Real total_weight_on_train() const;	
   Real total_weight_off_train() const;
   int num_correct() const;
   int num_incorrect() const;
   Real total_correct_weight() const;
   Real total_incorrect_weight() const;
   Real total_label_weight(Category label) const;
   Real true_label_weight(Category label) const;
   Real false_label_weight(Category label) const;
   Real total_mean_squared_error() const;
   Real total_mean_absolute_error() const;
   Real total_loss() const;
   Real error(ErrorType errType = Normal) const;
   Real accuracy(ErrorType errType = Normal) const;
   Real minimum_loss() const;
   Real maximum_loss() const;
   Real normalized_loss() const;
   Real total_log_loss() const;
   const StatData& loss_stats() const { return lossStats; }
   Real mean_loss() const { return total_loss() / total_test_weight(); }
   Real std_dev_loss() const { return loss_stats().std_dev(); }
   void confidence_loss(Real& confLow, Real& confHigh,
		               Real z = CONFIDENCE_INTERVAL_Z) const
   { loss_stats().percentile(z, confLow, confHigh); }

   const Array2<Real>& get_confusion_matrix() const { return confusionMatrix; }

   virtual InstanceRC get_instance(int num) const;
   virtual const AugCategory& label(int num) const;
   virtual const AugCategory& predicted_label(int num) const;
   // display_* show the instance and both labels (except for
   //   display_correct_instances() which shows only one label).
   // Instances which were in the training set say "(In TS)" on
   //   the display line.
   virtual void display_all_instances(MLCOStream& stream = Mcout) const;
   virtual void display_incorrect_instances(MLCOStream& stream = Mcout) const;
   virtual void display_correct_instances(MLCOStream& stream = Mcout) const;
   // category distribution displays an array where each cell i
   //   consists of: (1) number of test instances in class i,
   //   (2) number of test instances correctly predicted as class i,
   //   (3) number of test instances incorrectly predicted as class i.
   virtual void display_category_distrib(MLCOStream& stream = Mcout) const;
   // The confusion matrix displays for row i column j, the number
   //   of instances classified as j that should have been classified
   //   as i.
   virtual void display_confusion_matrix(MLCOStream& stream = Mcout) const;
   virtual void display_ascii_confusion_matrix(MLCOStream& stream = Mcout)
      const { display_ascii_confusion_matrix(stream, confusionMatrix,
					     testIL->get_schema()); }

   virtual void display_scatterviz_confusion_matrix(MLCOStream& stream,
						    MLCOStream& data) const;
   // display gives statistics (not displays)
   virtual void display(MLCOStream& stream = Mcout) const;
   // display_all dumps everything (display + display_all_instances).
   virtual void display_all(MLCOStream& stream = Mcout) const;
   virtual void display_scatterviz_lift_curve(const Category& labelValue,
					      MString configFileName) const;
   virtual void display_vrml_scattergram(const MString& fileName) const;

   // theoretical_std_dev gives the estimated standard deviation according to
   //   the binomial model, which assumes every test instance is a Bernoulli
   //   trial, thus std-dev=sqrt(acc*(1-acc)/n)
   // we use a static function here so that classes which don't use
   // a CatTestResult (like BaseInducer can still access this function.
   static Real theoretical_std_dev(Real acc, Real n);
   static void confidence(Real& confLow, Real& confHigh,
		  Real error, Real n, Real z = CONFIDENCE_INTERVAL_Z);
   static Real pessimistic_error_correction(Real numErrors, Real totalWeight,
					    Real zValue);
   
   static void set_compute_log_loss(Bool b) { computeLogLoss = b; }
   static Bool get_compute_log_loss() { return computeLogLoss; }
};

DECLARE_DISPLAY(CatTestResult)


// Manipulators for proper display of results

// Standardized display for performances (expected to be in percentages)
inline MLCOStream& perf_display(MLCOStream& stream) {
   stream << setprecision(GlobalOptions::printPerfPrecision)
	  << fixed_reals;
   return stream;
}
   
#endif

