// MLC++ - Machine Learning Library in -*- C++ -*-
// See Descrip.txt for terms and conditions relating to use and distribution.

// This is an include file.  For a full description of the class and
// functions, see the file name with a "c" suffix.

#ifndef _Categorizer_h
#define _Categorizer_h 1

#include <class_id.h>

// Define class IDs for Categorizers.  Note that these are not enums
//   because we want to give users the ability to define more of these
//   without changing the header file.
#define CLASS_CONST_CATEGORIZER          CATEGORIZER_ID_BASE +  1
#define CLASS_MULTITHRESH_CATEGORIZER    CATEGORIZER_ID_BASE +  2
#define CLASS_THRESHOLD_CATEGORIZER      CATEGORIZER_ID_BASE +  3
#define CLASS_ATTR_CATEGORIZER           CATEGORIZER_ID_BASE +  4
#define CLASS_RDG_CATEGORIZER            CATEGORIZER_ID_BASE +  5
#define CLASS_BAD_CATEGORIZER            CATEGORIZER_ID_BASE +  6
#define CLASS_TABLE_CATEGORIZER          CATEGORIZER_ID_BASE +  7
#define CLASS_IB_CATEGORIZER             CATEGORIZER_ID_BASE +  8
#define CLASS_LAZYDT_CATEGORIZER         CATEGORIZER_ID_BASE +  9
#define CLASS_NB_CATEGORIZER	         CATEGORIZER_ID_BASE + 10
#define CLASS_PROJECT_CATEGORIZER        CATEGORIZER_ID_BASE + 11
#define CLASS_DISC_CATEGORIZER           CATEGORIZER_ID_BASE + 12
#define CLASS_ATTR_EQ_CATEGORIZER        CATEGORIZER_ID_BASE + 13
#define CLASS_DTREE_CATEGORIZER          CATEGORIZER_ID_BASE + 14
#define CLASS_BAGGING_CATEGORIZER        CATEGORIZER_ID_BASE + 15
#define CLASS_LINDISCR_CATEGORIZER       CATEGORIZER_ID_BASE + 16
#define CLASS_CASCADE_CATEGORIZER        CATEGORIZER_ID_BASE + 17
#define CLASS_STACKING_CATEGORIZER       CATEGORIZER_ID_BASE + 18
#define CLASS_ATTR_SUBSET_CATEGORIZER    CATEGORIZER_ID_BASE + 19
#define CLASS_MULTI_SPLIT_CATEGORIZER    CATEGORIZER_ID_BASE + 20
#define CLASS_ONE_R_CATEGORIZER          CATEGORIZER_ID_BASE + 21
#define CLASS_CONSTRUCT_CATEGORIZER      CATEGORIZER_ID_BASE + 22
#define CLASS_LEAF_CATEGORIZER           CATEGORIZER_ID_BASE + 24
#define CLASS_DISC_NODE_CATEGORIZER      CATEGORIZER_ID_BASE + 25
#define CLASS_MAJORITY_CATEGORIZER       CATEGORIZER_ID_BASE + 95
#define CLASS_ODT_CATEGORIZER            CATEGORIZER_ID_BASE + 96
#define CLASS_CLUSTER_CATEGORIZER        CATEGORIZER_ID_BASE + 97
#define CLASS_OPTION_CATEGORIZER         CATEGORIZER_ID_BASE + 98

#include <InstanceRC.h>
#include <SchemaRC.h>
#include <MLCStream.h>
#include <AugCategory.h>
#include <LogOptions.h>
#include <DynamicArray.h>

class NodeInfo; // needed for friend function
class InstanceList;
class DisplayPref;
class CatDist;

class MInStore;
class MOutStore;

// @@ can't include this, since it's implemented in MTree. We should remove
// all references to NodeCategorizer here once we can use dynamic_cast (-n32)
class NodeCategorizer;

// The PerCategorizer_ class represents the persistent data within
// a categorizer.  It can be written to or read from a stream, as well.
class PerCategorizer_ {
   NO_DEFAULT_OPS(PerCategorizer_);
   
public:
   int numCat;
   MString descr;
   Array<Real> *distrArray;
   Array<Real> *origDistr;
   Real totalWeight;
   MString extraVizText;

   PerCategorizer_() { }
   virtual ~PerCategorizer_() { }

   virtual void read(MInStore& in);
   virtual void write(MOutStore& out);
};


// A categorizer returns a category given an instance.
class Categorizer { // ABC
   NO_DEFAULT_OPS(Categorizer);
   LOG_OPTIONS;

   // Member data
   int numCat; // not const to support clustering methods
   MString descr; // ASCII description.  Subclasses may have
                        // optional descriptors (e.g. graphics)
   Array<Real> *distrArray; // Probability distribution of categories (label)
   Array<Real> *originalDistr; // as opposed to backfit distribution
   Real totalWeight;  // total weight represented in distribution
   SchemaRC schema;  // schema used to build this categorizer.
   MString extraVizText;  // any extra text to be output at end of
                          // visualization.
public:
   virtual void OK(int level  = 0) const;

   Categorizer(int noCat, const MString& dscr, const SchemaRC& schema);
   Categorizer(const Categorizer& source, CtorDummy);
   Categorizer(const PerCategorizer_& data, const SchemaRC& schema);

   virtual ~Categorizer();
   virtual int num_categories() const { return numCat; }
   virtual const MString& description() const { return descr; }

   // We want to set the description after the Categorizer is initialized.
   // This function is virtual so that wrappers can change the wrapped
   // categorizer's description as well (should they want to).
   virtual void set_description(const MString& val) { descr = val; }
   virtual void set_schema(const SchemaRC& sch) { schema = sch; }
   
   // Categorization and scoring.  All categorizers support simple
   // categorization.  If a categorizer can return a distribution of
   // labels, it should return TRUE from supports_scoring(), and
   // the score() function should be operational.
   virtual AugCategory categorize(const InstanceRC&) const = 0;
   virtual CatDist *score(const InstanceRC&) const {
      err << "Categorizer does not support scoring" << fatal_error;
      return NULL;
   }
   virtual Bool supports_scoring() const {
      return FALSE;
   }

   virtual Bool supports_backfit() const { return FALSE; }
   virtual void backfit(const InstanceList&) {
      err << "Categorizer::backfit is not implemented for subclass "
	  << class_id() << fatal_error;
   }

   virtual void display_struct(MLCOStream& stream,
			       const DisplayPref& dp) const = 0;
   // Returns a pointer to a deep copy of this Categorizer
   virtual Categorizer* clone() const = 0;

   // Build the probability distribution;
   virtual void build_distr(const InstanceList& instList);

   // Copy the probability distribution;
   virtual void set_distr(const Array<Real>& val);

   // Add the probability distribution;
   virtual void add_distr(const Array<Real>& val);   
   virtual void add_distr(int label, Real delta); // distr[label] += delta

   // Get the probability distribution.
   virtual const Array<Real>& get_distr() const
      { ASSERT(distrArray); return *distrArray; }

   // Get/set the original (before backfit) distribution. This should only be
   // set by the backfitting code. The set function needs to be public so
   // that wrappers can call it on their wrapped categorizers.
   virtual const Array<Real>& get_original_distr() const
   { ASSERT(originalDistr); return *originalDistr; }
   
   virtual void reset_original_distr()
   { delete originalDistr; originalDistr = NULL; }
   virtual void set_original_distr(const Array<Real>& dist);

   virtual Bool has_original_distr() const
   { return originalDistr != NULL; }

   // Return the existence (true/false) of the probability distribution.
   virtual Bool has_distr() const
      { return distrArray != NULL;}

   // Get the majority of distribution array
   virtual Category majority_category(Category zeroBreaker =
				      FIRST_CATEGORY_VAL) const ;

   // Returns the total weight used to build this categorizer
   virtual Real total_weight() const;

   // Returns the class id
   virtual int class_id() const = 0;

   // return the schema
   virtual SchemaRC get_schema() const { return schema; }

   // Note:  The following, though pure virtual, has an implementation
   //   (defined in Categorizer.c).  This implementation MUST be invoked
   //   by the operator== of all directly-derived classes.
   virtual Bool operator==(const Categorizer& rhs) const  = 0;
   
   Bool operator!=(const Categorizer& rhs) const
      {return ! ((*this)== rhs);}

   // make persistent
   virtual PerCategorizer_ *build_persistent_object() const {
      return new PerCategorizer_; }
   virtual void make_persistent(PerCategorizer_& data) const;

   virtual void project_schema(const BoolArray& mask) {
      schema = schema.project(mask);
   }

   virtual Bool supports_persistence() const { return FALSE; }

   void set_used_attr(DynamicArray<Bool>& usedAttr) const;

   virtual void set_used_attr(BoolArray&) const
   {
      err << "Categorizer::set_used_attr is not defined for the subclass "
	  << class_id() << fatal_error;
   }

   virtual void project(const BoolArray&)
   {
     err << "Categorizer::project is not defined for the subclass "
	  << class_id() << fatal_error;
   }

   // @@ just a hack now
   PTHR_DECL(void change_ownership() const {
      schema.change_ownership();
   };)

   virtual void short_display(ostream&) const; // Used in Graph::display

   // @@ This is a general function and might be useful outside the class
   int projected_attribute_number (const BoolArray& mask, int attrNum)
      const;

   // functions to manipulate extra visualization text
   virtual const MString& get_extra_viz_text() const { return extraVizText; }
   virtual void set_extra_viz_text(const MString& txt) { extraVizText = txt; }

   // For casting to NodeCategorizers. @@ This should go away when we switch
   // to -n32 (then we can use dynamic_cast instead of these)
   virtual Bool can_cast_to_node_categorizer() const { return FALSE; }
   virtual NodeCategorizer& cast_to_node_categorizer();
   virtual const NodeCategorizer& cast_to_node_categorizer() const;
};

#endif
