// MLC++ - Machine Learning Library in -*- C++ -*-
// See Descrip.txt for terms and conditions relating to use and distribution.

// This is an include file.  For a full description of the class and
// functions, see the file name with a "c" suffix.

#ifndef _Centroid_h
#define _Centroid_h 1

#define CENTROID_1D_REAL_ID           CENTROID_ID_BASE + 1
#define CENTROID_1D_NOMINAL_DISTR_ID  CENTROID_ID_BASE + 2
#define CENTROID_1D_NOMINAL_MODE_ID   CENTROID_ID_BASE + 3
#define CENTROID_MULTID_ID            CENTROID_ID_BASE + 4

#include <Attribute.h>
#include <Instance.h>
#include <InstanceList.h>
#include <Persistence.h>
#include <PersistentObject.h>
#include <MRandom.h>


class CentroidMultiD;
class Centroid1D;
class Centroid1DReal;
class Centroid1DNominalDistr;


class PerCentroid_ : public PersistentHelper_ {
public:
   Real totalWeight;
   int initMethod;
   
   void read(MInStore& in);
   void write(MOutStore& out);
};


class Centroid : public PersistentObject { // ABC
   RAND_OPTIONS;
public:
   enum NominalType { distribution, mode };
   enum InitMethod { none, uniform, random };
   enum UnknownMethod { mean, one  };

private:
   Real totalWeight;
   SchemaRC schema;

   InitMethod initMethod;

   LOG_OPTIONS;
   NO_COPY_CTOR(Centroid);

protected:
   virtual const SchemaRC& get_schema() const { return schema; }

   virtual void update_total_weight(Real weight)
      { totalWeight += weight; }
   virtual void update(const InstanceRC& instance, Real weight) = 0;

public:
   virtual void OK(int level=0) const;
   Centroid(const SchemaRC& schema, Real totalWeight = 0.0);
   Centroid(const Centroid& other, CtorDummy );
   Centroid(const SchemaRC& schema, PerCentroid_& helper);
   virtual Centroid* clone() const = 0;

   virtual ~Centroid() { DBG(OK()); }
   virtual Bool operator==(const Centroid& rhs) const = 0;
   Bool operator!=(const Centroid& rhs) const 
      { return(! this->operator==(rhs) ); }

   virtual Bool can_cast_to_multiD() const { return FALSE; }
   virtual Bool can_cast_to_1D() const { return FALSE; }
   virtual Bool can_cast_to_1DNominal() const { return FALSE; }
   virtual Bool can_cast_to_1DReal() const { return FALSE; }

   virtual Centroid1D* cast_to_1D();
   virtual Centroid1DReal* cast_to_1DReal();
   virtual Centroid1DNominalDistr* cast_to_1DNominal();
   virtual CentroidMultiD* cast_to_multiD();

   //  virtual const Centroid1D& cast_to_1D() const;
   virtual const Centroid1DReal& cast_to_1DReal() const;
   virtual const Centroid1DNominalDistr& cast_to_1DNominal() const;
   virtual const CentroidMultiD& cast_to_multiD() const;

   virtual void set_user_options(const MString& preFix = "");
   virtual void set_init_method( InitMethod initMeth );

   virtual InitMethod init_method() { return initMethod; }

   virtual Real total_weight() const { return totalWeight; }
   virtual Bool zero_weight() const {
      return mlc.approx_equal((StoredReal)totalWeight, (StoredReal)0.0); }
   
   virtual void display_short(MLCOStream& stream = Mcout) const = 0;
   virtual void display(MLCOStream& stream= Mcout) const;

   virtual PtrArray<Centroid*>* init_centroids( int numCentroids, 
				  const InstanceList& instances) = 0;

   virtual Real distance(const InstanceRC& inst) const = 0;
   virtual Real distance(const Centroid& other) const = 0;

   virtual Real distance_if_removed(const InstanceRC& instance);
   virtual Real distance_if_added(const InstanceRC& instance);

   virtual void add_to( const InstanceRC& instance ) 
      { update(instance, instance.get_weight()); }
   virtual void del_from(const InstanceRC& instance)
      { update(instance, -instance.get_weight()); }

   virtual void reset( ) { totalWeight = 0.0; }

   virtual void fill_helper(PersistentHelper_& helper) const;
};

DECLARE_DISPLAY(Centroid)

// Persistence support functions and macros for centroids
typedef Centroid *
      (*CentroidBuilder)(const SchemaRC&, PersistentHelper_*&);

// Define macros to help declare and define persistent objects

// Use this macro at the top of the public section of any centroid
#define DECLARE_CENTROID(classID, Object, Helper) \
  static int staticClassID_;  \
  static Centroid *create_this_type(const SchemaRC&, PerCentroid_*&);  \
  virtual int class_id() const { return classID; }  \
  virtual PersistentHelper_ *create_my_helper() { \
     return new Helper; }

// Use this macro at the top of the public section of the persistent
// helper class
#define DECLARE_CENTROID_HELPER(classID, Object, Helper) \
  static int staticClassID_; \
  static PersistentHelper_ *create_this_type(); \
  virtual int associated_id() const { return classID; } \
  virtual int class_id() const { return staticClassID_; }

// Use this macro at the top of the .c file containing both the
// object and the helper
// @@ GNU need the (void *) casts
#define DEF_CENTROID(classID, Object, Helper) \
  PersistentHelper_ *Helper::create_this_type() { \
    return new Helper; } \
  int Helper::staticClassID_ = \
    mlcFactory.register_class("PersistentHelper_", # Helper, \
			      (void *)&Helper::create_this_type); \
  Centroid *Object::create_this_type(const SchemaRC& schema, \
				     PerCentroid_*& helper) \
    { return new Object(schema, (Helper&)(*helper)); } \
  int Object::staticClassID_ = \
    mlcFactory.register_class("Centroid", # Object, \
			      (void *) &Object::create_this_type, classID, \
			      Helper::staticClassID_);
  
// This function builds a centroid given a schema and a full helper.
Centroid *
create_centroid(const SchemaRC& schema, PersistentHelper_ *& helper);
   
#endif
