// MLC++ - Machine Learning Library in -*- C++ -*-
// See Descrip.txt for terms and conditions relating to use and distribution.

// This is an include file.  For a full description of the class and
// functions, see the file name with a "c" suffix.

#ifndef _Cluster_h
#define _Cluster_h 1

// Define class IDs for Clusterss.  Note that these are not enums
//   because we want to give users the ability to define more of these
//   without changing the header file.

#define CLASS_CLUSTER_1D_ID      0
#define CLASS_CLUSTER_MULTID_ID  1

#include <Array2.h>
#include <InstanceList.h>
#include <Centroid.h>
#include <Persistence.h>
#include <DistribDistFunctor.h>
#include <MEnum.h>

class Cluster {   // ABC
public:
   enum ClusteringMethod { kmeans, kmeansaggregate, kmeanssplit,
			   centroid, linkage };
private:
   LOG_OPTIONS;
   NO_COPY_CTOR(Cluster);

   // Member data
   const MString dscr;

   // User options
   int maxIterations;
   ClusteringMethod clustMethod;
   Centroid::NominalType nominalType;
   Centroid::InitMethod initMethod;

   Real clusterPenalty;
   DistribDistFunctor* nomDistFunctor;
  
   // Data and centroids
   InstanceList *TS;
   PtrArray<Centroid*>* centroids; // A pointer since this could change, both
				   // in size and content.
   // Statistics
   Real averageDistance;
   Real stdDevDistance;

   Array<Real>* avgDistByCluster;
   Array<Real>* stdDevDistByCluster;
protected:
   // Protected methods
   // Accessors for subclasses
   virtual Centroid::NominalType nominal_type() { return nominalType; }
   virtual Centroid::InitMethod init_method() { return initMethod; }
   virtual DistribDistFunctor*& nominal_dist_functor(){return nomDistFunctor;};

   // delets and nulls everything relating to the centroids
   virtual void reset_centroids();

   // Releases ownership of TS, gets ownership of newTS.
public:
   virtual InstanceList* assign_data(InstanceList*& newTS);
protected:


   // Releases ownership of centroids, gets ownership of newCents.
   virtual PtrArray<Centroid*>*
      assign_centroids(PtrArray<Centroid*>*& newCents);

   // Returns copy of centroids
   virtual PtrArray<Centroid*>* get_copy_of_centroids() const;

   virtual Centroid* get_prototype() = 0;
  
   // Clustering methods
   virtual void k_means( Array<int>& whichClust );
   virtual void k_means_aggregate( Array<int>& whichClust );
   virtual void k_means_split( Array<int>& whichClust, int maxNamClusts );
   virtual void centroid_method( Array<int>& whichClust );

   virtual PtrArray<Centroid*>* merge_clusters( Array<int>& whichClust );
   virtual PtrArray<Centroid*>* split_clusters( Array<int>& whichClust );
   PtrArray<Centroid*>* linkage_merge(Array<int>& whichClust,
				      Array<int>& distIndex,
				      Array2<Real>& centDists );
   virtual void assign_insts_to_centroids( Array<int>& whichClusts );
   virtual void assign_insts_to_empty_clusters( Array<int>& whichClust );
   virtual void calculate_statistics( );
public:
   virtual void OK(int /* level */=0) const;
  
   // Con/De-structors

   Cluster(const MString& description);
   Cluster(const Cluster& source, CtorDummy);
   virtual Cluster* clone() = 0;

   virtual ~Cluster();

   virtual int class_id() const = 0;
   virtual Bool operator==( const Cluster& rhs ) const = 0;
   Bool operator!=( const Cluster& rhs ) const
      {return ! ((*this)== rhs);}

   // Set Options
   virtual void set_user_options(const MString& prefix = "");

   virtual void set_max_iters(int iters);
   virtual void set_clust_method( ClusteringMethod clustMeth );
   virtual void set_nominal_type( Centroid::NominalType nomType );
   virtual void set_init_method( Centroid::InitMethod initMeth );

   virtual void set_nominal_dist_funct( DistribDistFunctor *distFunct );
   virtual void set_cluster_penalty(Real clustPen ) 
      { clusterPenalty = clustPen; }

   virtual const MString& description() const { return dscr; };

   virtual int num_attr() const;
   virtual int num_centroids() const;
   virtual int num_instances() const;
   virtual Real cluster_penalty() const { return clusterPenalty; }

   // read_data reads the training set.  The file name suffices for the data
   // file and names file default to ".data" and ".names"
   // respectively.  If the files have a different prefix, give an
   // empty file name, and give the full file name as suffix.

   virtual void read_data(const MString& file,
			  const MString& namesExtension = DEFAULT_NAMES_EXT,
			  const MString& dataExtension = DEFAULT_DATA_EXT);

   // release_data gives you back the training list.
   virtual InstanceList* release_data();
   virtual PtrArray<Centroid*>* release_centroids();
   virtual Bool has_data(Bool fatalOnFalse = TRUE) const;
   virtual Bool has_centroids(Bool fatalOnFalse = TRUE) const;

   // A couple of accessors
   virtual const InstanceList& instance_list() const;
   virtual const PtrArray<Centroid *>& centroid_array() const;

   // Cluster does the real work, returning the centroids.
   virtual const PtrArray<Centroid *>& cluster(int numClusts);

   virtual Real average_distance() const;
   virtual void display_statistics( MLCOStream& stream ) const;

   // Persistence ( of centroids )
   virtual void read_centroids( const MString& file );
   virtual void write_centroids( const MString& file );

   // linkage methods
   void init_linkage(Array<int>& whichClust, 
		     Array<int>& clustIndex,
		     Array2<Real>& centDists );
   void update_linkage(Array<int>& whichClust,
		       Array<int>& clustIndex,
		       Array2<Real>& centDists,
		       int merged1, int merged2  );
   void linkage_method(Array<int>& whichClust );

};

#endif
