// MLC++ - Machine Learning Library in -*- C++ -*-
// See Descrip.txt for terms and conditions relating to use and distribution.

// This is an include file.  For a full description of the class and
// functions, see the file name with a "c" suffix.

#ifndef _ClusterCategorizer_h
#define _ClusterCategorizer_h 1

#include <Categorizer.h>
#include <Array2.h>
#include <Centroid.h>
#include <Persistence.h>
#include <DistribDistFunctor.h>
#include <DisplayPref.h>


class ClusterCategorizer : public Categorizer {
private:
   LOG_OPTIONS;
   NO_COPY_CTOR(ClusterCategorizer);

   // Distance function for nominal distributions
   DistribDistFunctor* nomDistFunctor;

public:
   enum DistanceType { euclidian, manhattan };
   enum ScoringMethod { inverseDistance, inverseSquareDistance };
   
private:
   // Member data
   // Overall distance type.  Euclidian or Manhattan
   DistanceType distanceType;

   // Scoring method.  This determines how to convert centroids into
   // probability estimates given a single instance.
   ScoringMethod scoringMethod;

   // Centroids.  Stored as dynamic array because the number can change.
   // However, changing the number of centroids requires changing the
   // Schema to match.
   // NULL entries are permitted in the centroid array.  In fact, the
   // array is initially NULL.  This gives Inducers more freedom in
   // how they create and manipulate centroids.  However, any attempt
   // to classify or score with a NULL centroid entry will abort.
   DynamicArray<Centroid *> centroids; 

   // These variables all store the results of computing statistics
   // on this set of centroids.  Perhaps this should become part of
   // a ClusterResult (analog to CatTestResult) class?? @@
   Real averageDistance;
   Real stdDevDistance;
   Array<Real>* avgDistByCluster;
   Array<Real>* stdDevDistByCluster;

   // Private methods
   // Helper: determine if we have any NULL centroids
   Bool has_null_centroids(Bool fatalOnTrue = TRUE) const;

   // Helper: determine if the categorizer has statistics computed
   Bool has_statistics(Bool fatalOnFalse = TRUE) const;
   
public:
   virtual void OK(int level=0) const;

   // Schema's number of label values determines number of clusters
   // These will initially be NULL because we can't tell at this
   // point what type of Centroid to create.
   ClusterCategorizer(const MString& description, const SchemaRC& schema);

   ClusterCategorizer(const ClusterCategorizer& source, CtorDummy);
   virtual Categorizer *clone() const
     { return new ClusterCategorizer(*this, ctorDummy); }
   virtual ~ClusterCategorizer();

   virtual int class_id() const { return CLASS_CLUSTER_CATEGORIZER; }
   virtual Bool operator==( const Categorizer& other ) const;

   // Accessors
   // ClusterCategorizer is a "dumb" class--it gives Inducers access
   // to almost all internals which are safe to manipulate.  Algorithmic
   // smarts should be in the Inducer classes (i.e. ClusterInducer)

   // Gives full access to centroids.  Note that we return them as an
   // Array instead of a DynamicArray.  This prevents outsiders from
   // changing the number of clusters through this method.
   virtual Array<Centroid *>& get_centroids() { return centroids; }
   
   // Set the Categorizer's schema.  This also resets all centroids
   // to NULL since the number may have changed.  The schema MUST have
   // a nominal label, which determines the number of clusters (centroids)
   // to create.
   void set_schema(const SchemaRC& schema);

   // Calculate statistics on the existing centroids.  It is illegal
   // to call this function if there are any NULL centroids present.
   void calculate_statistics(const InstanceList& il);
   
   // Categorization/scoring.  Categorization is based on the closest
   // neighbor centroid based on some distance metric.  Scoring computes
   // distance to all centroids and converts to a probability estimate.
   // Note: if categorize/score detect NULL centroids, they abort.
   virtual AugCategory categorize(const InstanceRC&) const;
   virtual CatDist *score(const InstanceRC&) const;
   virtual Bool supports_scoring() const { return TRUE; }
   
   // Display
   virtual void display_statistics(MLCOStream& stream) const;
   virtual void display_struct(MLCOStream& stream,
			       const DisplayPref& dp =
			         defaultDisplayPref) const;

   // Classifier Options

   // Sets the distance properties used by categorize/score
   // Distance properties include choice of overall distance metric
   // (Euclidian or Manhattan), and method for determining distance
   // between two nominals.
   virtual void set_nominal_dist_funct( DistribDistFunctor *distFunct );
   virtual DistribDistFunctor *get_nominal_dist_funct() const;
   virtual void set_distance_type( DistanceType distType)
      { distanceType = distType; }
   virtual DistanceType get_distance_type() const
      { return distanceType; }
   virtual void set_scoring_method( ScoringMethod method)
      { scoringMethod = method; }
   virtual ScoringMethod get_scoring_method() const
      { return scoringMethod; }

   // grab results of statistics
   virtual Real average_distance()
      { has_statistics(); return averageDistance; }
   virtual Real std_dev_distance()
      { has_statistics(); return stdDevDistance; }
   virtual const Array<Real>& average_distance_by_cluster()
      { has_statistics(); return *avgDistByCluster; }
   virtual const Array<Real>& std_dev_distance_by_cluster()
      { has_statistics(); return *stdDevDistByCluster; }
};

#endif
