// MLC++ - Machine Learning Library in -*- C++ -*-
// See Descrip.txt for terms and conditions relating to use and distribution.

// This is an include file.  For a full description of the class and
// functions, see the file name with a "c" suffix.

#ifndef _ClusterInducer_h
#define _ClusterInducer_h 1

#include <Inducer.h>
#include <Array2.h>
#include <InstanceList.h>
#include <Centroid.h>
#include <ClusterCategorizer.h>
#include <DistribDistFunctor.h>
#include <MEnum.h>

class ClusterInducer : public Inducer {
public:
   enum Algorithm { kMeans, kMeansAggregate,
		    kMeansSplit, centroid, linkage };
   enum Dimensions { single, multiple };

private:
   NO_COPY_CTOR(ClusterInducer);
   LOG_OPTIONS;

   // Member data (also see protected data)
   // User options
   int maxIters;
   Algorithm algorithm;
   Dimensions dimensions;
   int numClusters;            // for algorithms requiring an exact number
   int lowerNumClusters;       // lower bound for some algorithms
   int upperNumClusters;       // upper bound for some algorithms
   Centroid::NominalType nominalType;
   Centroid::InitMethod initMethod;
   ClusterCategorizer::DistanceType distanceType;
   ClusterCategorizer::ScoringMethod scoringMethod;
   Real clusterPenalty;
   DistribDistFunctor* nominalDistFunctor;

   // Options to be set after we have data
   int clusterAttrNum;         // for one-dimensional clustering
   Array<Real> *attrWeights;   // for multi-dimensional clustering
   
protected:
   // Protected member data
   ClusterCategorizer *clusterCat;

   // Protected member functions
   // prototype/initial centroid creation
   virtual Centroid* build_prototype();

   // Statistics functions
   virtual Real get_fitness() const;

   // Output schema management functions
   SchemaRC create_output_schema(int numClusts) const;
   SchemaRC split_output_schema(const SchemaRC& oldSchema,
				int splitCluster) const;
   SchemaRC merge_output_schema(const SchemaRC& oldSchema,
				int merge1, int merge2) const;

   // Splitting and merging
   ClusterCategorizer *split_clusters( Array<int>& whichClust );
   ClusterCategorizer *merge_clusters( Array<int>& whichClust );
   ClusterCategorizer *linkage_merge( Array<int>& whichClust,
				      Array<int>& distIndex,
				      Array2<Real>& clustDist );  
   
   // Clustering methods
   void assign_insts_to_centroids(Array<int>& whichClust);
   void assign_insts_to_empty_clusters(Array<int>& whichClust);
   virtual void k_means( Array<int>& whichClust );
   virtual void k_means_aggregate( Array<int>& whichClust );
   virtual void k_means_split( Array<int>& whichClust );
   virtual void centroid_method( Array<int>& whichClust );

   // linkage-based clustering methods
   void init_linkage( Array<int>& whichClust, 
		      Array<int>& clustIndex,
		      Array2<Real>& centDists );
   void update_linkage( Array<int>& whichClust,
			Array<int>& clustIndex,
			Array2<Real>& centDists,
			int merged1, int merged2  );
   void linkage_method( Array<int>& whichClust );

   // copy options to and from the categorizer
   void copy_options_to_categorizer();
   void copy_options_from_categorizer();
   
public:
   virtual void OK(int level=0) const;
  
   ClusterInducer(const MString& description);
   ClusterInducer(const ClusterInducer& source, CtorDummy);
   virtual Inducer *copy() const {
      return new ClusterInducer(*this, ctorDummy); }
   virtual ~ClusterInducer();
   virtual int class_id() const { return CLUSTER_INDUCER; }

   // Override assign_data/release_data so attr weights can be managed
   virtual InstanceList* assign_data(InstanceList*& newTS);
   virtual InstanceList* release_data();
   
   // Categorizer management
   virtual Bool was_trained(Bool fatalOnFalse = TRUE) const;
   virtual const Categorizer& get_categorizer() const
     { was_trained(); return *clusterCat; }
   virtual Categorizer *release_categorizer();
   virtual void assign_categorizer(Categorizer*& newCat);
   
   // Option management
   virtual void set_user_options(const MString& prefix = "");

   virtual void set_max_iters(int iters);
   virtual int get_max_iters() const { return maxIters; }

   virtual void set_algorithm(Algorithm algo) { algorithm = algo; }
   virtual Algorithm get_algorithm() const { return algorithm; }

   virtual void set_dimensions(Dimensions dim) { dimensions = dim; }
   virtual Dimensions get_dimensions() const { return dimensions; }

   virtual void set_num_clusters(int nc) { numClusters = nc; }
   virtual int get_num_clusters() const { return numClusters; }

   virtual void set_lower_num_clusters(int lnc)
      { lowerNumClusters = lnc; }
   virtual int get_lower_num_clusters() const { return lowerNumClusters; }

   virtual void set_upper_num_clusters(int unc)
      { upperNumClusters = unc; }
   virtual int get_upper_num_clusters() const { return upperNumClusters; }
   
   virtual void set_nominal_type(Centroid::NominalType nomType)
      { nominalType = nomType; }
   virtual Centroid::NominalType get_nominal_type() const
      { return nominalType; }

   virtual void set_init_method(Centroid::InitMethod initMeth)
      { initMethod = initMeth; }
   virtual Centroid::InitMethod get_init_method() const
      { return initMethod; }

   virtual void set_distance_type(ClusterCategorizer::DistanceType dt)
      { distanceType = dt;  copy_options_to_categorizer(); }
   virtual ClusterCategorizer::DistanceType get_distance_type() const
      { return distanceType; }  

   virtual void set_scoring_method(ClusterCategorizer::ScoringMethod sm)
      { scoringMethod = sm;  copy_options_to_categorizer(); }
   virtual ClusterCategorizer::ScoringMethod get_scoring_method() const
      { return scoringMethod; }  

   virtual void set_cluster_penalty(Real penalty)
      { clusterPenalty = penalty; }  
   virtual Real get_cluster_penalty() const { return clusterPenalty; }
   
   virtual void set_nominal_dist_funct( DistribDistFunctor *distFunct );
   virtual DistribDistFunctor *get_nominal_dist_funct() const
      { return nominalDistFunctor; }  

   virtual void set_cluster_attr_num(int attrNum);
   virtual int get_cluster_attr_num() const;

   virtual void set_cluster_attr_weight(int attr, Real weight);
   virtual Real get_cluster_attr_weight(int attr) const;

   // train--note: for 1D clustering, user must call set_cluster_attr_num
   // before training.
   virtual void train();
   
};

#endif
