// MLC++ - Machine Learning Library in -*- C++ -*-
// See Descrip.txt for terms and conditions relating to use and distribution.

// This is an include file.  For a full description of the class and
// functions, see the file name with a "c" suffix.

#ifndef _DDTInducer_h
#define _DDTInducer_h 1

#include <Inducer.h>
#include <Splitter.h>
#include <SplitSelector.h>
#include <Pruner.h>
#include <Array.h>
#include <LogOptions.h>
#include <MEnum.h>

class DTCategorizer;
class Categorizer;
class DecisionTree;
class SubtreeInfo;

class DDTInducer : public Inducer {
   NO_DEFAULT_OPS(DDTInducer);
   LOG_OPTIONS;
public:
   enum PruningMethod { none, confidence };
   static MEnum pruningMethodEnum;

   enum LeafDistType { allOrNothing, frequencyCounts,
                       laplaceCorrection, evidenceProjection };
   static MEnum leafDistTypeEnum;
   
private:
   Real totalTrainingSetWeight; // may be used for progress estimation
   Bool hasRealAttributes;      // so that we know when to transpose   

   DynamicArray<Bool> multiValAttributes; // for gain ratio computation

   // options
   int maxLevel;
   SplitScore::SplitScoreCriterion splitCriterion;
   Real lowerBoundMinSplit;
   Real upperBoundMinSplit;
   Real minSplitPercent;
   Bool nominalLowerBoundOnly;
   PruningMethod pruningMethod;
   Real pruningFactor;
   Bool emptyUnknownEdges;
   LeafDistType leafDistType;
   Real leafDistParameter;
   Bool parentTieBreaking;
   Bool adjustThresholds;
   Bool continuousMDLPenalty;
   Bool debugDisplay;
   
   // These actually do the work
   PtrArray<Splitter*> *splitters;
   SplitSelector *selector;
   PtrArray<Pruner*> *pruners;

   // The categorizer that we'll generate in train()
   DTCategorizer *categorizer;

protected:
   virtual Bool multi_val_attribute(int attrNum) const;
   virtual void set_global_training_info();
   virtual void accumulate_subtree_info(SubtreeInfo& subtreeInfo,
					const SubtreeInfo& childInfo) const;
   virtual void initialize_subtree_info(SubtreeInfo& subtreeInfo) const;
   virtual void get_split_suggestions(SubtreeInfo& subtreeInfo,
				      DblLinkList<SplitSuggestion*>&
				      suggestions) const;
   virtual NodePtr induce_decision_tree(SubtreeInfo& subtreeInfo) const;
   
public:
   virtual void OK(int level = 0) const;

   DDTInducer(const MString& description);
   DDTInducer(const DDTInducer& rhs, CtorDummy);
   virtual ~DDTInducer();

   virtual Bool is_initialized(Bool fatalOnFalse = TRUE) const;
   virtual Bool was_trained(Bool fatalOnFalse = TRUE) const;
   virtual const Categorizer& get_categorizer() const; 
   virtual Categorizer *release_categorizer();

   virtual int class_id() const { return DDT_INDUCER; }
   
   virtual Inducer* clone() const { return new DDTInducer(*this, ctorDummy); }
   virtual Inducer* copy() const { return clone(); }

   void reset();

   void initialize(PtrArray<Splitter*>*& newSplitters, SplitSelector*& newSel,
		   PtrArray<Pruner*>*& newPruners);

   virtual void build_categorizer(DecisionTree*& dTree, NodePtr rootNode);

   virtual void train();

   virtual const DTCategorizer& get_dt_categorizer() const
   { was_trained(TRUE); return *categorizer; }

   // @@ These two functions are questionable, since we don't actually set the
   // values of the variables until we call set_global_training_info. It would
   // be better if we were to intercept assign_data and read_data from
   // BaseInducer.
   Real get_total_training_set_weight() const
   { has_data(TRUE); return totalTrainingSetWeight; }
   Bool get_has_real_attributes() const
   { has_data(TRUE); return hasRealAttributes; }
   
   int num_nodes() const;
   int num_nontrivial_nodes() const;
   int num_leaves() const;
   int num_nontrivial_leaves() const;

   // option sets, so that users don't have to call initialize themselves
   virtual void set_default_functors();

   virtual void set_default_options();
   virtual void set_c45_options();
   virtual void set_user_options(const MString& prefix);

   virtual void display_options(MLCOStream& stream = Mcout,
				const DisplayPref& dp =
				DisplayPref::ASCIIDisplay) const; 

   // options
   void set_max_level(int maxDepth);
   int get_max_level() const { return maxLevel; }

   void set_split_by(SplitScore::SplitScoreCriterion criterion)
   { splitCriterion = criterion; }
   SplitScore::SplitScoreCriterion get_split_by() const
   { return splitCriterion; }

   void set_lower_bound_min_split(Real lbound);
   Real get_lower_bound_min_split() const { return lowerBoundMinSplit; }

   void set_upper_bound_min_split(Real ubound);
   Real get_upper_bound_min_split() const { return upperBoundMinSplit; }

   void set_min_split_percent(Real percent);
   Real get_min_split_percent() const { return minSplitPercent; }

   void set_nominal_lbound_only(Bool b) { nominalLowerBoundOnly = b; }
   Bool get_nominal_lbound_only() const { return nominalLowerBoundOnly; }

   void set_pruning_method(PruningMethod method);
   PruningMethod get_pruning_method() const { return pruningMethod; }

   void set_pruning_factor(Real factor);
   Real get_pruning_factor() const { return pruningFactor; }

   void set_empty_unknown_edges(Bool b) { emptyUnknownEdges = b; }
   Bool get_empty_unknown_edges() const { return emptyUnknownEdges; }

   void set_leaf_dist_type(LeafDistType leafType) { leafDistType = leafType; }
   LeafDistType get_leaf_dist_type() const { return leafDistType; }

   void set_leaf_dist_parameter(Real parameter);
   Real get_leaf_dist_parameter() const { return leafDistParameter; }

   void set_parent_tie_breaking(Bool b) { parentTieBreaking = b; }
   Bool get_parent_tie_breaking() const { return parentTieBreaking; }

   void set_adjust_thresholds(Bool b) { adjustThresholds = b; }
   Bool get_adjust_thresholds() const { return adjustThresholds; }
   
   void set_cont_mdl_adjust(Bool b) { continuousMDLPenalty = b; }
   Bool get_cont_mdl_adjust() const { return continuousMDLPenalty; }

   void set_debug_display(Bool b) { debugDisplay = b; }
   Bool get_debug_display() const { return debugDisplay; }
};

#endif
