// MLC++ - Machine Learning Library in -*- C++ -*-
// See Descrip.txt for terms and conditions relating to use and distribution.

// This is an include file.  For a full description of the class and
// functions, see the file name with a "c" suffix.

#ifndef _DiscDispatch_h
#define _DiscDispatch_h 1

#include <EntropyDisc.h>

class RealDiscretizor;

class DiscDispatch {
   LOG_OPTIONS;
   NO_DEFAULT_OPS(DiscDispatch);
public:
   enum DiscretizationType { oneR, uniformRange, uniformWeight,
			     entropy, c45Disc, t2Disc};
   // AlgoHeuristic means algorithm specific.
   enum NumBins { AlgoHeuristic, FixedValue, MDL};
private:
   // Member data (also see private data) 
   Array<int>* discVect;
   SchemaRC* schema;
   PtrArray<RealDiscretizor*>* disc;
   DiscretizationType discType;
   NumBins binHeuristic;
   EntropyDiscretizor::StopCrit stoppingCriterion;
   Real lowerBoundMinSplitWeight; //used for entropy
   Real upperBoundMinSplitWeight; //used for entropy
   Real minSplitWeightPercent;    //used for entropy
   int initVal;                   //initial discretization value
   int minInstPerLabel;           //used for 1R
   int numPast;                   //used for normalized entropy/gain ratio
   Bool pessError;          //used for normalized entropy/gain ratio
   Real zValue;                   //used for normalized entropy/gain ratio
   Real trimFraction;             //used for trimming uniform-range/weight

   // Private methods
   void set_disc_vector(const InstanceList& instList);
   void allocate(const InstanceList& instList);
   void free();
public:
   // Public data
   // defaults
   static const Real defaultLowerBoundMinSplitWeight;
   static const Real defaultUpperBoundMinSplitWeight;
   static const Real defaultMinSplitWeightPercent;
   static const int defaultNumIntervals;
   static const EntropyDiscretizor::StopCrit defaultStopCrit;
   static const int defaultNumPast;
   static const Bool defaultPessError;
   static const Real defaultZValue;
   static const Real defaultTrimFraction;

   // Methods
   DiscDispatch();
   ~DiscDispatch() { free(); }
   virtual void display(MLCOStream& stream = Mcout) const;
   void set_user_options(const MString& prefix);
   void set_disc_vect(Array<int>*& dv); //gets ownership
   const Array<int>* disc_vect() const;
   PtrArray<RealDiscretizor*>* discretizors();
   // @@ This should probably return a const ref and maybe
   // @@ have a different name
   const PtrArray<RealDiscretizor*>* discretizors() const;
   void OK(int level = 0) const;
   void set_bin_heuristic(NumBins h);
   NumBins get_bin_heuristic() const {return binHeuristic;}
   int get_initial_val() const { return initVal;}
   Real get_lower_bound_min_split_weight() const
   { return lowerBoundMinSplitWeight; }
   Real get_upper_bound_min_split_weight() const
   { return upperBoundMinSplitWeight; }
   Real get_min_split_weight_percent() const
   { return minSplitWeightPercent; }
   int get_min_inst() const { return minInstPerLabel; }
   int get_num_past() const { return numPast;}
   Bool get_pessimistic_error() const { return pessError; }
   Real get_z_value() const { return zValue; }
   DiscretizationType get_disc_type() const { return discType; }
   EntropyDiscretizor::StopCrit get_stopping_criterion() const
   { return stoppingCriterion;}
   Real get_trim_fraction() const { return trimFraction; }
   void set_initial_val(int newVal);
   void set_lower_bound_min_split_weight(Real in)
   { lowerBoundMinSplitWeight = in; }
   void set_upper_bound_min_split_weight(Real in)
   { upperBoundMinSplitWeight = in; }
   void set_min_split_weight_percent(Real in)
   { minSplitWeightPercent = in; }
   void set_min_inst(int min) {minInstPerLabel = min;}
   void set_num_past(int num) {numPast = num;}
   void set_pessimistic_error(Bool use) {pessError = use;}
   void set_z_value(Real val) { zValue = val; }
   void set_disc_type(DiscretizationType discType);
   void set_stopping_criterion(EntropyDiscretizor::StopCrit crit)
   { stoppingCriterion = crit; }
   void set_trim_fraction(Real val) { trimFraction = val; }
   InstanceList* discretize_instance_list(const InstanceList& instList); 
   // must call before discretize_instance_list
   void create_discretizors(const InstanceList& instList);   
   PtrArray<RealDiscretizor*>* disc_copy(); 
   DiscDispatch *make_shallow_copy() const;
};

DECLARE_DISPLAY(DiscDispatch)

#endif

