// MLC++ - Machine Learning Library in -*- C++ -*-
// See Descrip.txt for terms and conditions relating to use and distribution.

// This is an include file.  For a full description of the class and
// functions, see the file name with a "c" suffix.

#ifndef DynamicArray_h
#define DynamicArray_h

#include <Array.h>

template <class Element>
class DynamicArray : public Array<Element> {
   NO_COPY_CTOR(DynamicArray);
protected:
   // Protected member data
   Element initialValue;
   int sizeAllocated; // Physically allocated size. Note that this size is
		      //   different from arraySize which is the greatest
                      //   size that actually has ever been accessed.
   Bool wasInitialized;
   int incrementSize;

   // Protected methods
   //   These are virtual because it makes sense to override them.
   //   See RandCharArray
   virtual int increase_size();
   virtual void extend_array(int sizeRequested);
public:
   DynamicArray(const DynamicArray& source, CtorDummy);
   DynamicArray(int base, int size, const Element& initialVal);   
   DynamicArray(int size, const Element& initialVal);
   DynamicArray(int size = 0);

   virtual ~DynamicArray() {}

   void set_increment_size(int incrSize);
   int get_increment_size() const {return incrementSize;}
   DynamicArray& operator=(const DynamicArray& source);
   // *** WARNING WARNING WARNING WARNING WARNING WARNING ***
   // These two functions return references that are valid only
   //    until the array extends itself.  Once that happens,
   //    the references are invalid.  For example, the line
   //         int s = a.size(); a.index(s) = a.index(s-1)
   //    is incorrect because index(s-1) returns a reference to an
   //    element, but index(s) extends the array and if the
   //    array will actually extend, (delete/new/copy), the first
   //    reference points to freed memory.
   //    A common solution is to do (void)index(s); to make
   //    sure the array first extends.
   Element& index(int i);
   Element& operator[](int i);

   // These two methods do not have dynamic feature, because they
   //   are const functions. So just call the base methods.
   // Note that due to ObjectCenter bug, you cannot inline and not
   //   inline based on const alone, so we force these in the .c file
   Array<Element>& operator=(const Array<Element>& source) {
      return Array<Element>::operator=(source);}
   const Element& index(int i) const {
      return Array<Element>::index(i); }
   const Element& operator[](int i) const {
      return Array<Element>::operator[](i); }
   void append(const Array<Element>& other);
   void truncate(int size);

   void insert(const Element& elem, int pos);
   void erase(int start, int size=1);
};

// GNU compiler is unable to find the operator<< for Array
#if defined(GNU)
template <class Element>
DECLARE_DISPLAY(DynamicArray<Element>)

// GNU's disambiguation policies aren't good enough to get this
//   bounds_are_equal
template <class Element>    
Bool bounds_are_equal(const DynamicArray<Element>& lhs,
		      const Array<Element>& rhs)
   { return bounds_are_equal((const Array<Element>&) lhs, rhs); }

template <class Element>   
Bool bounds_are_equal(const Array<Element>& lhs,
		      const DynamicArray<Element>& rhs)
{ return bounds_are_equal(lhs, (const Array<Element>&) rhs); }

template <class Element>   
Bool bounds_are_equal(const DynamicArray<Element>& lhs,
		      const DynamicArray<Element>& rhs)
   { return bounds_are_equal((const Array<Element>&) lhs,
			     (const Array<Element>&) rhs); }

// operator==
template <class Element> 
Bool operator==(const Array<Element>& lhs, const DynamicArray<Element>& rhs)
   { return lhs==(const Array<Element>&)rhs;}

template <class Element> 
Bool operator==(const DynamicArray<Element>& lhs, const Array<Element>& rhs)
   { return (const Array<Element>&)lhs==rhs;}

template <class Element> 
Bool operator==(const DynamicArray<Element>& lhs,
		const DynamicArray<Element>& rhs)
   { return (const Array<Element>&)lhs==(const Array<Element>&)rhs;}

// operator !=
template <class Element>
Bool operator!=(const Array<Element>& lhs, const DynamicArray<Element>& rhs)
   { return lhs!=(const Array<Element>&)rhs;}

template <class Element> 
Bool operator!=(const DynamicArray<Element>& lhs, const Array<Element>& rhs)
   { return (const Array<Element>&)lhs!=rhs;}

template <class Element> 
Bool operator!=(const DynamicArray<Element>& lhs,
		const DynamicArray<Element>& rhs)
   { return (const Array<Element>&)lhs!=(const Array<Element>&)rhs;}
#endif
   
#ifdef COMPILE_TIME_TEMPLATES
#include "DynamicArray.ct"
#endif

#endif 


