// MLC++ - Machine Learning Library in -*- C++ -*-
// See Descrip.txt for terms and conditions relating to use and distribution.

// This is an include file.  For a full description of the class and
// functions, see the file name with a "c" suffix.

#ifndef _EntropyDiscretizor_h
#define _EntropyDiscretizor_h 1

#include <RealDiscretizor.h>

class EntropyDiscHelper {
   NO_DEFAULT_OPS(EntropyDiscHelper);

   // Member data
   RealAndLabelColumn* realAndLabelCol;	// instances
   Real origEntropy;		// the entropy of the whole
   Real bestEntropy;		// best entropy for the best split
   Real bestGain;		// best gain; can not be negative
   int bestSplitIndex;		// best index of realAndLabelCol to split on
   StoredReal bestThreshold;    // best threshold to split
   Real leftEntropy;		// entropy of the left part
   Real rightEntropy;		// entropy of the right part

   Array<Real> leftCount;
   Array<Real> rightCount;
   Array<Real> unknownCount;
   
   // Private methods
   void set_no_split_found(Array2<Real>& splitAndLabelDist,
			   Array<Real>& splitDist);
   void find_best_split(Real minSplit);
   void build_distributions(Array2<Real>*& splitAndLabelDist,
			    Array<Real>*& splitDist,
			    const DynamicArray<EntropyDiscHelper*> &splits)
      const;  

   // See Effective C++ by S. Meyers, page 107
   enum { defaultSplitIndex = 0 };

public:
   EntropyDiscHelper(RealAndLabelColumn*& column, Real minSplit);
   ~EntropyDiscHelper() { delete realAndLabelCol; }

   // Best split index can not be zero if a split exists
   Bool best_split_exists() const {
      return (bestSplitIndex != defaultSplitIndex);
   };
   Real best_threshold() const {
      ASSERT(best_split_exists());
      return bestThreshold;
   };
   Real best_gain() const {
      ASSERT(best_split_exists());
      return bestGain;
   };

   // Split the InstanceList and estimate the minimum description length
   //   condition.
   EntropyDiscHelper* split(Real minSplit);

   // Conditions checked during split creation.
   Bool mdl_condition() const;
   Bool error_condition(Real inZValue) const;
   Bool norm_mut_info_condition(const DynamicArray<EntropyDiscHelper*> &splits,
				Real totalEntr, int totalInsts,
				Real &lastVal) const;
   Bool gain_ratio_condition(const DynamicArray<EntropyDiscHelper*> &splits,
			     Real totalEntr, int totalInsts,
			     Real &lastVal) const;
   
   
   // Block further splits from this set.
   void block_further_splits()
   { bestSplitIndex = defaultSplitIndex;
     bestGain = -REAL_MAX; }

   // Accessors
   Real get_total_known_weight() const
   { return realAndLabelCol->known_weight(); }
   Real get_total_unknown_weight() const
   { return realAndLabelCol->unknown_weight(); }
   Real get_total_weight() const
   { return realAndLabelCol->total_weight(); }
   Real get_original_entropy() const
   { return origEntropy; }
   
   // Needed for finding max.
   Bool operator>(const EntropyDiscHelper& other) const
   { return bestGain > other.bestGain; };

   void display(MLCOStream& stream = Mcout) const;

};

DECLARE_DISPLAY(EntropyDiscHelper)


class EntropyDiscretizor : public RealDiscretizor{
   NO_DEFAULT_OPS(EntropyDiscretizor);
public:
   enum StopCrit { MDL, normMutInfo, gainRatio };
private:
   // Member data
   Real minSplit; // minimum number of instances in a split
   int numIntervals;
   int numPast;
   Bool pessError;
   Real zValue; // used in pessimistic_error_correction
   StopCrit crit;

   // Private methods
    int class_id() const { return CLASSID_ENTROPYDISC;}
   void auto_mdl_thresholds(DynamicArray<EntropyDiscHelper*> &splits,
			    DynamicArray<Real> &threshVals);
   void auto_non_mdl_thresholds(DynamicArray<EntropyDiscHelper*> &splits,
				DynamicArray<Real> &threshVals);
   void build_num_intervals_thresholds(DynamicArray<EntropyDiscHelper*> &splits,
				       DynamicArray<Real> &threshVals);
   Bool try_split(DynamicArray<EntropyDiscHelper*> &splits,
		  Real &lastVal, Real totalWeight, Real totalEntr,
		  Bool &errorFailed);

protected:
   // Protected methods
   void build_thresholds(RealAndLabelColumn*& column, Real trim);

public:
   //@@ void column_thresholds(RealAndLabelColumn*& column);
   EntropyDiscretizor(int attrNum, const SchemaRC& schema);
   EntropyDiscretizor(const EntropyDiscretizor& disc, CtorDummy);
   virtual Bool operator==(const RealDiscretizor& rhs) const;
   RealDiscretizor* clone() const; //returns copy of self

   //member access functions
   int get_num_intervals() const
   { return numIntervals; }
   void set_num_intervals(int val)
   { numIntervals = val; }
   int get_num_past() const
   { return numPast; }
   void set_num_past(int val)
   { numPast = val; }
   StopCrit get_stopping_crit() const
   { return crit; }
   void set_stopping_crit(StopCrit inCrit)
   { crit = inCrit; }
   Real get_min_split() const { return minSplit;}
   void set_min_split(Real aMinSplit) { minSplit = aMinSplit; }
   Bool get_pessimistic_error() const
   { return pessError; }
   void set_pessimistic_error(Bool val)
   { pessError = val; }
   Real get_z_value() const
   { return zValue; }
   void set_z_value(Real val)
   { zValue = val; }
};

#ifdef PTHREADS

// Class to wrap pthread calls nicely (see MPthread.h)

class EDThread : public MPthread {
   NO_DEFAULT_OPS(EDThread);

   LOG_OPTIONS;

   // all data for the discretization are here
   int attrNumber;
   const SchemaRC& schema;
   const int attrIntervals;
   const int inPast;
   const EntropyDiscretizor::StopCrit inCrit;
   const Real minSplit;
   const Bool inError;
   const Real inZValue;
   RealAndLabelColumn* realAndLabelCol;
   const Real columnTrim;

   // @@ need to free it in the parent space
   EntropyDiscretizor* entropyDisc;

protected:
   // virtual methods
   void do_work();
   void copy_data();

public:
   // the constructor has all the data
   EDThread(ThreadNo num, const Real dribbleScale, const int number, const
	    SchemaRC& aSchema, const int intervals, const int past, const
	    EntropyDiscretizor::StopCrit crit, const Real mSplit, const Bool
	    error, const Real zvalue, const LogOptions& logOptions,
	    RealAndLabelColumn*& column, const Real trim);
   ~EDThread();

   RealDiscretizor* get_result() { return entropyDisc; }

};

#endif /* PTHREADS */

PtrArray<RealDiscretizor*>*
create_entropy_discretizors(const LogOptions& logOptions,
			    const InstanceList& instList,
			    Real inLowerBoundMinSplitWeight,
			    Real inUpperBoundMinSplitWeight,
			    Real inMinSplitWeightPercent,
			    int inPast, Bool inError, Real inZValue,
			    EntropyDiscretizor::StopCrit inCrit,
			    const Array<int>& attrIntervals,
			    const Array<Real>& columnTrim);


PtrArray<RealDiscretizor*>*
create_entropy_discretizors(const LogOptions& logOptions,
			    const InstanceList& instList,
			    Real inLowerBoundMinSplitWeight,
			    Real inUpperBoundMinSplitWeight,
			    Real inMinSplitWeightPercent,
			    int inPast, Bool inError, Real inZValue,
			    EntropyDiscretizor::StopCrit inCrit,
			    int numIntervalsForEachAttr = 0,
			    Real trim = 0);


#endif




