// MLC++ - Machine Learning Library in -*- C++ -*-
// See Descrip.txt for terms and conditions relating to use and distribution.

// This is an include file.  For a full description of the class and
// functions, see the file name with a "c" suffix.

#ifndef _ErrorData_h
#define _ErrorData_h 1

#include <StatData.h>


// ErrorData is a class which encapsulates statistical data collected
// during error estimation.  Each ErrorData represents a single
// performance metric, while the combined class PerfData represents
// a full suite of performance metrics.
class ErrorData : public StatData {
   NO_COPY_CTOR(ErrorData);

   // Member data (also see public data)
   // test set error--stored here if provided
   Real testSetError;
   Real totalTestWeight;

   // flag determines if values stored here are bounded from 0-1
   Bool isPercent;   
public:
   // Public data
   static const int defaultPrecision;

   // Methods
   virtual void OK() const;

   ErrorData(Bool isPer = TRUE) : testSetError(UNDEFINED_REAL),
      totalTestWeight(0), isPercent(isPer) { }
   
   virtual void set_test_set(Real testError, Real weight);
   
   virtual Real error(Real trim = 0) const;
   virtual Real std_dev(Real trim = 0) const;
   virtual void confidence(Real& low, Real& high) const;

   virtual Real test_set_error() const;
   virtual Real theo_std_dev() const;
   virtual void theo_confidence(Real& low, Real& high) const;
   virtual Real bias(Real trim = 0) const {
      return error(trim) - test_set_error(); }

   virtual Bool has_test_set() const { return testSetError >= 0.0; }     
   virtual void check_test_set() const;
   virtual Bool has_estimated() const { return size() > 0; }
   virtual void check_estimated() const;
   virtual Bool has_std_dev() const { return size() > 1; }
   virtual void check_std_dev() const;
   virtual Bool has_theo_std_dev() const
      { return (isPercent && has_test_set() &&
		mlc.approx_greater(totalTestWeight, 1.0)); }
   virtual void check_theo_std_dev() const;

   virtual void clear() {
      StatData::clear();  testSetError = UNDEFINED_REAL;
      totalTestWeight = 0; }

   virtual void append(const StatData& other) { StatData::append(other); }
   virtual void append(const ErrorData& other);
   virtual int class_id() const {return CLASS_ERROR_DATA;}
   ErrorData& operator=(const ErrorData& rhs);
   virtual Bool operator==(const StatData& rhs) const;
   virtual void display(MLCOStream& stream = Mcout, Real trim = 0, 
			int precision = defaultPrecision) const;
   virtual void dot_display(MLCOStream& stream = Mcout, Real trim = 0,
			    int precision = defaultPrecision) const;
};

DECLARE_DISPLAY(ErrorData)


#endif
