// This is an include file.  For a full description of the class and
// functions, see the file name with a "c" suffix.

#ifndef _FileSchema_h
#define _FileSchema_h 1

#include <SchemaRC.h>
#include <Array2.h>
#include <OptionServer.h>

const MString MLC_SPECIAL_CHARS = "\\ \t\r,.:";
const int MAX_DISPLAYED_LOSS_ENTRIES = 100;

class FSLossEntry {
public:
   MString predName;
   MString actName;
   Real loss;
};

class FileSchema {
public:
   enum LossKeyword { nomatrix, nodefault, adefault, distance };
private:
   
   DynamicArray<AttrInfo *> attrInfos;

   LossKeyword lossKeyword;
   Array<Real> lossArgs;
   DynamicArray<FSLossEntry> lossEntries;
   int labelColumn;
   int weightColumn;
   Bool ignoreWeightColumn;
   
   NO_COPY_CTOR(FileSchema);
   NO_OPERATOR_EQ(FileSchema);

   void read_attributes(MLCIStream& namesFile, Bool lossOK);
   void apply_config(OptionServer& configOptions);
   void read_names(MLCIStream& namesIn);

   void check_for_duplicates() const;

   // support functions for read_loss_spec()
   Bool read_loss_override_spec(MLCIStream& in);
   void read_loss_default_spec(MLCIStream& in);
   void check_loss_entry_name(const MString& name,
			      int& anyMissing,
			      const SchemaRC& schema) const;
   
public:
   void OK() const;
   FileSchema(const MString& namesFile);
   FileSchema(MLCIStream& namesIn);
   FileSchema(const Array<AttrInfo *>& attrInfos);
   FileSchema(const FileSchema& other, CtorDummy);
   ~FileSchema();

   // create an MLC++ schema from this description
   SchemaRC create_schema() const;

   // get basic information about the file schema
   int num_attr() const { return attrInfos.size(); }
   const AttrInfo& operator()(int i) const
     { return *(attrInfos.index(i)); }

   // find an attribute by name
   int find_attribute(const MString& name, Bool fatalOnNotFound=TRUE) const;
   
   // get information about label and weight columns.  These are used by
   // InstanceReader
   int get_label_column() const { return labelColumn; }
   int get_weight_column() const { return weightColumn; }

   Bool get_ignore_weight_column() const { return ignoreWeightColumn; }
   void set_ignore_weight_column(Bool i) { ignoreWeightColumn = i; }
   
   // set label/weight columns.  These can be set either to attribute
   // numbers within this file schema, or by name
   void set_label_column(int which);
   void set_label_column(const MString& labelColName) {
      set_label_column(find_attribute(labelColName)); }
   void set_weight_column(int which);
   void set_weight_column(const MString& weightColName) {
      set_weight_column(find_attribute(weightColName)); }
   
   // set an attr info.  This will delete the old one automatically
   void set_attr_info(int i, const AttrInfo& newInfo);

   // set default loss condition
   void set_loss_default(LossKeyword keyword, const Array<Real>& args);
   
   // add a loss entry
   void add_loss_entry(const MString& predVal,
		       const MString& actVal, Real loss);

   // read a full loss specification from a file
   void read_loss_spec(MLCIStream& in);

   // apply the loss specification in this FileSchema to a Schema
   void apply_loss_spec(SchemaRC& schema) const;
   
   // output the file schema in our standard format
   void display(MLCOStream& stream) const;

   // display status information from the file schema
   void display_status(MLCOStream& stream) const;
};

DECLARE_DISPLAY(FileSchema)

   
#endif




