// MLC++ - Machine Learning Library in -*- C++ -*-
// See Descrip.txt for terms and conditions relating to use and distribution.

// This is an include file.  For a full description of the class and
//   functions, see the file name with a "c" suffix.

#ifndef _InstanceHash_h
#define _InstanceHash_h 1


#include <InstanceList.h>
#include <MLCStream.h>
#include <UnivHashTable.h>

// This class will be used only for the purpose of providing KeyType
//   and DataType for the InstanceHashTable class.
// It has a regular (default) copy constructor instead of a dummy
//   argument-copy constructor because it will be used as a template
//   argument of DblLinkList.

#define InstanceHashList InstanceList

class InstanceListSameAttr_ {
private:
   InstanceHashList* instHashList;

public:
   // Create an InstnaceHashList with the instance.   
   InstanceListSameAttr_(const InstanceRC& instance);
   // Normal copy constructor.
   InstanceListSameAttr_(const InstanceListSameAttr_& source);
   InstanceListSameAttr_& operator=(const InstanceListSameAttr_& source);
   virtual ~InstanceListSameAttr_();

   virtual void OK(int level=0) const;
   const InstanceHashList& get_instance_list() const { return *instHashList; }
   // Length is required for universal hashing KeyType.
   int length() const { return instHashList->get_schema().num_attr(); }

   // Return TRUE iff the instances in the InstanceHashLists match EXCLUDING
   //   the label (note that this is different from the usual semantics of
   //   operator=().
   Bool operator==(const InstanceListSameAttr_& rhs) const;
   Bool operator!=(const InstanceListSameAttr_& rhs) const {
      return !(operator==(rhs));
   }
					       
   // The same as operator==() because KeyType and DataType are the
   //   same.
   Bool contains_key(const InstanceListSameAttr_& rhs) const {
      return operator==(rhs);
   }

   InstanceListSameAttr_ get_key() const { return *this; }
   
   // Overwrite methods to meet the condition that there must be at
   //   least one element and all the instances are the same.
   virtual ILPix add_instance(const InstanceRC& instance);
   virtual InstanceRC remove_instance(ILPix& pix);

   // Operator<<() is needed by HashTable class.
   friend MLCOStream& operator<<(MLCOStream& stream, const
				 InstanceListSameAttr_& source);
};   



// InstanceHashTable_ is a universal hash table specifically designed
//   for InstanceListSameAttr_ data types. The purpose of defining this class
//   is to redefine the virtual hash() function in UniversalHashTable.

class InstanceHashTable_ :
   public HashTable<InstanceListSameAttr_,
                             InstanceListSameAttr_> {
   NO_COPY_CTOR(InstanceHashTable_);
   NO_OPERATOR_ASSIGN(InstanceHashTable_);
protected :
   RandCharArray randomData;   

public:
   InstanceHashTable_(int estimatedNum);
   InstanceHashTable_(int estimatedNum, unsigned int seed);   
   InstanceHashTable_(const InstanceHashTable_& source, CtorDummy);
   virtual int hash(const InstanceListSameAttr_& instance) const;
};


// InstanceHashTable is a clean interface to the user of hash
//   table of instances. Notice that it is not derived from any hash
//   table but has a private data member of type InstanceHashTable_.

class InstanceHashTable {
private:
   NO_COPY_CTOR(InstanceHashTable);
   NO_OPERATOR_ASSIGN(InstanceHashTable);
   InstanceHashTable_ hashTable;
   int numInstances; // It counts all the instances in each InstanceList in the
                     // hash table.
   
public:
   void OK() const { hashTable.ok_members(); }
   InstanceHashTable(int estimatedNum, const InstanceHashList* ihl = NULL);
   InstanceHashTable(int estimatedNum, unsigned int seed,
		     const InstanceHashList* ihl = NULL);    
   InstanceHashTable(const InstanceHashTable& source, CtorDummy);
   
   // Accessors
   int num_instances() const { return numInstances; }
   int num_items() const { return hashTable.num_items(); }

   // Return the number of matched instances in the hash table with
   //   the ihl instances.
   int num_matched_labelled_instances(const InstanceHashList& ihl) const;
   Bool find_labelled_instance(const InstanceRC& instance) const;
   // Add new instance into the hash table.
   //   Adding duplicate instances is allowed.
   void insert(const InstanceRC& instance);
   void del(const InstanceRC& instance);
   void del_all_unlabelled(const InstanceRC& instance);
   const InstanceHashList* find(const InstanceRC& instance) const;
   void display(MLCOStream& strm = Mcout) const { hashTable.display(strm); }
   Bool operator==(const InstanceHashTable& rhs) const
      {return ((numInstances == rhs.numInstances) &&
	       (hashTable == rhs.hashTable));}
};

#endif





