// MLC++ - Machine Learning Library in -*- C++ -*-
// See Descrip.txt for terms and conditions relating to use and distribution.

// This is an include file.  For a full description of the class and
// functions, see the file name with a "c" suffix.

#ifndef _InstanceList_h
#define _InstanceList_h 1

#include <InstanceRC.h>
#include <DblLinkList.h>
#include <Categorizer.h>
#include <MLCStream.h>
#include <StatData.h>
#include <BagCounters.h>
#include <InstanceOp.h>
#include <LogOptions.h>
#include <AttributeColumn.h>

class BoolArray;
class RealAndLabelColumn;
class InstanceList; // defined later in the file
class InstanceReader;

class AttrEqCategorizer; // @@ should be removed before NodeCat check in.

typedef DblLinkList<InstanceRC> InstancePtrList;
class ILPix;

// @@ change to array of pointers that are more efficient.
// @@ This should be a PixArray
typedef PtrArray<ILPix*> InstanceListIndex;
typedef PtrArray<InstanceList*> InstanceListArray;

class MRandom;         // defined in MRandom.h

// used in feature construction methods
typedef AttrValue_ (*FeatureFuncPtr) (const AttrValue_&, const AttrInfo&,
				      const AttrValue_&, const AttrInfo&,
				      const AttrInfo&);  

struct InstSampleElement
{   // used for sample_with_replacement--should not be
                   // global, but needed for template instantiation
   NO_DEFAULT_OPS(InstSampleElement);

   // Public member data
   ILPix *pix;     
   Bool flag;

   // Methods
   InstSampleElement()
      : pix(NULL), flag(FALSE) {}
};

extern const MString DEFAULT_DATA_EXT;
extern const MString DEFAULT_NAMES_EXT;
extern const MString DEFAULT_TEST_EXT;

class FileSchema;

class InstanceList {
   LOG_OPTIONS;
   NO_COPY_CTOR(InstanceList);

   friend class InstanceReader;
   friend class ILPix; // @@ try to make only constructor a friend

   // Member data
   InstancePtrList* instances;
   
   // schemas:  these are pointers because they may be NULL during
   // construction.  schema is the actual schema of the instances stored
   // in this list.
   // FileSchema refers to the names file used to construct this list.
   SchemaRC *schema;
   FileSchema *fileSchema;
   
   Bool weighted;     // FALSE implies all weights are 1.0
   Real totalWeight;  // cached sum of all weights

   static Real normalizeConfInterval;
   
   BagCounters* bagCounters;
   
   static Bool ignoreNames;
   static Bool removeZeroWeights;
   static int maxLabelVals;
   static int maxAttrVals;

   // Private methods
   void check_init_counters();
   InstancePtrList& instance_list();
   const InstancePtrList& instance_list() const;
   
   // feature construction functions defined in InstanceListFeature.c.
   void define_feature_info(int attrNum1, int attrNum2,
			    const MString& separator);
   void fill_in_feature_values(int attrNum1, int attrNum2,
			       const FeatureFuncPtr newFeatureFuncPtr);

   virtual void extreme_normalize_attr(int attrNum,const RealAttrInfo& rai,
				       RealAttrValue_& min,
				       RealAttrValue_& max) const;
   virtual void interquartile_normalize_attr(int attrNum,
					     const RealAttrInfo& rai,
					     RealAttrValue_& min,
					     RealAttrValue_& max) const;

   virtual void read_data_line(MLCIStream& dataFile, Bool isTest,
			       InstanceReader& reader);

   // internal functions used by the InstanceReader
   InstanceRC reader_add_instance(const Array<AttrValue_>&,
				  const AttrValue_ *labelVal,
				  Real weight,
				  Bool allowUnknownLabels);
   virtual void update_for_overflows(const BoolArray& projMap);
public:

   static void set_ignore_names(Bool ignore) { ignoreNames = ignore; }
   static Bool get_ignore_names()            { return ignoreNames;}
   static void set_remove_zero_weights(Bool b) { removeZeroWeights = b; }
   static Bool get_remove_zero_weights() { return removeZeroWeights; }
   static void set_max_label_vals(int maxVals) { maxLabelVals = maxVals; }
   static void set_max_attr_vals(int maxVals) { maxAttrVals = maxVals; }
   static int get_max_label_vals() { return maxLabelVals; }
   static int get_max_attr_vals() { return maxAttrVals; }
   enum NormalizationMethod { none, extreme, interquartile };
 
   virtual void OK(int level = 0) const;
   InstanceList(const SchemaRC& schema);
   
   // The file name suffices if the names file and data file default to
   // ".names" and ".data" respectively.  If the files have a different
   // prefix, give an empty file name, and give the full file name as suffix.
   // Gets ownership of instance info and instances
   // Note: do NOT use this constructor for test files.  Use one of the
   // next two constructors instead.
   InstanceList(const MString& file,
		const MString& namesExtension = DEFAULT_NAMES_EXT,
		const MString& dataExtension = DEFAULT_DATA_EXT);

   // Most general test file constructor.  You provide a names file,
   // a data file, and a schema.  The constructed list will be assimilated
   // to this existing schema so that it may be applied to a classifier.
   InstanceList(const SchemaRC& catSchema,
		const MString& file,
		const MString& namesExtension = DEFAULT_NAMES_EXT,
		const MString& testExtension = DEFAULT_TEST_EXT);
   
   // Create an instance list from a FileSchema representing a names file,
   // and a data file.
   InstanceList(const FileSchema& names,
		const MString& dataName);

   // Create an instance list (empty) from just a FileSchema
   InstanceList(const FileSchema& names);

   // Create an empty instance list with assimilation ability.  The list
   // will assimilate to the given schema
   InstanceList(const SchemaRC& catSchema,
		const FileSchema& names);
   
   // Create an instance list assimilated to the given catSchema,
   // using a FileSchema and a dataName
   InstanceList(const SchemaRC& catSchema,
		const FileSchema& names,
		const MString& testName);

   // Create an instance list assimilated to the given list.  This list
   // MUST have been built with a FileSchema, which will be used to
   // interpret the test data
   InstanceList(const InstanceList& trainList,
		const MString& testName);
   
   // copying and equality testing
   InstanceList(const InstanceList& source, CtorDummy,
		Bool preserveCounters=FALSE);
   InstanceList& operator=(const InstanceList& source);
   InstanceList *clone(Bool preserveCounters = FALSE) const {
      return new InstanceList(*this, ctorDummy, preserveCounters); }
   virtual Bool operator==(const InstanceList& other) const;
   Bool operator!=(const InstanceList& rhs) const
      {return ! ((*this)== rhs);}
   
   virtual ~InstanceList();

   // The given file must match the list's instance info
   // Deletes old data
   virtual void read_data(const MString& file, Bool isTest = FALSE);
   
   SchemaRC get_schema() const;
   virtual void set_schema(const SchemaRC& schemaRC);
   virtual ILPix add_instance(const InstanceRC& instance);
   virtual InstanceRC remove_instance(ILPix& pix);
   virtual void remove_all_instances();
   // Check if the instance list has no instances.  Note that there
   //   may be instances of weight zero
   Bool no_instances() const {DBG(ASSERT(instances != NULL));
			      return instances->empty();}

   int num_instances() const;
   int num_categories() const;
   
   // If we have a tie and one of them is the given tieBreaker, we prefer it.
   // The idea is to give the majority of some higher level.
   virtual Category majority_category(Category tieBreaker =
				      UNKNOWN_CATEGORY_VAL) const;
   virtual Category majority_category(const Array<Real> *tieBreakerDist)
      const;
   virtual Category majority_category(const Array<int>& tieBreakingOrder)
      const;

   virtual Array<int> *get_distribution_order() const;

   // search.  Returns NULL if not found.
   virtual ILPix find_labelled(const InstanceRC&,
			       Bool ignoreWeight = TRUE) const;
   virtual ILPix find_unlabelled(const InstanceRC&) const;
   virtual InstanceList *find_attr_val(int attrNum,
				       NominalVal val,
				       Bool exclude,
				       Bool excludeUnknowns) const;
   
   // splitting does not preserve counters or FileSchemas
   virtual InstanceListArray* split_by_label() const;
   
   // convenience functions from instance info
   const AttrInfo& attr_info(int attrNum) const {
      return get_schema().attr_info(attrNum);}
   const NominalAttrInfo& nominal_attr_info(int attrNum) const {
      return attr_info(attrNum).cast_to_nominal();}
   const AttrInfo& label_info() const {
      return get_schema().label_info();}
   const NominalAttrInfo& nominal_label_info() const {
      return label_info().cast_to_nominal();}
   int num_label_values() const {
      return nominal_label_info().num_values();}
   int num_attr_values(int attrNum) const {
      return nominal_attr_info(attrNum).num_values();}
   int num_attr() const {
      return get_schema().num_attr();}

   // order-based functions
   virtual InstanceListIndex* create_inst_list_index() const;
   virtual InstanceList* independent_sample(int size,
				   MRandom* mrandom = NULL,
				   InstanceListIndex* index = NULL) const;
   virtual InstanceList* independent_sample(int size,
				   InstanceList *restOfList,
				   MRandom* mrandom = NULL,
			           InstanceListIndex* index = NULL) const;
   virtual InstanceList*
      sample_with_replacement(int size, InstanceList *restOfList,
			      MRandom* mrandom = NULL) const;
   virtual InstanceList*
      weighted_sample_with_replacement(int size, InstanceList *restOfList,
				       MRandom* mrandom = NULL,
				       Bool rollInstances = TRUE) const;
   virtual InstanceList*
      weighted_perturbation(Real stdDev, Real lboundFactor,
			    MRandom* mrandom = NULL) const;
   InstanceList* shuffle(MRandom* mrandom = NULL,
			 InstanceListIndex* index = NULL,
			 Bool keepFileSchema = FALSE) const;
   virtual InstanceRC remove_front();
   virtual InstanceList* split_prefix(int numInSplit,
				      Bool keepFileSchema=FALSE);
   virtual void unite(InstanceList*& list);

   // unknown and conflicting instance handling
   virtual InstanceList* impute_unknown_nominals();
   virtual void remove_inst_with_unknown_attr();
   virtual void remove_inst_with_unknown_attr(int attrNumX);
   virtual void remove_conflicting_instances();

   // decompose the InstanceList into arrays of attribute label pair
   PtrArray<RealAndLabelColumn*>* transpose(const BoolArray& mask)
      const;
   PtrArray<AttributeColumn*>* transpose_all() const;
   AttributeColumn *transpose_labels() const;
   
   // corrupt_values changes each attribute value in each instance
   // to unknown with the given probability.
   virtual void corrupt_values_to_unknown(Real rate, MRandom& mrandom);
   
   virtual InstanceList* project(const BoolArray& attrMask) const;
   virtual void project_in_place(const BoolArray& attrMask);
   
   // weighting of instances
   virtual Bool is_weighted() const;
   virtual void unweight();
   virtual void make_weighted();
   virtual void set_weight(ILPix pix, Real wt);
   virtual Real get_weight(ILPix pix) const;
   virtual void normalize_weights(Real normFactor = 1.0,
				  Bool allowZeros = TRUE);
   virtual Real total_weight(Bool recalculate = FALSE) const;
   virtual InstanceList *unroll_integral_weights() const;
   virtual InstanceList *rollup_weights() const;
   virtual InstanceList *remove_zero_weights() const;
   virtual Array<Real> *get_weights() const;
   virtual void detect_outlier_weights(Real zThreshold = 2.25,
				       int numWarn = 10) const;
   
   // display
   virtual void display(MLCOStream& stream = Mcout,
			Bool protectChars = TRUE,
			Bool normalizeReal = FALSE) const;
   virtual void buntine_display(MLCOStream& stream = Mcout,
			        Bool test = FALSE) const;
   virtual void display_names(MLCOStream& stream = Mcout,
			      Bool protectChars = TRUE,
                              const MString& header = EMPTY_STRING) const;
   virtual void display_mlc_names(MLCOStream& stream,
				  Bool enumerateNominalValues = FALSE,
				  const MString& header = EMPTY_STRING) const;
   virtual void display_mlc_data(MLCOStream& stream,
				 const FileSchema *fileSchema=NULL) const;
   
   // normalization
   virtual void normalize_attr(int attrNum, NormalizationMethod method);
   virtual void normalize_attr(int attrNum, NormalizationMethod method,
			       RealAttrValue_& min, RealAttrValue_& max) const;
   virtual void normalize_list(NormalizationMethod method);

   // feature construction function defined in InstanceListFeature.c.
   virtual InstanceList *copy_with_blank_feature() const;
   
   // should be used ONLY for lists created with
   // copy_with_blank_feature(). Defined in InstanceListFeature.c.
   virtual void change_feature(int attrNum1, int attrNum2,
			       const MString& separator,
			       const FeatureFuncPtr newFeatureFuncPtr);

   // counter functions
   virtual const BagCounters& counters() const;
   virtual Bool has_counters() const { return (bagCounters != NULL); }
   virtual void ensure_counters() const;
   virtual void drop_counters() const;

   // file schema manipulation functions
   virtual const FileSchema& get_original_schema() const;
   virtual void set_names(const FileSchema& newNames);

   // finalize the loss specification--uses the FileSchema to secure
   // the loss matrix in the Schema now that all the data is read
   virtual void finalize_loss_matrix();

   // remove any attributes with only unknown values (nominals only)
   virtual void remove_unknown_attributes();
   
   // loss matrix functions
   virtual void read_loss_matrix(const MString& fileName);
   virtual Bool has_loss_matrix() const { return schema->has_loss_matrix(); }
   virtual void remove_loss_matrix();
   virtual void set_loss_matrix(const Array2<Real>& lossMatrix);

   // convenience function for zero-weight lists
   virtual Bool no_weight() const {
      return mlc.approx_equal((StoredReal)totalWeight, (StoredReal)0);
   }
   
   // @@ a hack to change ownership
   PTHR_DECL(void change_ownership() {
      if (schema)
	 schema->change_ownership();
      if (bagCounters)
	 bagCounters->change_ownership();
   })
};

class ILPix : public DLLPix<InstanceRC> {
public:
   ILPix(const InstanceList& il, int dir = 1) :
      DLLPix<InstanceRC>(il.instance_list(), dir) {}
   ILPix(const InstanceList& il, const DLLPix<InstanceRC>& pix) :
      DLLPix<InstanceRC>(il.instance_list(), 0) { 
          DLLPix<InstanceRC>::operator=(pix);}
};

DECLARE_DISPLAY(InstanceList)

#endif


