// MLC++ - Machine Learning Library in -*- C++ -*-
// See Descrip.txt for terms and conditions relating to use and distribution.

// This is an include file.  For a full description of the class and
// functions, see the file name with a "c" suffix.

#ifndef _InstanceReader_h
#define _InstanceReader_h 1

#include <InstanceList.h>
#include <FileSchema.h>
#include <QuarkTable.h>
#include <LogOptions.h>

struct QuarkAndString {
   NO_DEFAULT_OPS(QuarkAndString);

   // Public member data
   int quark;
   MString string;

   // Methods
   QuarkAndString()
      : quark(UNDEFINED_INT), string(EMPTY_STRING) {}
};

class InstanceReader {
   NO_DEFAULT_OPS(InstanceReader);
   LOG_OPTIONS;

   // Member data
   InstanceList* instList;  // This is the list we're building
   Array<AttrValue_> vals;  // Storage for attribute values for the
                            // instance we're building.  There is one
                            // for each entry in the list's FileSchema

   // the following are state flags on this reader
   BoolArray setAttr;  // TRUE for each attribute which has a value set
   Bool anySet;        // TRUE if any attribute has a value set
   
   int attrValueLimit;  // Limit on the number of values per attribute.
                        // crossing this limit results in
                        // automatic projection of that
                        // attribute.
   Real weight;         // Weight of the instance we're building
   Bool makeUnknowns;   // TRUE if unseen nominal values should be converted
                        // to unknowns when this instance is added
   Bool allowUnknownLabels;  // TRUE if unknown labels are allowed into the
                             // final list.  Useful for unsupervised
                             // learning.
   
   FileSchema *fileSchema;  // Specifies the form of the input data
                            // NOTE: aliased to InstanceList's FileSchema
   Array<int> assimMap;     // Maps each attribute in the above fileSchema
                            // to an attribute in the schema of the
                            // list we're building.
   BoolArray projMap;       // Keeps track of which attributes have run
                            // over the limit on number of values
   BoolArray listProjMap;   // Keeps track of which attributes have been
                            // projected out--from the list's
                            // point of view
   Array<QuarkTable *> quarkTables;   // stores quark mappings for
                                      // nominals (if needed)
   Bool warnOnSetComplete;  // If TRUE, the set_complete() function will
                            // generate warnings when it creates unknowns.

   // Private methods
   void prepare_to_set(int attrNum, Bool isNominal, Bool isReal);
   void construct_assim_map();
   void update_assim_map(const BoolArray& projMap);
   void build_proj_maps(BoolArray& readerProjMap,
			BoolArray& listProjMap);
   void update_for_overflows();
   int warn_projected_columns();
   void check_has_list();
   
public:
   // this enum contains special values for all mapping operations.
   // Since positive integers are legal for mapping operations as well,
   // this enum has no name.
   enum { unmapped = -1, mapToLabel = -2 };

   void OK();
   
   InstanceReader(InstanceList *& listToBuild,
		  int limit=0,
		  Bool makeUnknowns=FALSE,
		  Bool allowUnknownLabels=FALSE);
   ~InstanceReader();

   // Releasing the built list
   InstanceList *release_list();
   
   // Functions for setting values of a new instance
   void set_from_file(int attrNum, MLCIStream& stream);
   void set_nominal(int attrNum, const MString& attrVal);
   void set_nominal_quark(int attrNum, int quark, const char *quarkStr);
   void set_real(int attrNum, Real attrVal);
   void set_unknown(int attrNum);
   void set_complete();

   // Quark table management functions
   void preload_quark_table(int attrNum,
			    const Array<QuarkAndString>& quarkData);
   void establish_quark_table(int attrNum);

   // Functions for adding new instances
   InstanceRC add_instance();
   void assimilate_list(const InstanceList& instList);

   // Accessors
   Bool has_list() { return instList != NULL; }
   const InstanceList& get_list() { check_has_list(); return *instList; }
   SchemaRC get_schema() { check_has_list();
                           return instList->get_schema(); }
   const FileSchema& get_file_schema() { check_has_list();
     ASSERT(fileSchema);  return *fileSchema; }
   Bool is_labelled() { return get_schema().is_labelled(); }
   int get_label_column() { return get_file_schema().get_label_column(); }
   int get_weight_column() { return get_file_schema().get_weight_column(); }
   
   // write the contents of the list to a file, in the order defined by
   // the FileSchema (using the assimilation map)
   void display_data(MLCOStream& out);
};

#endif

