/*******************************************************************************
+
+  LEDA 3.5
+
+
+  d3_rat_plane.h
+
+
+  Copyright (c) 1995, 1996, 1997  by  LEDA Software GmbH
+  Postfach 151101, 66041 Saarbruecken, Germany
+  All rights reserved.
+ 
*******************************************************************************/


#ifndef LEDA_D3_RAT_PLANE_H
#define LEDA_D3_RAT_PLANE_H

#if !defined(LEDA_ROOT_INCL_ID)
#define LEDA_ROOT_INCL_ID 350028
#include <LEDA/REDEFINE_NAMES.h>
#endif


#include <LEDA/rat_vector.h>
#include <LEDA/d3_rat_point.h>
#include <LEDA/d3_plane.h>

class __exportC d3_rat_plane;

//------------------------------------------------------------------------------
// d3_rat_planes
//------------------------------------------------------------------------------

class __exportC d3_rat_plane_rep  : public handle_rep {

friend class __exportC d3_rat_plane;
friend __exportF int side_of(const d3_rat_plane&, const d3_rat_point&);

static leda_mutex mutex_id_counter;
static unsigned long id_counter;
   
   d3_rat_point a;
   integer     nx;
   integer     ny;
   integer     nz;

   unsigned long id;

public:
    
   d3_rat_plane_rep(const d3_rat_point&, const integer&, const integer&, 
                                                         const integer&);
  ~d3_rat_plane_rep() {}

friend inline unsigned long ID_Number(const d3_rat_plane&);
   
};

/*{\Manpage {d3_rat_plane} {} {Rational Planes}}*/

class __exportC d3_rat_plane  : public handle_base 
{
/*{\Mdefinition
An instance $P$ of the data type |\Mname| is an oriented rational plane in the 
three-dimensional space $\real^3$. It can be defined by a tripel ($a$,$b$,$c$)
of non-collinear rational points or a single rational point $a$ and a 
normal vector $v$. }*/


d3_rat_plane_rep* ptr() const { return (d3_rat_plane_rep*)PTR; }

public:

/*{\Mcreation p }*/

 d3_rat_plane();
/*{\Mcreate introduces a variable |\Mvar| of type |\Mname| initialized 
            to the trivial plane. }*/

 d3_rat_plane(const d3_rat_point& a, const d3_rat_point& b, const d3_rat_point& c);
/*{\Mcreate introduces a variable |\Mvar| of type |\Mname| initialized to 
            the plane through $(a,b,c)$.\\
            \precond $a$, $b$, and $c$ are
            not collinear. }*/

 d3_rat_plane(const d3_rat_point& a, const rat_vector& v);
/*{\Mcreate introduces a variable |\Mvar| of type |\Mname| initialized 
            to the plane that contains $a$ with normal vector $v$.\\
            \precond |v.dim()| $= 3$ and |v.length() > 0|. }*/

 d3_rat_plane(const d3_rat_point& a, const d3_rat_point& b);
/*{\Mcreate introduces a variable |\Mvar| of type |\Mname| initialized 
            to the plane that contains $a$ with normal vector $b-a$.}*/

 d3_rat_plane(const d3_rat_plane& p) : handle_base(p) {}
~d3_rat_plane() {}

d3_rat_plane& operator=(const d3_rat_plane& p) 
{ handle_base::operator=(p); return *this; }


/*{\Moperations 2 4}*/

d3_rat_point point1() const { return ptr()->a; }
/*{\Mop     returns the a point of |\Mvar|.}*/

rat_vector  normal() const 
{ return rat_vector(ptr()->nx,ptr()->ny,ptr()->nz,1); }
/*{\Mop     returns a normal vector of |\Mvar|.}*/

d3_plane to_d3_plane() const;
/*{\Mop  returns a floating point approximation of |\Mvar|. }*/


rational sqr_dist(const d3_rat_point& q) const;
/*{\Mop     returns the square of the Euclidean distance between |\Mvar|
            and $q$.}*/

rat_vector normal_project(const d3_rat_point& q) const;
/*{\Mop     returns the vector pointing from |q| to its projection
            on |\Mvar| along the normal direction.}*/


d3_rat_plane translate(const rational& dx, const rational& dy, 
                                           const rational& dz) const;
/*{\Mopl    returns |\Mvar| translated by vector $(dx,dy,dz)$.}*/


d3_rat_plane translate(integer dx, integer dy, integer dz, integer dw) const;
/*{\Mopl    returns |\Mvar| translated by vector $(dx/dw,dy/dw,dz/dw)$.}*/


d3_rat_plane translate(const rat_vector& v) const;
/*{\Mop     returns \Mvar$+v$, i.e., |\Mvar| translated by vector $v$.\\
	    \precond $v$.dim() = 3.}*/

d3_rat_plane operator+(const rat_vector& v) const { return translate(v); }
/*{\Mbinop  returns |\Mvar| translated by vector $v$.}*/


d3_rat_plane reflect(const d3_rat_plane& Q) const;
/*{\Mop     returns |\Mvar| reflected  across plane $Q$.}*/

d3_rat_plane reflect(const d3_rat_point& q) const;
/*{\Mop     returns |\Mvar| reflected across point $q$. }*/

d3_rat_point reflect_point(const d3_rat_point& q) const;
/*{\Mop     returns |q| reflected across plane |\Mvar|.}*/

int side_of(const d3_rat_point& q) const;
/*{\Mop computes the side of |\Mvar| on which |q| lies. }*/

bool contains(const d3_rat_point& q) const { return side_of(q) == 0;}
/*{\Mop     returns true if point |q| lies on plane |\Mvar|, i.e.,
            (|\Mvar.side_of(q) == 0|), and false otherwise .}*/

bool parallel(const d3_rat_plane& Q) const;
/*{\Mop     returns true if planes |\Mvar| and |Q| are parallel, and 
            false otherwise. }*/

int operator==(const d3_rat_plane& Q) const;
int operator!=(const d3_rat_plane& Q) const { return !operator==(Q);}


friend __exportF ostream& operator<<(ostream& O, const d3_rat_plane& p) ;
/*{\Mbinopfunc  writes |\Mvar| to output stream $O$.}*/

friend __exportF istream& operator>>(istream& I, d3_rat_plane& p);
/*{\Mbinopfunc  reads |\Mvar| from input stream $I$.}*/


friend unsigned long ID_Number(const d3_rat_plane&);

};


inline unsigned long ID_Number(const d3_rat_plane& p) { return p.ptr()->id; }
inline char* leda_tname(const d3_rat_plane*) { return "d3_rat_plane"; }


#if LEDA_ROOT_INCL_ID == 350028
#undef LEDA_ROOT_INCL_ID
#include <LEDA/UNDEFINE_NAMES.h>
#endif

#endif

