// MLC++ - Machine Learning Library in -*- C++ -*-
// See Descrip.txt for terms and conditions relating to use and distribution.

// This is an include file.  For a full description of the class and
// functions, see the file name with a "c" suffix.

#ifndef _MLCFactory_h
#define _MLCFactory_h 1

#include <DynamicArray.h>
#include <StringMap.h>

// Base class for use by the factory.  Supports only class_id().
// All classes used by the factory must be derived from this class.
class FactoryObject {
public:
   virtual int class_id() = 0;
};

// This class represents a single row in the factory's table
// Each row contains:
//   classID:  numerical ID for this class--must be unique across library.
//   baseName: name of base class for this class's hierarchy.
//   className: name of the registered class.
// NOTE:  This class is a "final" class; do not derive from it.
class FactoryElement {
public:
   // Public member data
   int classID;
   MString baseName;
   MString className;
   void *buildFunction;
   int associatedID;

   // Methods
   FactoryElement() : classID(UNDEFINED_INT), buildFunction(NULL),
                      associatedID(UNDEFINED_INT) { }
};

class MLCFactory {
   NO_DEFAULT_OPS(MLCFactory);

   // Member data (also see public data)
   // This array is a pointer so that external_ctor will work 
   DynamicArray<FactoryElement> *builders;

   // StringMap for mapping classnames to their ID numbers
   StringMap<int> *nameMapper;
   
   // Hint to help find the next available ID faster.  All
   // IDs lower than this index are guaranteed to be full.
   int nextIDHint;

public:
   // Public data
   static int initialized;

   // Methods
   MLCFactory() { };
   ~MLCFactory();
   
   // need to initialize via this call instead of constructor so we
   // can control initialization order.
   void external_ctor();

   int id_from_name(const MString& className); 
   void *get_builder(const MString& baseName, int classID);
   void *get_builder(const MString& baseName, const MString& className)
      { return get_builder(baseName, id_from_name(className)); }
   int get_associated_id(int classID);
   int register_class(const MString& baseName,
		      const MString& className,
		      void *builder,
		      int classID=UNDEFINED_INT,
		      int associatedID=UNDEFINED_INT);
};

// there is only one factory
extern MLCFactory mlcFactory;


// This is a generic object builder function.  Use this for heirarchies
// with default constructors
typedef FactoryObject *
      (*FactoryObjectBuilder)();

// Use this macro at the top of the public section of the
// factory object class declaration in the header file
// Use DECLARE_FACTORY_OBJECT_ID if you have a class ID for this object
// Use DECLARE_FACTORY_OBJECT otherwise.
#define DECLARE_FACTORY_OBJECT_ID(classID, Object) \
  static int staticClassID_;  \
  static FactoryObject *create_this_type();  \
  virtual int class_id() { return classID; } 

#define DECLARE_FACTORY_OBJECT(Object) \
  static int staticClassID_; \
  static FactoryObject *create_this_type(); \
  virtual int class_id() { return staticClassID_; }

// Use this macro at the top of the .c file for the object class.
// DEF_FACTORY_OBJECT_ID if you have a class ID,
// DEF_FACTORY_OBJECT if not.
#define DEF_FACTORY_OBJECT_ID(classID, Object) \
  int Object::staticClassID_ = \
    mlcFactory.register_class("FactoryObject", # Object, \
			      &Object::create_this_type, classID); \
  FactoryObject *Object::create_this_type() \
  { return new Object(); }

#define DEF_FACTORY_OBJECT(Object) \
int Object::staticClassID_ = \
  mlcFactory.register_class("FactoryObject", # Object, \
			    (void *) &Object::create_this_type); \
  FactoryObject *Object::create_this_type() \
  { return new Object(); }


// This declares a convenience function for constructing a
// FactoryObject using the factory.  All casts and checks are performed.
FactoryObject *create_factory_object(int classID);
FactoryObject *create_factory_object(const MString& className);

#endif
