// MLC++ - Machine Learning Library in -*- C++ -*-
// See Descrip.txt for terms and conditions relating to use and distribution.

// This is an include file.  For a full description of the class and
// functions, see the file name with a "c" suffix.

#ifndef _MLCStream_h
#define _MLCStream_h 1
#include <iomanip.h>
#include <fstream.h>
#include <MString.h>
#include <error.h>
#include <fstream.h>

//@@ Fixed limit on the size of the option stack forced by a
//   bug in GNU.
#define OPTION_STACK_DEPTH_LIMIT  25

// Due to archaic 8+3 PC naming conventions, strstream.h
// is sometimes called strstrea.h
#ifdef PC_MSVC
#include <strstrea.h>
#else
#include <strstream.h>
#endif

#include <checkstream.h>

// For saving, restoring options
struct MStreamOptions {
   // default copy constructor required by MLCOStream::get_options

   // Public member data
   // Wrapping options
   int wrapWidth;
   MString wrapPrefix;
   MString newLinePrefix;

   // Numerical output options
   int fieldWidth;
   int precision;
   MString::RealNotation realNotation;

   //@@ Add display preference options here later

   // Methods
   MStreamOptions()
      : wrapWidth(UNDEFINED_INT), wrapPrefix(), newLinePrefix(),
        fieldWidth(0), precision(0),
        realNotation(MString::mlcGeneral) { }

   MStreamOptions& operator=(const MStreamOptions& rhs);
   Bool operator==(const MStreamOptions& rhs) const;
   Bool operator!=(const MStreamOptions& rhs) const
      {return ! ((*this)== rhs);}
};


enum OutputType {FileStream, MemStream, XStream, PrinterStream};

// Note:  this class needs to be buildable by MLCInit::startup().  This
//   means that all data members must be able to be initialized through
//   and external_ctor() function as well as standard constructors.
//   as a result, we may have no reference variables or straight members
//   which are classes.  These have all been converted to pointers and
//   access functions have been provided to duplicate the older references.
//   A do-nothing constructor which takes the enum type UseExternalCtor
//   is provided to prevent external_ctor from conflicting with normal
//   construction.

class MLCIStream {
   NO_DEFAULT_OPS(MLCIStream);

   // safe static construction information
   friend class MLCInit;

   // Member data (also see protected data)
   Bool usedExternalCtor;

   // Private methods
   void external_ctor(const MString& fileName);
   void external_ctor(const MString& dscr, ifstream& strm);
   void external_ctor(istream& strm);
   void external_ctor(int fd, const MString& dscr);
   void external_dtor();
   
protected:
   enum {INPUT_BUFFER_SIZE = 50000}; // this is like a #define for the class
   // Protected member data
   char buf[INPUT_BUFFER_SIZE];
   
   MString *descrPtr;
   
   // Note order. strStream & fileStream need to be
   //   declared before stream in order for c-tor's to work.
   istrstream* strStream;
   ifstream *fileStream;
   
   istream *streamPtr;         // Points to either an strStream or an ifstream
   Bool closed;             // Flag for safety
   long tellpLast; 
   int lineCount;
   int posInLine;

   // Protected methods
   inline istream& stream() const;
   
   void open_read();
   void update_line_count();
   void init();
public:
   MLCIStream(UseExternalCtor) { }
   void attach(int fd);
   inline istream& get_stream();
   void OK(int level = 0) const;
   MLCIStream(const MString& fileName);
   MLCIStream(const MString& dscr, ifstream& strm);
   MLCIStream(istream& strm);
   MLCIStream(int fd, const MString& dscr);
   ~MLCIStream();
   void close();
   void check_stream_is_valid() const;
   inline const MString& description() const;
   int line_count();
   int pos_in_line();
   ifstream& get_fstream();
   // The function below is *ONLY* used in OK() because of the oc ref. bug.
   istrstream& get_strstream() { return *strStream;}  
   void setbuf(char *p, int l);
   Bool eof(){ return get_stream().eof(); }
   inline char get();
   MLCIStream& get(char&);
   void ignore(int len = 1, int delim = EOF);
   inline int peek();
   void skip_white();
   void read_bin(int&);
   void read_bin(unsigned char&);
   void read_bin(char&);
   void read_bin(long&);
   void read_bin(float&);
   void read_bin(short&);
   void read_bin(double&);
   MLCIStream& operator>>(int&);
   MLCIStream& operator>>(char*);
   MLCIStream& operator>>(char&);
   MLCIStream& operator>>(long&);
   MLCIStream& operator>>(Real&);
   MLCIStream& operator>>(short&);
   MLCIStream& operator>>(float&); 
   MLCIStream& operator>>(unsigned int&);
   MLCIStream& operator>>(unsigned char&);
   MLCIStream& operator>>(unsigned long&);
   MLCIStream& operator>>(unsigned short&);
   MLCIStream& operator>>(istream& (*)(istream&));
#if defined(CFRONT) || defined(MIPS_PRO) || defined(PC_MSVC)
   friend MLCIStream& operator>>(MLCIStream& s, const SMANIP(int)& m);
#elif defined(GNU)
   friend MLCIStream& operator>>(MLCIStream& s, const smanip<int>& m);
#else
#error "compiler not supported"
#endif

};

class MLCOStream {
   NO_DEFAULT_OPS(MLCOStream);

   // safe static construction information
   friend class MPStream;
   friend class MLCInit;

   // Member data (also see protected data)
   Bool usedExternalCtor;

#ifdef PTHREADS
   // need access to external destructors
   friend class MLCThreads;
   // The id of the owning thread; all other threads output to a buffer instead
   pthread_t streamOwner;
   Bool is_stream_owner() const {
      return pthread_equal(streamOwner, pthread_self()); }
#else
   Bool is_stream_owner() const { return TRUE; }
#endif /* PTHREADS */

   // Private methods
   void external_ctor(const OutputType otype);
   // For memory streams (strstream based MLCOStream).
   void external_ctor(const MString& description, char *buf, int nBytes); 
   void external_ctor(const MString& fileName,
		      const OutputType otype = FileStream,
		      Bool append = FALSE, int prot = defaultOpenProt);
   void external_ctor(const MString& dscr, ofstream& strm,
		      const OutputType otype = FileStream);
   void external_ctor(ostream& strm, const OutputType otype);
   void external_ctor(int fd, const MString& dscr,
		      const OutputType otype = FileStream);
   void external_dtor();
   
protected:
   enum {MAX_BUFFER_SIZE = 50000}; // this is like a #define for the class

   // Protected member data
   OutputType outputType;

   MString *descrPtr;

   //@@ Because of a bug in GNU, we are unable to use Array or Dynamic
   //   Array within this class.  Therefore, the option stack is
   //   implemented using a C-style array with a fixed size.  Manual
   //   bounds checks occur in all the accessor functions.
   int optionStackDepth;
   MStreamOptions *optionStack;
   
   // Note order. strStream & fileStream need to be
   //   declared before stream in order for c-tor's to work.
   ostrstream* strStream;
   ofstream* fileStream;
   
   ostream *streamPtr;		// Can point to a string stream or an ofstream

   char *formatStreamBuf;	// string buffer is allocated from free store
   long lineCount, linePos; 
   Bool atLeftmostPosition;	// For checking to see if we are at the
				// left most position on output.
   
   ostrstream *formatStreamPtr; // strstream used for incore formatting
   ostrstream& format_stream() const;
   
   char *memBuf;		// Returned from an strsrtream-based MLCOStream

   Bool closed;

   // Protected methods
   ostream& stream() const;
   void init();
   void print_trailing_space();
   void output_format_stream(Bool flushOutput);
   Bool output_format_stream_no_newline(Bool outputFlush, unsigned long lenny);
   void gen_newline(); 
   void get_inner_precision_and_width(int& precision, int& width) const;
   MStreamOptions& get_options_(); // Non-const version
   
public:
   MLCOStream(UseExternalCtor);
   void attach(int fd);
   inline ostream& get_stream();
   inline const MString& description() const;

   // Wrapping methods.
   int wrap_prefix_length() const { return get_options().wrapPrefix.length(); }
   int newline_prefix_length() const
     { return get_options().newLinePrefix.length(); }
   void reset_pos_in_line() { output_format_stream(TRUE); linePos = 1;}

   // General Option handling
   void set_options(const MStreamOptions& options);
   void push_options();
   void push_options(const MStreamOptions& options);
   const MStreamOptions& get_options() const;
   void pop_options();
   int stack_depth();

   // Wrapping options
   void set_wrap_width(int newWidth);
   int get_wrap_width() const { return get_options().wrapWidth; }
   void set_wrap_prefix(const MString& str);
   const MString& get_wrap_prefix() const { return get_options().wrapPrefix; }
   void set_newline_prefix(const MString& str);
   const MString& get_newline_prefix() const
     { return get_options().newLinePrefix; }

   // Numerical options
   void set_field_width(int newWidth);
   int get_field_width() const { return get_options().fieldWidth; }
   void set_precision(int prec);
   int get_precision() const { return get_options().precision; }
   void set_notation(MString::RealNotation notation) {
      get_options_().realNotation = notation; }
   MString::RealNotation get_notation() { return get_options().realNotation; }
   
   // Everything else.
   void OK(int level = 0) const;
   MLCOStream(const OutputType otype);
   // For memory streams (strstream based MLCOStream).
   MLCOStream(const MString& description, char *buf, int nBytes); 
   MLCOStream(const MString& fileName, const OutputType otype = FileStream,
	      Bool append = FALSE, int prot = defaultOpenProt);
   MLCOStream(const MString& dscr, ofstream& strm,
	      const OutputType otype = FileStream);
   MLCOStream(ostream& strm, const OutputType otype);
   MLCOStream(int fd, const MString& dscr,
	      const OutputType otype = FileStream);
   ~MLCOStream();
   void close();
   void check_stream_is_valid() const;
   ofstream& get_fstream();
   ostrstream& get_strstream();
   int line_count() const;
   int pos_in_line();
   void setbuf(char *p, int l);
   MString mem_buf();
   OutputType output_type() const;
   void include_file(const MString& filename);
   void write_bin(int&);
   void write_bin(char&);
   void write_bin(unsigned char&);
   void write_bin(long&);
   void write_bin(float&);
   void write_bin(short&);
   void write_bin(double&);
   MLCOStream& operator<<(int);
   MLCOStream& operator<<(char);
   MLCOStream& operator<<(long);
   MLCOStream& operator<<(short);
   MLCOStream& operator<<(float);
   MLCOStream& operator<<(double);
   MLCOStream& operator<<(const char*);
   MLCOStream& operator<<(unsigned int);
   MLCOStream& operator<<(unsigned char);
   MLCOStream& operator<<(unsigned long);
   MLCOStream& operator<<(unsigned short);
   MLCOStream& operator<<(void*);
   MLCOStream& operator<<(ostream& (*)(ostream&));
   MLCOStream& operator<<(MLCOStream& (*)(MLCOStream&));
#if defined(CFRONT) || defined(MIPS_PRO) || defined(PC_MSVC)
   friend MLCOStream& operator<<(MLCOStream& s, const SMANIP(int)& m);
#elif defined(GNU)
   friend MLCOStream& operator<<(MLCOStream& s, const smanip<int>& m);
#else
#error "compiler not supported"
#endif
};

MLCOStream& reset_pos_in_line(MLCOStream&);

MLCOStream& push_options(MLCOStream&);
MLCOStream& pop_options(MLCOStream&);


/***************************************************************************
  Description : Returns the istream associated with the stream.
  Comments    : Currently all constructors set it so it would never be the
                  case that this returns an invalid stream.
***************************************************************************/
istream& MLCIStream::get_stream()
{
   return stream();
} 


/***************************************************************************
  Description : Extracts a character from the stream and returns it.
  Comments    : Aborts if extraction encounters end of file.
***************************************************************************/
char MLCIStream::get()
{
   int ch = stream().get();

   if (ch == EOF) 
      err << "MLCIStream::get: Tried to get EOF in stream " << description()
	  << fatal_error;
   OK();
   return ch;
}


/***************************************************************************
  Description : If stream is at end of file, returns EOF.  Otherwise,
                  returns next character without extracting it.
  Comments    :
***************************************************************************/
int MLCIStream::peek()
{
   // Notice that there's no OK() check here because we want the
   //   caller to check EOF.
   // OK() here is also a time killer (10% of time in mushroom/ID3).
   return get_stream().peek();
}   

inline const MString& MLCIStream::description() const
{
   ASSERT(descrPtr);  return *descrPtr;
}

/***************************************************************************
  Description : Checks that the stream is valid, returns the ostream associated
                  with the stream.
  Comments    : Need to move here because of is_master()
***************************************************************************/
ostream& MLCOStream::get_stream()
{
  if (outputType == XStream)
      err << "MLCOStream::get_stream(): operation undefined for XStream "
	  << description() << fatal_error;
  if (get_wrap_width())
     return ((MLCOStream*)this)->format_stream(); // cast away constness
  else
     return stream();
}

inline istream& MLCIStream::stream() const
{
   ASSERT(streamPtr); return *streamPtr;
}

inline const MString& MLCOStream::description() const
{
   ASSERT(descrPtr);  return *descrPtr;
}

/***************************************************************************
  Description : New inlined manipulators for the stream:
                  push_opts:      pushes duplicate options onto the stack
		  pop_opts:       pops options from the stack
                  general_reals:  sets mlcGeneral notation for reals
		  fixed_reals:    sets mlcFixed notation for reals
		  std_e_reals:    sets standard "e" notation for reals
		  std_f_reals:    sets standard "f" notation for reals
		  std_g_reals:    sets standard "g" notation for reals
  Comments    : 
***************************************************************************/
inline MLCOStream& push_opts(MLCOStream& stream) {
   stream.push_options();  return stream; }
inline MLCOStream& pop_opts(MLCOStream& stream) {
   stream.pop_options();  return stream; }
inline MLCOStream& general_reals(MLCOStream& stream) {
   stream.set_notation(MString::mlcGeneral);  return stream; }
inline MLCOStream& fixed_reals(MLCOStream& stream) {
   stream.set_notation(MString::mlcFixed);  return stream; }
inline MLCOStream& std_e_reals(MLCOStream& stream) {
   stream.set_notation(MString::exponential);  return stream; }
inline MLCOStream& std_f_reals(MLCOStream& stream) {
   stream.set_notation(MString::fixed);  return stream; }
inline MLCOStream& std_g_reals(MLCOStream& stream) {
   stream.set_notation(MString::general);  return stream; }


// allow display of MStreamOptions
MLCOStream& operator<<(MLCOStream& stream, const MStreamOptions& opt);


#endif

