// MLC++ - Machine Learning Library in -*- C++ -*-
// See Descrip.txt for terms and conditions relating to use and distribution.

// This is an include file.  For a full description of the class and
// functions, see the file name with a "c" suffix.

#ifndef _MLCThreads_h
#define _MLCThreads_h 1

#ifdef PTHREADS
#include <stdio.h> // For syserrlist
#include <errno.h>
#include <error.h>		// for ERRMESSAGEBUF_SIZE
#include <TmpFileName.h>	// for shmFile

// @@ we will get into recursion problem if we are going to use
// THREAD_MEMORY in the Array.h
#include <Array.h>		// need for the PtrArray class

class MPthread;
typedef PtrArray<MPthread*> MPthreads;

class ThreadStreamInfo;		// dynamic array of open streams

// These can be also defined in
// (a) corresponding files before the basics.h inclusion
// (b) compiler command line via -D...
//
// #define TEST_UNIPROC
// #define INCREASE_STACK_SIZE
//

// have exactly the same number of threads as processors
const int THREAD_FACTOR = 1;

// The workaround for the 6.2 and 6.4 OS bug #525481
#define NEW_STACK_SIZE 0x200000
// insert this at the start of main()
#define GROW_STACK auto int local = 0; \
                   *(((char*) &local) - NEW_STACK_SIZE) = local

class MLCThreads {
   friend class GlobalOptions;	// need to get and set numAvailProcs
   friend class MPthread;	// need direct access to the scheduler
   friend class MLCInit;	// for id initialization

   NO_DEFAULT_OPS(MLCThreads);

   // statics below are initialized in a separate static function init()
   static Bool initialized;

   // scheduler control; per class variables
   static pid_t masterPid;
   static pthread_t masterId;
   static volatile int numAvailProcs;
   static volatile int numWorkingThreads;

   // fatal error in one of the threads; bail out!
   static volatile Bool fatalErrorFlag;

   // needed for scheduling policy and increased stack
   static pthread_attr_t attr;

   // LEDA lock
   static pthread_mutex_t ledaSem;

   // PROTECT lock
   static pthread_mutex_t protectSem;

   // scheduler control: semaphor and conditional variables
   static pthread_mutex_t schedulerSem;
   static pthread_cond_t schedulerCondVar;

   // per thread data to be read by pthread_getspecific()
   static pthread_key_t arenaKey;
   static pthread_key_t mpthreadKey;

   // we would like to keep thread specific buffers
   static pthread_key_t mcoutKey;
   static pthread_key_t mcerrKey;
   static pthread_key_t errKey;

   // these functions need to be executed just before a thread joins
   static void delete_thread_stream(void* stream);

public:

   MLCThreads();
   ~MLCThreads();

   // the actual initialization of the static locks and keys
   static void init();

   // is it a pthread ?
   static Bool is_master() {
      return (pthread_equal(masterId, pthread_self())); }

   // scheduler group
   static int get_num_working() { return numWorkingThreads; }

   static Bool slot_ready() {
      return (numWorkingThreads < THREAD_FACTOR * numAvailProcs);
   }

   static void incr_working() {
      DBG(ASSERT(initialized));
      ASSERT(pthread_mutex_lock(&schedulerSem) == 0);
      while(numWorkingThreads >= THREAD_FACTOR * numAvailProcs)
	 pthread_cond_wait(&schedulerCondVar, &schedulerSem);
      numWorkingThreads++;
      ASSERT(pthread_mutex_unlock(&schedulerSem) == 0);
   }

   static void decr_working() {
      DBG(ASSERT(initialized));
      ASSERT(pthread_mutex_lock(&schedulerSem) == 0);
      numWorkingThreads--;
      pthread_cond_signal(&schedulerCondVar);
      ASSERT(pthread_mutex_unlock(&schedulerSem) == 0);
   }

   static void thread_fatal_error();
   static void check_fatal_error_flag(Bool abortOnFatal = TRUE);
   
   // returns true if all threads have finished working
   static Bool all_finished(const MPthreads& threads);
   
   // dribble group
   static void dribble(const MPthreads& threads);
   static void dribble_and_wait(const MPthreads& threads, int time);
   static void dribble_till_end(const MPthreads& threads, int time);   

   // locks
   static void mpthread_leda_lock() {
      DBG(ASSERT(initialized));
      if (pthread_mutex_lock(&ledaSem))
	 err << "MLCThread: leda lock error" << fatal_error;
   }

   static void mpthread_leda_unlock() {
      DBG(ASSERT(initialized));
      if (pthread_mutex_unlock(&ledaSem))
	 err << "MLCThread: leda unlock error" << fatal_error;
   }   

   static void mpthread_protect_lock() {
      DBG(ASSERT(initialized));
      if (pthread_mutex_lock(&protectSem))
	 err << "MLCThread: global protect lock error" << fatal_error;
   }

   static void mpthread_protect_unlock() {
      DBG(ASSERT(initialized));
      if (pthread_mutex_unlock(&protectSem))
	 err << "MLCThread: global protect unlock error" << fatal_error;
   }

   // thread specific data access
   static const void* get_arena();

   static MPthread& get_mpthread() {
      return *((MPthread*)pthread_getspecific(mpthreadKey));
   }

   static MLCOStream* get_mcout_stream() {
      return (MLCOStream*) pthread_getspecific(mcoutKey);
   }

   static MLCOStream* get_mcerr_stream() {
      return (MLCOStream*) pthread_getspecific(mcerrKey);
   }

   static MLCOStream* get_err_stream() {
      return (MLCOStream*) pthread_getspecific(errKey);
   }
   
   // @@ this is a shortcut for 2.1, will be expanded to a dynamic array
   static MLCOStream* get_own_stream(const MLCOStream* stream) {
      DBG(ASSERT(!is_master()));

      if (stream == &Mcout)
         return get_mcout_stream();
      else if (stream == &Mcerr)
         return get_mcerr_stream();
      else if (stream == &err)
         return get_err_stream();

      // @@ need to look up a dynamic array
      err << "stream " << stream->description() << " is not implemented"
	  << fatal_error;

      ABORT_IF_REACHED;		// should never get here
      return NULL;		// keeps compiler happy
   }

};

// in the MPArena and MPStream only MPthread has access to the constructors
// and destructors; thus they are private

class MPArena {
   NO_DEFAULT_OPS(MPArena);
   friend class MPthread;

   const void* arenaPtr;
   const TmpFileName shmFile;
   const void* shmAddr;
   usptr_t *usPtr;
   int shmId;

   MPArena();
   ~MPArena();
};


class MPStream {
   NO_DEFAULT_OPS(MPStream);
   friend class MPthread;

   MLCOStream stream;
   const MLCOStream* actualStream;
   char* buffer;
   int buffSize;

   MPStream(MLCOStream& stream, char* parentBuffer, int size);
   ~MPStream();

   void init();

   MString get_buf() { return stream.mem_buf(); }
   operator const char *() const { return buffer; }
   char operator[](int k) { return buffer[k]; }
      
};


class MPthread {		// ABC

   NO_DEFAULT_OPS(MPthread);

protected:
   typedef int ThreadNo;

private:   
   ThreadNo number;		// thread id to be printed

   const Real dribbleScale;	// how much it contributes to total dribble
   volatile Real dribble;	// dribble from [0; 1]

   Bool joined;		 	// TRUE if joined with the parent
   Bool working;		// TRUE if doing work, FALSE if dribbling
   const pthread_t parentId;    // the id of the parent
   pthread_t threadId;		// the id of the created thread

   sigjmp_buf fatalErrorJump;	// context for long jump on fatal error

   MPArena arena;

   // @@ the dynamic array with thread specific streams will appear here
   ThreadStreamInfo* threadStreamInfo;
   
   // shortcut for 2.1
   MPStream errStream;
   MPStream mcerrStream;
   MPStream mcoutStream;

   char errMessageBuf[ERRMESSAGEBUF_SIZE];

   typedef void* (*pthreadFunPnt)(void *);

   // pthread_main contains all thread initialization; called by a child
   static void pthread_main(void *ptr);

protected:

   Bool volatile fatalErrorFlag;

   // should contain unset_working and dribble; called by a child
   virtual void do_work() = 0;

   // should move the results from the child arena; called by the parent
   virtual void copy_data() = 0;

public:
   MPthread(ThreadNo number, double dribbleScale);
   virtual ~MPthread();

   Bool is_working() const { return working; }
   Bool parent_is_child() const { return pthread_equal(parentId, threadId); }
   Bool is_parent() const { return pthread_equal(parentId, pthread_self()); }
   Bool is_child() const { return pthread_equal(threadId, pthread_self()); }

   Real get_current_dribble() const { return dribbleScale * dribble; }
   void set_dribble(Real newDribble) { dribble = newDribble; }

   // spawn a new thread with new arena malloc; called by the parent
   void execute() {

      // wait for the slot and increment MLCThreads::numWorkingThreads
      MLCThreads::incr_working();

#ifdef TEST_UNIPROC
      // initialize the streams
      errStream.init();
      mcerrStream.init();
      mcoutStream.init();

      // call do_work() directly
      do_work();
#else
      // spawn the thread and pass the current class pointer to pthread_main
      if (pthread_create((pthread_t*) &threadId, &MLCThreads::attr,
			 (pthreadFunPnt) pthread_main, (void*) this))
	 err << "MPthread: pthread_create" << fatal_error;
#endif /* TEST_UNIPROC */
   }

   void unset_working() {
      ASSERT(working);
      MLCThreads::decr_working();
      working = FALSE;
   }

   // finalize the data and copy thread data to the parent if necessary
   void finalize() {
      // do not wait for a thread twice

      ASSERT(is_parent());

      if (!joined) {

	 ASSERT(!working);
	 
#ifdef TEST_UNIPROC
	 // need to delete arena
	 adelete((void*)arena.arenaPnt);
#else
	 if (pthread_join(threadId, NULL)) {
	    Mcerr << "Warning: MPthread::finalize() pthread_join: "
		  << sys_errlist[errno] << endl;
	 }

	 if (mcerrStream[0]) {
	    Mcerr << "--- thread " << number << " Mcerr output:" << endl;
	    Mcerr << (const char*) mcerrStream;
	    Mcerr << "--- end of Mcerr ---" << endl;
	 }

	 if (fatalErrorFlag) {
	    // @@ do we want to abort here ?
	    Mcerr << "fatal error in thread " << number << ": "
		  << errMessageBuf << endl;
	    // MLCThreads::fatalErrorFlag is set in pthread_main();
	 } else
	    copy_data();

	 if (mcoutStream[0]) {
	    Mcout << "--- thread " << number << " Mcout output:" << endl;
	    Mcout << (const char*) mcoutStream;
	    Mcout << "--- end of Mcout ---" << endl;
	 }

	 // @@ should go through the dynamic array here
	 
#endif /* TEST_UNIPROC */

	 joined = TRUE;
	 
      }

   }

   virtual void recover();

};


inline Bool is_master() { return MLCThreads::is_master(); }


#define PROTECT(stmt)\
MLCThreads::mpthread_protect_lock();\
               stmt;\
MLCThreads::mpthread_protect_unlock()

#ifdef PTHREADS_DBG
#define DBGPROTECT(stmt)\
MLCThreads::mpthread_protect_lock();\
               stmt;\
MLCThreads::mpthread_protect_unlock()
#else
#define DBGPROTECT(stmt)	//
#endif /* PTHREADS_DBG */


// distributed malloc to be defined in threaded classes
#define THREAD_MEMORY(type)\
\
void* operator new(size_t bytes)\
{\
   if (is_master())\
      return ::operator new(bytes);\
   void *p = amalloc(bytes, (void*) MLCThreads::get_arena());\
   if (p == NULL)\
      err << "type::new out of thread memory" << fatal_error;\
   return p;\
}\
\
void operator delete(void* p, size_t)\
{\
   if (is_master())\
      ::operator delete(p);\
   else\
      afree(p, (void*) MLCThreads::get_arena());\
}

// is a global variable with scheduling information
extern MLCThreads threadControl;


#else /* PTHREADS */

inline Bool is_master()		{ return TRUE; }
#define PROTECT(stmt)		stmt
#define GROW_STACK		//
#define THREAD_MEMORY(type)	//

#endif /* PTHREADS */

#endif
