// MLC++ - Machine Learning Library in -*- C++ -*-
// See Descrip.txt for terms and conditions relating to use and distribution.

// This is an include file.  For a full description of the class and
// functions, see the file name with a "c" suffix.

#ifndef _MOption_h
#define _MOption_h 1

class MEnum;

class MOption {
   NO_DEFAULT_OPS(MOption);
public:
   enum PromptLevel { requiredOnly, promptBasic, promptAll };
   enum OptionFormat { mlc, sh, csh };
private:
   // Member data (also see protected data)
   friend class MLCInit;
   static MEnum *boolEnumPtr;
   static MEnum *promptLevelEnumPtr;
   
protected:
   // Protected member data
   MString name;
   MString defValue;
   MString help;
   Bool nuisance;
   Bool noPrompt;
   static PromptLevel promptLevel;

   // Protected methods
   static MEnum& promptLevelEnum() { ASSERT(promptLevelEnumPtr);
                                     return *promptLevelEnumPtr; }
   virtual Bool result_OK(const MString&) const = 0;
   virtual void prompt_user(const MString& value, Bool full=FALSE) const;
   
   MOption(const MString& aName, const MString& defVal,
	   const MString& help, Bool aNuisance, Bool noPrompt);

public:
   static MEnum& boolEnum() { ASSERT(boolEnumPtr);  return *boolEnumPtr; }
   virtual void OK() const;
   static PromptLevel get_prompt_level() { return promptLevel; }
   static void set_prompt_level(PromptLevel pl) { promptLevel = pl; }
   virtual MString get(Bool hasDefault) const;
   static void dump_option(MLCOStream& out, OptionFormat format,
			   const MString& name, const MString& val);
   static void initialize();
   static void finish();
};


class StringOption : public MOption {
   NO_DEFAULT_OPS(StringOption);
protected:
   // Protected methods
   virtual void prompt_user(const MString& value, Bool full = FALSE) const;
   virtual Bool result_OK(const MString&) const { return TRUE; }

public:
   StringOption(const MString& aName, const MString& defVal,
		const MString& help = EMPTY_STRING, Bool aNuisance = FALSE,
		Bool noPrompt = FALSE);
};


class IntOption : public MOption {
   NO_DEFAULT_OPS(IntOption);
protected:
   // Protected methods
   virtual void prompt_user(const MString& value, Bool full = FALSE) const;
   virtual Bool result_OK(const MString& result) const;
   
public:
   IntOption(const MString& aName, const MString& defVal,
	     const MString& help = EMPTY_STRING, Bool aNuisance = FALSE,
	     Bool noPrompt = FALSE);
};

class IntRangeOption : public IntOption {
   NO_DEFAULT_OPS(IntRangeOption);
protected:
   // Preotected member data
   int lowerBound;
   int upperBound;

   // Protected methods
   virtual Bool result_OK(const MString& result) const;
   virtual void prompt_user(const MString& value, Bool full = FALSE) const;

 public:
   virtual void OK() const;
   IntRangeOption(const MString& aName, const MString& defVal,
		  int lower, int upper,
		  const MString& help = EMPTY_STRING, Bool aNuisance = FALSE,
		  Bool noPrompt = FALSE);
};

class RealOption : public MOption {
   NO_DEFAULT_OPS(RealOption);
protected:
   // Protected methods
   virtual Bool result_OK(const MString& result) const;
   virtual void prompt_user(const MString& value, Bool full = FALSE) const;
   
public:
   RealOption(const MString& aName, const MString& defVal,
	      const MString& help, Bool aNuisance = FALSE,
	      Bool aNoPrompt = FALSE);
};

class RealRangeOption : public RealOption {
   NO_DEFAULT_OPS(RealRangeOption);
protected:
   // Protected member data
   Real lowerBound;
   Real upperBound;
   // Protected methods
   virtual Bool result_OK(const MString& result) const;
   virtual void prompt_user(const MString& value, Bool full = FALSE) const;

public:
   virtual void OK() const;
   RealRangeOption(const MString& aName, const MString& def,
		   Real lower, Real upper,
		   const MString& hlp = EMPTY_STRING, Bool aNuisance = FALSE,
		   Bool noPrompt = FALSE);
};

class MEnumOption : public MOption {
   NO_DEFAULT_OPS(MEnumOption);
protected:
   // Protected member data
   const MEnum& enumSpec;
   // Protected methods
   virtual Bool result_OK(const MString& result) const;
   virtual void prompt_user(const MString& value, Bool full = FALSE) const;

public:
   MEnumOption(const MString& aName, const MEnum& mEnum, const MString& defVal,
	       const MString& help = EMPTY_STRING, Bool aNuisance = FALSE,
	       Bool noPrompt = FALSE);
};


#endif




   


