// MLC++ - Machine Learning Library in -*- C++ -*-
// See Descrip.txt for terms and conditions relating to use and distribution.

// This is an include file.  For a full description of the class and
// functions, see the file name with a "c" suffix.

#ifndef _NaiveBayesCat_h
#define _NaiveBayesCat_h 1

#include <Categorizer.h>
#include <MLCStream.h> 
#include <InstanceList.h>
#include <StatData.h>
#include <Array2.h>
#include <MEnum.h>
#include <DisplayPref.h>

enum NBWidthOption { fillRow, leftAlign, centerAlign, justify };
enum UnknownIsValueEnum { unknownNo, unknownYes, unknownAuto };
extern const MEnum unknownIsValueMEnum;

// structure to hold values needed to express Normal Density function.
// allow default copy constructor
class NBNorm {
   NO_COPY_CTOR(NBNorm);
   friend class NaiveBayesCat;
   friend class PerNaiveBayesCat_;

   // Member data
   Real mean;
   Real var;
   Bool hasData;
public:
   NBNorm( Real m=0.0, Real v=0.0, Bool hd=FALSE );
   NBNorm& operator=(const NBNorm& rhs);
   void set_mean_and_var(Real m, Real v);
   Bool operator==(const NBNorm& rhs) const;
   Bool operator!=(const NBNorm& rhs) const;
};


// NaiveBayes Persistence class; holds NB data in a simpler form to make
// it easier to use with persistent storage
class PerNaiveBayesCat_ : public PerCategorizer_ {
   NO_DEFAULT_OPS(PerNaiveBayesCat_);
public:
   // 1.1 members
   PerBagCounters_ counters;
   Array2<NBNorm> *continNorm;
   Real trainWeight;
   Bool useLaplace;
   Real mEstimateFactor;
   Array<Real> *attrImportance;

   // 1.2 members
   UnknownIsValueEnum unknownIsValue;
   Real klThreshold;
   Real noMatchesFactor;

   // Evidence projection
   Bool useEvidenceProjection;
   Real evidenceFactor;

   PerNaiveBayesCat_();
   virtual void read(MInStore& in);
   virtual void write(MOutStore& out);
};


// NaiveBayes Categorizer returns the category with the highest likelihood 
// given the contents of the training set, assuming independence between
// attributes.
class NaiveBayesCat : public Categorizer {
   NO_DEFAULT_OPS(NaiveBayesCat);

   // Member data (also see public data)
   BagCounters nominCounts;		// hold data on nominal attributs
   Array2<NBNorm>* continNorm;		// hold data on real attributes
   Real trainWeight;
   int numAttributes;
   Bool useLaplace;                     // turn on to activate Laplace
                                        // correction
   Real mEstimateFactor;                // noise in Laplace correction
   Array<Real> *attrImportance;         // importance values per attribute

   Array<Bool> *unkIsVal;               // should unknowns be special values?
                                        // decisions per attribute

   // How to treat unknowns
   UnknownIsValueEnum unknownIsValue;
   Real klThreshold;
   
   // Fraction of a single occurence to use in cases when a label
   // has no occurences of a given nominal value in the training set:
   Real noMatchesFactor;

   // Evidence projection: flag to turn it on, and scale factor
   Bool useEvidenceProjection;
   Real evidenceFactor;

   // Private methods
   void log_prob(const Array<Real>& prob, const SchemaRC& instSchema) const;
   Real generate_probability_prior(Real labelCount, int numLabels) const;
   Real generate_cond_probability(Real labelValCount, Real labelCount,
				  int numAttrVals, int numAttr) const;
   Real kl_distance(int attrNum) const;
   void compute_kl_distances();

public:
   // Public data
   // Categorizer option defaults
   static const Real defaultMEstimateFactor;
   static const Bool defaultLaplaceCorrection;
   static const UnknownIsValueEnum defaultUnknownIsValue;
   static const Real defaultKLThreshold;
   static const Real defaultNoMatchesFactor;
   static const Bool defaultUseEvidenceProjection;
   static const Real defaultEvidenceFactor;
   // Value to use for Variance when actual variance = 0:
   static Real epsilon;
   // Value to use for Vaiance when actual variance is undefined becase there
   // is only one occurance.
   static const Real defaultVariance;

   // Methods
   virtual void OK(int = 0) const;

   NaiveBayesCat(const MString& dscr, const InstanceList& instList);
   NaiveBayesCat(const NaiveBayesCat&, CtorDummy);
   NaiveBayesCat(const PerNaiveBayesCat_& data, const SchemaRC& schema);
   
   virtual ~NaiveBayesCat();
   virtual AugCategory categorize(const InstanceRC&) const;
   virtual CatDist *score(const InstanceRC&) const;
   virtual Bool supports_scoring() const { return TRUE; }
   
   virtual void display_struct(MLCOStream& stream = Mcout,
			const DisplayPref& dp = defaultDisplayPref) const;
   
   virtual void generate_viz(MLCOStream& stream,
			     const BoolArray *autoDiscVector = NULL,
			     int evivizVersion = 200) const;

   // Returns total weight used in training
   virtual Real total_train_weight() const { return trainWeight; }

   virtual Bool supports_backfit() const { return TRUE; }
   virtual void backfit(const InstanceList& instList);

   // Returns a pointer to a deep copy of this NaiveBayesCat
   virtual Categorizer* clone() const;

   virtual int class_id() const {return CLASS_NB_CATEGORIZER;}

   virtual Bool operator==(const Categorizer& rhs) const;

   virtual void set_use_laplace(Bool lap) { useLaplace = lap; }
   virtual Bool get_use_laplace() const { return useLaplace; }

   virtual void set_m_estimate_factor(Real m);
   virtual Real get_m_estimate_factor() const { return mEstimateFactor; }

   virtual void set_unknown_is_value(UnknownIsValueEnum unk) {
      unknownIsValue = unk;
   }
   virtual UnknownIsValueEnum get_unknown_is_value() const {
      return unknownIsValue;
   }

   virtual void set_use_evidence_projection(Bool b)
     { useEvidenceProjection = b;}
   virtual Bool get_use_evidence_projection() const
     { return useEvidenceProjection; }

   virtual void set_evidence_factor(Real f) {
      evidenceFactor = f; }
   virtual Real get_evidence_factor() const { return evidenceFactor; }

   virtual void set_no_matches_factor(Real nm) { noMatchesFactor = nm; }
   virtual Real get_no_matches_factor() const { return noMatchesFactor; }

   virtual void set_kl_threshold(Real th) { klThreshold = th; }
   virtual Real get_kl_threshold() const { return klThreshold; }
   
   // persistence
   virtual PerCategorizer_ *build_persistent_object() const {
      return new PerNaiveBayesCat_;
   }
   virtual void make_persistent(PerCategorizer_&) const;
   virtual Bool supports_persistence() const { return TRUE; }
   
};

#endif



