// MLC++ - Machine Learning Library in -*- C++ -*-
// See Descrip.txt for terms and conditions relating to use and distribution.

// This is an include file.  For a full description of the class and
// functions, see the file name with a "c" suffix.

#ifndef _NodeCategorizer_h
#define _NodeCategorizer_h 1

#include <Categorizer.h>
#include <TDDTInducer.h>

// structure used for node accuracies.  Stores the total weight of
// instances this node is responsible for, and also the total
// losses incurred by this node.
struct NodeLoss {
   // Default copy ctor is fine.

   // Public member data
   Real totalWeight;
   Real totalLoss;
   Real totalLossSquared; // Allow computing std-dev

   // Methods
   NodeLoss()
      : totalWeight(UNDEFINED_REAL), totalLoss(UNDEFINED_REAL),
	totalLossSquared(UNDEFINED_REAL)
   {}
   void display(MLCOStream& stream = Mcout) const;
   void update(Real weight, Real loss)
   {
      totalWeight += weight;
      totalLoss += weight * loss;
      totalLossSquared += weight * loss * loss;
   }
   Bool operator==(const NodeLoss& rhs) const;
   Bool operator!=(const NodeLoss& rhs) const
      {return ! ((*this)== rhs);}
   NodeLoss& operator=(const NodeLoss& rhs);
};

DECLARE_DISPLAY(NodeLoss)

class PerNodeCategorizer_ : public PerCategorizer_ {
   NO_DEFAULT_OPS(PerNodeCategorizer_);
public:
   PerNodeCategorizer_() { }
   virtual void read(MInStore& in) { PerCategorizer_::read(in); }
   virtual void write(MOutStore& out) { PerCategorizer_::write(out); }
};

class NodeCategorizer : public Categorizer {  // ABC
   NO_DEFAULT_OPS(NodeCategorizer);

   // Member data
   NodeLoss lossInfo;
   NodePtr nodePtr;
   CGraph *cGraph;
   Bool smoothDistribution;
   Real smoothFactor;
   DBG_DECLARE(Bool checkGraph;)
protected:
   // Protected methods
   const CGraph& get_graph() const;  // @@ these two should be private, but
   NodePtr get_node() const;         // @@ MultiSplitCat needs them
   void update_loss(Real weight, Real loss) { lossInfo.update(weight, loss); }
   
public:
   virtual void OK(int level = 0) const;
   virtual void OK_expect(NodePtr expectedNode, CGraph *expectedGraph,
			  int level) const;
   
   NodeCategorizer(int noCat, const MString& dscr, const SchemaRC& schema);
   NodeCategorizer(const NodeCategorizer& source, CtorDummy);
   NodeCategorizer(const PerNodeCategorizer_& data, const SchemaRC& schema);
   virtual ~NodeCategorizer() { DBG(OK(0)); }

   virtual Bool operator==(const Categorizer& rhs) const = 0;

   // Pure virtual function: returns an edge (number + description)
   // given an instance--this determines how this categorizer splits
   virtual AugCategory branch(const InstanceRC& inst) const = 0;
   virtual AugCategory *get_edge_label(int branchNum) const = 0;
      
   // Node / graph functions
   void set_graph_and_node(CGraph* graph, NodePtr node);
   Bool in_graph() const { return cGraph != NULL; }
   const NodeLoss& get_loss() const { return lossInfo; }  
   const NodeCategorizer& get_child_categorizer(AugCategory branch) const;
   const NodeCategorizer& get_child_categorizer(const InstanceRC& inst) const;
   int num_children() const { return get_graph().outdeg(get_node()); }
   virtual void add_child(NodePtr childNode, int branchNum);
   
   void reset_node_loss();
   virtual void add_instance_loss(const InstanceRC& instance,
				  const CatDist& pred);

   // Scoring functions
   virtual AugCategory categorize(const InstanceRC&) const;
   virtual Bool supports_scoring() const { return TRUE; }
   virtual CatDist *score(const InstanceRC& inst) const
   { return score(inst, FALSE); }
   virtual CatDist *score(const InstanceRC& inst, Bool addLoss) const;

   virtual InstanceListArray* split_instance_list(const InstanceList& il)
      const;

   virtual Bool supports_backfit() const { return TRUE; }
   virtual void backfit(const InstanceList&) {
      err << "NodCategorizer::backfit: call distribute_instances instead"
	  << fatal_error;
   }
   virtual void distribute_instances(const InstanceList& il,
				     Real pruningFactor,
				     Real &pessimisticErrors,
				     TDDTInducer::LeafDistType ldType,
				     Real leafDistParameter,
				     const Array<Real> *parentWeightDist,
				     Bool saveOriginalDistr);

   // Persistence
   virtual PerCategorizer_ *build_persistent_object() const
   { return new PerNodeCategorizer_; }
   virtual void make_persistent(PerCategorizer_& data) const
   { Categorizer::make_persistent(data); }
   virtual Bool supports_persistence() const { return TRUE; }

   // For casting from Categorizers. @@ This should go away when we switch
   // to -n32 (then we can use dynamic_cast instead of these)
   virtual Bool can_cast_to_node_categorizer() const { return TRUE; }
   virtual NodeCategorizer& cast_to_node_categorizer() { return *this; }
   virtual const NodeCategorizer& cast_to_node_categorizer() const
   {return *this; }
};
   
#endif
