// MLC++ - Machine Learning Library in -*- C++ -*-
// See Descrip.txt for terms and conditions relating to use and distribution.

// This is an include file.  For a full description of the class and
// functions, see the file name with a "c" suffix.

#ifndef _ODTInducer_h
#define _ODTInducer_h 1

#include <Array.h>
#include <DecisionTree.h>
#include <DTCategorizer.h>
#include <IncrInducer.h>
#include <InstanceList.h>
#include <ODTCategorizer.h>

// @@ add persistance

class LeafCategorizer;

// Abstract Base Class for Oblivious Decision Trees
// Hierarchy: ODTInducer --> IncrInducer --> Inducer --> BaseInducer
class ODTInducer : public IncrInducer {
   NO_COPY_CTOR(ODTInducer);

   // Member data (also see protected data)
   ODTCategorizer::MajorityType zeroWeightBehavior;
   ODTCategorizer::MajorityType tieBehavior;

   // Private methods
   // helper functions
   // No ownership transfer, but currentNodeSet gets modified
   void update_bad_cats(DecisionTree &decisionTree,
			DynamicArray<NodePtr> &currentNodeSet,
			DynamicArray<const InstanceList*>&currentLevelILs,
			const NodeCategorizer *splitCat) const;
   // No ownership transfer, but decisionTree, currentLevelILs, newNodeSet,
   // and nextLevelILs are modified.
   void split_current_level(DecisionTree &decisionTree,
			    DynamicArray<NodePtr> &currentNodeSet,
			    DynamicArray<const InstanceList*> &currentLevelILs,
			    DynamicArray<NodePtr> &newNodeSet,
			    DynamicArray<const InstanceList*> &nextLevelILs,
			    const NodeCategorizer *splitCat,
			    Array<MString>* catNames,
			    int &nodeNumber, int&) const;
   // Gets ownership of the instance list.
   virtual LeafCategorizer* create_leaf_categorizer(InstanceList*&) const;
   // No ownership transfer, but currentNodeSet gets modified
   void update_leaf_bad_cats(DecisionTree &decisionTree,
			     DynamicArray<NodePtr> &currentNodeSet,
			     DynamicArray<const InstanceList*> &currentLevelILs)
      const;
   NodePtr induce_oblivious_decision_tree(CGraph& aCgraph);

protected:
   // Protected member data
   CGraph *cgraph; // Graph to build this in, or NULL if to create one.
   ODTCategorizer *ODTCat;
   AttrOrder ao; // for use in OrderFSS
   int current; // counter in the AttrOrder array

   // Protected method
   // Best_split finds the best split in the node and returns a categorizer
   // implementing it.  Allocates catNames which must be freed by the caller.
   virtual NodeCategorizer *
       best_split(DynamicArray<const InstanceList*> &IL_array,
		  Array<MString> *& catNames );

public:

   // Check invariants
   virtual void OK(int level = 0) const;

   // Constructor, Destructor
   ODTInducer(const MString& descr, CGraph* aCgraph = NULL);
   ~ODTInducer();

   // Pure Virtual functions from abstract class Inducer

   // Train the inducer and obtain an ODTCategorizer
   virtual void train();
   virtual Bool was_trained(Bool fatalOnFalse = TRUE) const;

   // Pure Virtual functions from abstract class IncrInducer

   // (Incremental operations)
   virtual ILPix add_instance(const InstanceRC &);   
   virtual InstanceRC del_instance(ILPix& pix);
   virtual Bool can_cast_to_incr_inducer() const { return TRUE; }
   virtual IncrInducer& cast_to_incr_inducer() { return *this; }

   // Access Functions
   // Returns the Class ID
   virtual int class_id() const { return ODT_INDUCER; }
   // Returns a read-only copy of the ODTCat
   virtual const Categorizer& get_categorizer() const; 
   virtual Categorizer *release_categorizer();
   ODTCategorizer::MajorityType get_zero_weight_behavior() const
                                                { return zeroWeightBehavior; }
   void set_zero_weight_behavior(ODTCategorizer::MajorityType type)
                                                { zeroWeightBehavior = type; }
   ODTCategorizer::MajorityType get_tie_behavior() const
                                                { return tieBehavior; }
   void set_tie_behavior(ODTCategorizer::MajorityType type)
                                                { tieBehavior = type; }
   virtual AttrOrder& get_attr_order_info() { return ao; }
   // set_order() is meant for interfacing with OrderFSS.
   void set_order(const Array<int>& order) { ao.set_order(order);}
   // @@ what about the ODTCat's copy of AttrOrder ?

   // ask user to set options that control the behavior of ODTInducer.  See
   // .c file for description of options.
   virtual void set_user_options(const MString& preFix);

   // Output Functions
   virtual void display_struct(MLCOStream& stream = Mcout,
			       const DisplayPref& dp = defaultDisplayPref)
      const;

};

#endif
