// MLC++ - Machine Learning Library in -*- C++ -*-
// See Descrip.txt for terms and conditions relating to use and distribution.

// This is an include file.  For a full description of the class and
// functions, see the file name with a "c" suffix.

#ifndef _PerfEstDispatch_h
#define _PerfEstDispatch_h 1

#include <PerfData.h>
#include <InstanceList.h>
#include <BaseInducer.h>
#include <MEnum.h>
#include <LogOptions.h>
#include <PerfEstimator.h>

class PerfEstDispatch {
   LOG_OPTIONS;
   NO_COPY_CTOR(PerfEstDispatch);
public:
   enum ErrorType {classError, meanSquaredError, meanAbsoluteError,
		   classLoss};
   // note: PerfEstimationMethod is public but must be declared BEFORE
   // perfEstimationMethod, which is private.  Hence, the violation of
   // ordering rules here.
   enum PerfEstimationMethod {cv, stratCV, testSet, bootstrap, holdOut,
			      automatic};

   // Public data
   static const MEnum perfEstimationMethodEnum;

   static const MEnum errorTypeEnum;
private:
   // Member data (also see public data)
   // PerfEstimator level options
   Real errTrim;
   Real stdDevTrim;
   unsigned int randSeed;
   Real cvWeightThreshold;
   
   // CV options
   int cvFolds;
   int cvTimes;
   int maxTimes;
   Real cvFraction;
   Real desStdDev;

   // Bootstrap options
   int bootstrapTimes;
   Real bootstrapFraction;

   // HoldOut options
   int hoTimes;
   int hoNumber;
   Real hoPercent;

   // results
   PerfData perfData;
   int actualTimes;   

   PerfEstimationMethod perfEstimationMethod;
   ErrorType errorType;

   // Private method
   Real estimate_performance(BaseInducer& inducer,
			     InstanceList *trainInstList,
			     InstanceList *testInstList,
			     PerfData *providedPerfData);
   
public:
   PerfEstDispatch() { set_defaults(); }

   void set_defaults();
   void set_user_options(const MString& prefix = EMPTY_STRING);

   // almost like the copy constructor, but doesn't copy perfData.  
   void copy_options(const PerfEstDispatch& source);

   Real estimate_performance(BaseInducer& inducer,
			  const InstanceList& trainList,
			  const InstanceList& testList,
			  PerfData *pPerfData = NULL);
   Real estimate_performance(BaseInducer& inducer,
			  const InstanceList& trainList,
			  PerfData *pPerfData = NULL);
      
   // generic parameters
   void set_perf_estimator(PerfEstimationMethod perfEst);
   PerfEstimationMethod get_perf_estimator() const
   { return perfEstimationMethod; }
   void set_error_type(ErrorType errType) { errorType = errType; }
   ErrorType get_error_type() const { return errorType; }
   void set_error_trim(Real trim) { errTrim = trim; }
   Real get_error_trim() const { return errTrim; }
   void set_seed(const unsigned int seed) { randSeed = seed; }
   unsigned int get_seed() const { return randSeed; }
   void set_cv_weight_threshold(Real threshold)
   { cvWeightThreshold = threshold; }
   Real get_cv_weight_threshold() const { return cvWeightThreshold; }
   
   // Parameters for cross validation.
   void set_cv_folds(int folds);
   int  get_cv_folds() const { return cvFolds; }
   void set_cv_times(int times) { cvTimes = times; }
   int  get_cv_times() const { return cvTimes; }
   void set_cv_fraction(Real fract) { cvFraction = fract; }
   Real get_cv_fraction() const { return cvFraction; }
   void set_max_times(int times) { maxTimes = times; }
   int  get_max_times() const { return maxTimes; }
   void set_desired_std_dev(Real dstd) { desStdDev = dstd; }
   Real get_desired_std_dev() const { return desStdDev; }

   // Parameters for bootstrap.
   void set_bootstrap_times(int times);
   int  get_bootstrap_times() const { return bootstrapTimes; }
   void set_bootstrap_fraction(Real fract);
   Real get_bootstrap_fraction() const { return bootstrapFraction; }

   // Parameters for holdout
   void set_holdout_times(int times) { hoTimes = times; }
   int get_holdout_times() const { return hoTimes; }
   void set_holdout_number(int num) { hoNumber = num; }
   int get_holdout_number() const { return hoNumber; }
   void set_holdout_percent(Real pct) { hoPercent = pct; }
   Real get_holdout_percent() const { return hoPercent; }

   // get results; get_error_data returns your chosen error data based
   // on errorType.  get_perf_data gets the full suite of error stats.
   virtual const ErrorData& get_error_data() const;
   virtual const PerfData& get_perf_data() const;

   // get number of times auto-cv was actually run
   virtual int get_actual_times() const { return actualTimes; }
   
   // display:  we can display settings, performance information, or both
   void display_settings(MLCOStream& stream) const;
   void display_performance(MLCOStream& stream) const;
   void display(MLCOStream& stream) const;
};

DECLARE_DISPLAY(PerfEstDispatch)


#endif
