// MLC++ - Machine Learning Library in -*- C++ -*-
// See Descrip.txt for terms and conditions relating to use and distribution.

// This is an include file.  For a full description of the class and
// functions, see the file name with a "c" suffix.

#ifndef _PersistentObject_h
#define _PersistentObject_h 1

#include <Persistence.h>
#include <Array.h>
#include <MLCFactory.h>

// Persistent helpers have two states: full and empty.  Full
// helpers have data and are enabled for writing and object
// construction.  Empty helpers have no data and are enabled for
// reading and filling from an object.
// The associated_id() returns the class ID of the associated
// object.
class PersistentHelper_ {
   friend class PersistentObject;
   NO_DEFAULT_OPS(PersistentHelper_);

   // Member data
   Bool isFull;   
public:
   PersistentHelper_();
   virtual ~PersistentHelper_();
   virtual int class_id() const = 0;
   virtual int associated_id() const = 0;
   Bool is_full() const { return isFull; }

   virtual void read(MInStore& in);
   virtual void write(MOutStore& out);
};


class PersistentObject {
   NO_DEFAULT_OPS(PersistentObject);
 public:
   PersistentObject() { };
   PersistentObject(PersistentHelper_& helper);
   virtual ~PersistentObject() { };
   virtual int class_id() const = 0;
   virtual PersistentHelper_ *create_my_helper() = 0;
   virtual void fill_helper(PersistentHelper_& helper) const;
};


// Definitions so that persistent objects and helpers work in the factory
typedef PersistentObject *
      (*PersistentObjectBuilder)(PersistentHelper_*&);
typedef PersistentHelper_ *
      (*PersistentHelperBuilder)(void);

// Define macros to help declare and define persistent objects

// Use this macro at the top of the public section of the
// persistent object class
#define DECLARE_PERSISTENT_OBJECT(classID, Object, Helper) \
  static int staticClassID_;  \
  static PersistentObject *create_this_type(PersistentHelper_*&);  \
  virtual int class_id() const { return classID; }  \
  virtual PersistentHelper_ *create_my_helper() { \
     return new Helper; }

// Use this macro at the top of the public section of the persistent
// helper class
#define DECLARE_PERSISTENT_HELPER(classID, Object, Helper) \
  static int staticClassID_; \
  static PersistentHelper_ *create_this_type(); \
  virtual int associated_id() const { return classID; } \
  virtual int class_id() const { return staticClassID_; }

// Use this macro at the top of the .c file containing both the
// object and the helper
// GNU needs the explicit casts to (void *).
#define DEF_PERSISTENT(classID, Object, Helper) \
  PersistentHelper_ *Helper::create_this_type() { \
    return new Helper; } \
  int Helper::staticClassID_ = \
    mlcFactory.register_class("PersistentHelper_", # Helper, \
			      (void*) &Helper::create_this_type); \
  PersistentObject *Object::create_this_type(PersistentHelper_*& helper) \
    { return new Object(*helper); } \
  int Object::staticClassID_ = \
    mlcFactory.register_class("PersistentObject", # Object, \
			      (void *) &Object::create_this_type, classID, \
			      Helper::staticClassID_);
  
// This function builds a persistent object given only its helper.
PersistentObject *
create_persistent_object(PersistentHelper_ *& helper);

// This function builds a persistent helper, either from the OBJECT's
// classID, or from the OBJECT's className.
PersistentHelper_ *create_persistent_helper(int objClassID);
PersistentHelper_ *create_persistent_helper(const MString& objClassName);

#endif
