// MLC++ - Machine Learning in -*- C++ -*-
// See Descrip.txt for terms and conditions relating to use and distribution.

/*****************************************************************************
  Description  : This is a class for an Universal pointer hash table. It
                    derives from PtrHashTable<KeyType, DataType> and 
		    defines a Universal hash function. The universal hashing
		    function is chosen at run time from a collection
		    of such functions and the expected number of
		    collisions involving a particular key is less than
		    1. For details, refer to 'Introduction to
		    Algorithms', Cormen et al, p230-p231, MIT press.
  Assumptions  : hash() function assumes KeyType can be cast to char*
                    (without loss of semantics, i.e. that key does not
		    contain pointers within it). In other cases, new
		    hash() function may be defined  properly in any
		    derived classes from this class.
		 The KeyType should be a class which has a length()
		    member function returning the length of key in bytes.
  Comments     : The size of hash table, if it is less than 2^8+1, is
                    extended to 2^8+1. This is a necessary condition for
		    universal hashing functions to work properly.
  Complexity   : hash() takes O(m), where m is the length of keyValue.
  Enhancements :
  History      : Clay Kunz                                        5/30/96
                   Made PtrUniversalHashTable from UniversalHashTable
                 Yeogirl Yun                                      4/17/94
                   initial revision
*****************************************************************************/ 

#include <basics.h>
#include <PtrUnivHash.h>



/******************************************************************************
  Description  : Copy constructor which generates a fatal error when
                   called.
  Comments     :
******************************************************************************/
template <class KeyType, class DataType>
PtrUniversalHashTable<KeyType, DataType>::PtrUniversalHashTable
(const PtrUniversalHashTable<KeyType, DataType>& )
   : PtrHashTable<KeyType, DataType>(1), randomData(0)
{
   err << "PtrUniversalHashTable<>::PtrUniversalHashTable : copy"
          " constructor is not allowed. Use copy constructor with"
          " dummy arguments" << fatal_error;
}



/******************************************************************************
  Description  : Copy constructor with dummy argument.
                 It initializes randomData and pass the source argument over
      		   to base class' copy constructor.
		 It should not be called by any derived class when it defined
		   its own hash function because within this copy constructor
		   PtrUniversalHashTable<>::hash() would be used.
  Comments     :
******************************************************************************/
template <class KeyType, class DataType>
PtrUniversalHashTable<KeyType, DataType>::PtrUniversalHashTable
   (const PtrUniversalHashTable<KeyType, DataType>& source, CtorDummy)
   : PtrHashTable<KeyType, DataType>
   (MLC::max((int)(source.num_items()/loadFactor),
	    source.size())),
     randomData(0)
{
   merge(source);
}




/******************************************************************************
  Description  : Constructor with estimated number arguments.
  Comments     : The size samller than (UCHAR_MAX + 1) is extened to
                   the size of UCHAR_MAX + 1 without any error
		   message. This is to make hash table size greater
		   than 2^8. This will work even in the case
		   loadFactor is 1.
******************************************************************************/

// we need to cast UCHAR_MAX to an integer to allow the templated max
// function to instantiate properly

template <class KeyType, class DataType>
PtrUniversalHashTable<KeyType, DataType>::
   PtrUniversalHashTable(int estimatedNum)
 : randomData(0),
   PtrHashTable<KeyType, DataType>(MLC::max((int)(UCHAR_MAX) + 1,
					    estimatedNum))
{
   if (estimatedNum <= 0)
      err << "PtrUniversalHashTable::PtrUniversalHashTable: bad estimated "
	  << "number of elements: " << estimatedNum << fatal_error;
}



/******************************************************************************
  Description  : Constructor with estimated number arguments and seed for
                   RandCharArray.
  Comments     : The size samller than (UCHAR_MAX + 1) is extened to
                   the size of UCHAR_MAX + 1 without any error
		   message. This is to make hash table size greater
		   than 2^8. This will work even in the case
		   loadFactor is 1.
******************************************************************************/
template <class KeyType, class DataType>
PtrUniversalHashTable<KeyType, DataType> :: PtrUniversalHashTable
   (int estimatedNum, unsigned int seed) : randomData(0, seed),
   PtrHashTable<KeyType, DataType>(MLC::max((int)(UCHAR_MAX) + 1,
					    estimatedNum))
{
   if (estimatedNum <= 0)
      err << "PtrUniversalHashTable::PtrUniversalHashTable: bad "
	  << "estimated number of elements: " << estimatedNum << fatal_error;
}





/******************************************************************************
  Description  : Universal hash function. It assumes the length of
                   keyValue is greater than 0( >= 1 ).
  Comments     : It requires the size of hash table should be greater
                   than 2^8, or 256.
******************************************************************************/
template <class KeyType, class DataType>
int PtrUniversalHashTable<KeyType, DataType>::hash(const KeyType& keyValue)
     const
{

   const unsigned char* ptr = keyValue.get_internal_rep();
   long hashValue = 0;

   // note : To iterate from the end helps speed execution time since
   //        it obviates unnecesary iterative extensions of array.
   //        For details, see the DynamicArray.c file.
   for(int i = keyValue.length()-1; i >= 0 ; i-- ) 
      hashValue += ptr[i] * // constness cast away : randomData, being a
			    // dynamic array, does not allow const function.
      ((PtrUniversalHashTable<KeyType, DataType> *)this)->
	 randomData.index(i);

   return hashValue  % size();
}
    





