// MLC++ - Machine Learning Library in -*- C++ -*-
// See Descrip.txt for terms and conditions relating to use and distribution.

// This is an include file.  For a full description of the class and
// functions, see the file name with a "c" suffix.

#ifndef _RDGCategorizer_h
#define _RDGCategorizer_h 1

#include <Categorizer.h>
#include <RootCatGraph.h>
#include <TDDTInducer.h>
#include <entropy.h>
#include <NodeCategorizer.h>

class NodeAndThreshold {
   NO_COPY_CTOR(NodeAndThreshold);
   //  Compiler-generated operator= is ok here.

   // Member data
   Real threshold;
   NodePtr nodePtr;
public:
   NodeAndThreshold();
   NodeAndThreshold(NodePtr node, Real thresh);
   NodeAndThreshold(const NodeAndThreshold & nodeAndThreshold, CtorDummy);
   Real get_threshold() const {return threshold;}
   void set_threshold(Real newThreshold) { threshold = newThreshold;}
   NodePtr get_node() const {return nodePtr;}
   void set_node(NodePtr newNodePtr) {nodePtr = newNodePtr;}
   void display(MLCOStream& stream = Mcout) const;

   static void logNodeThreshold(const int logLevel, const MString& preamble,
				const CGraph& graph1, const NodePtr node1,
				const CGraph& graph2, const NodePtr node2);

   // The following arithmetic comparison methods use exact comparisons,
   //   as they were designed for use while sorting.  For approximate
   //   comparisons (taking into account rounding errors), the MLC
   //   functions declared in basics.h should be used.
   friend Bool operator<(const NodeAndThreshold& s1,
			 const NodeAndThreshold& s2);
   friend Bool operator<=(const NodeAndThreshold& s1,
			  const NodeAndThreshold& s2);
   friend Bool operator==(const NodeAndThreshold& s1,
			  const NodeAndThreshold& s2);
   friend Bool operator!=(const NodeAndThreshold& s1,
			  const NodeAndThreshold& s2);
   friend Bool operator>=(const NodeAndThreshold& s1,
			  const NodeAndThreshold& s2);
   friend Bool operator>(const NodeAndThreshold& s1,
			 const NodeAndThreshold& s2);
};

DECLARE_DISPLAY(NodeAndThreshold)


class PerRDGCategorizer_ : public PerCategorizer_ {
public:
   PerRootedCatGraph_ rcGraph;
   TDDTInducer::LeafDistType leafDistType;
   Real leafDistParameter;

   void read(MInStore& in);
   void write(MOutStore& out);
};


// RDG is rooted decision graph
class RDGCategorizer : public Categorizer {
   NO_DEFAULT_OPS(RDGCategorizer);
   
// DTCategorizer needs access to this member
protected:
   // Protected member data
   RootedCatGraph* rcGraph;
   TDDTInducer::LeafDistType leafDistType;
   Real leafDistParameter;
   Bool hasNodeLosses;
public:
   virtual void OK(int level = 0)const;
   // Gets ownership of RootedCatGraph
   RDGCategorizer(RootedCatGraph*& rcg, const MString& dscr, int numCat,
		  const SchemaRC& schema);
   RDGCategorizer(const RDGCategorizer& source, CtorDummy);
   RDGCategorizer(const PerRDGCategorizer_& data,
		  const SchemaRC& schema,
		  NodeInfo*& prototype,
		  const Array<NodeInfo *>& nodes);

   virtual ~RDGCategorizer();
   virtual AugCategory categorize(const InstanceRC&) const;   
   virtual CatDist *score(const InstanceRC&) const;
   virtual Bool supports_scoring() const { return TRUE; }
   
   // The nodeLoss struct allows incrementing counts for nodes
   // that are blamed for making a wrong prediction.
   virtual CatDist *score_from_node(const InstanceRC&, 
				    NodePtr currentNode,
				    Bool computeNodeLosses = FALSE) const;
   void compute_node_losses(const InstanceList& testList) const;
   Array<NodeLoss> *get_node_losses() const;
   virtual void display_struct(MLCOStream& stream = Mcout,
			       const DisplayPref& dp = defaultDisplayPref)
      const;
   virtual Categorizer* clone() const;
   virtual const RootedCatGraph& rooted_cat_graph() const;

   virtual void reset_original_distr();

   // Convenience functions
   virtual int num_nodes() const
       {return rooted_cat_graph().num_nodes();}
   virtual int num_leaves() const 
       {return rooted_cat_graph().num_leaves();}
   virtual int num_nontrivial_nodes() const
       {return rooted_cat_graph().num_nontrivial_nodes();}
   virtual int num_nontrivial_leaves() const 
       {return rooted_cat_graph().num_nontrivial_leaves();}

   virtual int class_id() const {return CLASS_RDG_CATEGORIZER; }
   virtual Bool is_isomorphic(const CGraph& graph1, const NodePtr node1, 
                              const CGraph& graph2, const NodePtr node2,
			      const MString& preamble = EMPTY_STRING) const;
   virtual Bool operator==(const Categorizer& rhs) const ;
   virtual RootedCatGraph* release_rooted_cat_graph();

   virtual Array< DynamicArray <NodeAndThreshold> >*
      create_threshold_array(const CGraph & graph, int numAttributes);
   virtual DynamicArray<Real>* get_attribute_values(int whichAttr,
      const InstanceList& instances);
   virtual void adjust_thresholds_to_attributes(CGraph& graph,
      DynamicArray<NodeAndThreshold>& attrThresholds,
      const DynamicArray<Real>& realInstanceVal);
   virtual void adjust_real_thresholds(const InstanceList& instances);

   virtual void set_used_attr(BoolArray& usedAttr) const;
   virtual void project(const BoolArray& mask);

   virtual void set_leaf_dist_params(TDDTInducer::LeafDistType ldType,
				     Real mEst, Real eviFactor);

   virtual PerCategorizer_ *build_persistent_object() const {
      return new PerRDGCategorizer_; }
   virtual void make_persistent(PerCategorizer_& data) const;
};

#endif
