// MLC++ - Machine Learning Library in -*- C++ -*-
// See Descrip.txt for terms and conditions relating to use and distribution.

// This is an include file.  For a full description of the class and
// functions, see the file name with a "c" suffix.

#ifndef _RealAndLabelColumn_h
#define _RealAndLabelColumn_h 1

#include <Array.h>
#include <Array2.h>
#include <Attribute.h>

class AttrLabelPair {
   NO_OPERATOR_ASSIGN(AttrLabelPair);
public:
   // Public member data
   // A compact representation is preferrable, one might even pack
   //   both the value and the label into 4 bytes in the future
   StoredReal value;
   int label;
   Real weight;

   // Methods
   AttrLabelPair()
      : value(-1), label(UNDEFINED_INT), weight(UNDEFINED_REAL) {}
   AttrLabelPair(const AttrLabelPair& rhs)
      {value = rhs.value; label = rhs.label; weight = rhs.weight;}
   // Operators needed for sorting
   Bool operator<(const AttrLabelPair& rhs) const;
   Bool operator==(const AttrLabelPair& rhs) const;
   Bool operator>(const AttrLabelPair& rhs) const
      {return ! ((*this)== rhs || (*this) < rhs); };
   Bool operator!=(const AttrLabelPair& rhs) const
      {return ! ((*this)== rhs);}
   void display(MLCOStream& stream = Mcout) const;
};

DECLARE_DISPLAY(AttrLabelPair)

class RealAndLabelColumn {
   NO_DEFAULT_OPS(RealAndLabelColumn);

   friend class InstanceList;	  // needed to call add_pair;
   friend class AttributeColumn;  // needed for same reason (for FC)

   // Member data
   Bool ordered;
   const RealAttrInfo& attrInfo;
   const NominalAttrInfo* labelInfo;
   
   int knownCount;		  // Number of known values
   Real knownWeight;              // Total weight of known values
   Array<AttrLabelPair>* known;   // The data with known values; can not
                                  //   use DynamicArray because of split
   Array<Real> knownLabelWeights; // The total weight of knowns associated
                                  //   with each label

   int unknownCount;                 // Number of unknown values
   Real unknownWeight;               // Total weight of unknown values
   Array<AttrLabelPair>* unknown;    // The data with unknown values
   Array<Real> unknownLabelWeights;  // The total weight of unknowns
                                     //   associated with each label

   // Private methods
   void zero_known_weights();
   // Only InstanceList and AttributeColumn can use the following two methods.
   void add_known(StoredReal value, int label, Real weight);
   void add_unknown(int label, Real weight);
   
   void add_known(StoredReal value, Real weight);
   void add_unknown(Real weight);

public:   
   RealAndLabelColumn* split(int splitIndex);
   RealAndLabelColumn(int maxLength, int labelCount,
		      const RealAttrInfo& rai,
		      const NominalAttrInfo& nai);
   RealAndLabelColumn(int maxLength,
		      const RealAttrInfo& rai);
   ~RealAndLabelColumn() {
      delete known; delete unknown;};
   void OK() const;
   int label_count() const { return knownLabelWeights.size(); };
   int known_count() const { return knownCount; };
   Real known_weight() const { return knownWeight; };
   int unknown_count() const { return unknownCount; };
   Real unknown_weight() const { return unknownWeight; };
   int total_count() const { return knownCount + unknownCount; };
   Real total_weight() const { return knownWeight + unknownWeight; };
   // Return the AttrLabelPair block for efficiency
   AttrLabelPair& index(int index) { return known->index(index); };
   void sort() { known->truncate(knownCount); known->sort(); ordered = TRUE; };
   // A shortcut again for the efficiency reasons
   int init_dist_counts(Array2<Real>& splitAndLabelDist,
			Array<Real>& labelDist) const;
   int distinct_value_count();
   void trim_count(Real percent);
   void trim_value(Real percent);
   void trim_weight(Real percent);
   StoredReal min_value();
   StoredReal max_value();
   InstanceList* gen_inst_list();
   void display(MLCOStream& stream = Mcout) const;
   
   const RealAttrInfo& attr_info() const { return attrInfo; }
   RealAndLabelColumn* clone() const;
};


DECLARE_DISPLAY(RealAndLabelColumn)     

#endif


