// MLC++ - Machine Learning Library in -*- C++ -*-
// See Descrip.txt for terms and conditions relating to use and distribution.

// This is an include file.  For a full description of the class and
// functions, see the file name with a "c" suffix.

#ifndef _SGIDTInducer_h
#define _SGIDTInducer_h 1

#include <ID3Inducer.h>
#include <MRandom.h>

enum PartitionMethodType { searchSplit, cartSplit, cartSearch };

class PartitionSplit;
class PartitionEvaluator;
class AttrSubsetCat;
class MultiSplitCat;
class DiscDispatch;

class SGIDTInducer : public ID3Inducer {
   NO_DEFAULT_OPS(SGIDTInducer);
   RAND_OPTIONS;
public:
   enum SplitDecisionType { bestSplit, randomSplit, multipleSplit };
   enum WeightType { noWeight, sizeWeight, entropyWeight };
private:
   // Member data (also see protected data)
   //options
   struct SGIDTOptions{
      Bool multiThresholding;       // Should local multithresholding be used
      Bool alwaysBinary;            // Should splits on continuous attributes
                                    // always be binary?
      int discThreshold;            // Should we stop local multithresholding
                                    // at some point?
      Bool partitionNominals;       // group nominal attribute values
				    // together? 
      PartitionMethodType partitionMethod;   // how to find partition?
      SplitDecisionType splitDecision;  // how to choose which split to make?
      Real fitnessInitialRatio;     // ratio of choices to pick from
      Real fitnessMultRatio;        // how much to reduce previous ratio
      unsigned int randomSeed;      // random seed to use
      int multiSplitMaxDepth;       // maximum depth to allow
                                    // multiple splits 0=none
      int multiSplitMaxSize;        // maximum number of splits in a
                                    //    multi-split node
      Real multiSplitMinWeight;     // min weight to make an OR node
      int multiSplitWidthChange;    // amount to change maxSize at each level
      Bool multiBottomSplit;        // Create OR nodes near bottom of tree
      int multiBottomSplitDepth;    // depth at which to branch near bottom
      WeightType multiSplitWeight;  // Weight method for multi split nodes
      Bool useLossWeight;           // Reweigh instance list according to loss
				    //   matrix before training.
      Bool useCCP;		    // Should cost-complexity pruning be used
				    //   to find the optimal pruning factor?
      Bool CCPuseCV;		    // Should CCP use cross-validation? (as
				    //   opposed to hold-out)
      Bool CCPuseLoss;              // Should CCP take the loss matrix into
				    //   account?
      int CCPfolds;		    // Holds the number of folds when
				    //   performing CV for CCP
      Real CCPholdout;		    // Holds the fraction of instances to use
				    //   for holdout when performing CCP
   } sgiOptions;

   Bool ownDispatcher; // Do we own the dispatcher pointer?

   Bool haveBottomSplit; // Have we already created a bottom split?
   Real currentFitnessRatio;
   
   Bool ownMRandom;
   MRandom *randomGenerator;
   Bool ownSplitter;
   PartitionSplit *partitionSplitter;

   // Private methods
   void try_multi_thresholding(SplitAttr &sa, int attrNum,
			       DiscDispatch *dispatcher,
			       const SchemaRC &schema) const;
   AttrSubsetCat *build_subset_cat(SplitAttr& sa,
				   const MString& attrName,
				   Array<MStringRC> *& catNames) const;
   void replace_tree_with_option(DecisionTree &decisionTree,
				 const Array<int>& tieBreakingOrder,
				 Real& numSubtreeErrors,
				 Real& pessimisticSubtreeErrors,
				 int& numLeaves) const;
protected:
   // Protected member data
   DiscDispatch *dispatcherPrototype; // holds options for multithresholds

   // Protected methods
   virtual void induce_tree_from_split(DecisionTree& decisionTree,
				       NodeCategorizer& splitCat,
				       Array<MStringRC> *catNames,
				       const Array<int>& tieBreakingOrder,
				       Real& numSubtreeErrors,
				       Real& pessimisticSubtreeErrors,
				       int& numLeaves,
				       int remainingSiblings) const;
   virtual void create_option_node(DecisionTree& decisionTree,
				   NodeCategorizer& splitCat,
				   const Array<int>& tieBreakingOrder,
				   Real& numSubtreeErrors,
				   Real& pessimisticSubtreeErrors,
				   int& numLeaves,
				   int remainingSibling) const;
   virtual void prune_option_node(DecisionTree &decisionTree,
				  Real& numSubtreeErrors,
				  Real& pessimisticSubtreeErrors,
				  const Array<int>& tieBreakingOrder,
				  int& numLeaves) const;
   virtual void set_option_weights(DecisionTree &decisionTree,
				   MultiSplitCat& multiSplitCat) const;
   virtual Real pessimistic_subtree_errors(const LogOptions& logOptions,
				   DecisionTree& dt, NodePtr subtree,
				   const InstanceList& il, Real zValue,
				   const Array<int>& tieBreakingOrder) const;
   virtual void do_cost_complexity_pruning();
public:
   DynamicArray<Real>* determine_alpha_and_accuracy(
				     const LogOptions& logOptions,
				     InstanceList*& trainingList,
				     const InstanceList* testingList,
				     DynamicArray<Real>*& alphaValues,
				     PruningMethod pruningMethod,
				     Real totalTrainingSetWeight,
				     Bool useLoss,
				     LeafDistType leafDistType,
				     Real leafDistParameter,
				     DTCategorizer* dtCat=NULL);
   static Real post_prune_tree(const LogOptions& logOptions,
			       DecisionTree& dTree, NodePtr subtree,
			       Real& numSubtreeErrors, int& numSubtreeLeaves,
			       PruningMethod pruningMethod, Real pruningFactor,
			       const SchemaRC& schema,
			       Real totalTrainingSetWeight,
			       Bool useLoss,
			       LeafDistType leafDistType,
			       Real leafDistParameter);
   static Real compute_loss(const SchemaRC& schema,
			    const Array<Real>& weightDist,
			    Category majority,
			    LeafDistType leafDistType, Real leafDistParam);
private:
   static NodeCategorizer* create_leaf_categorizer_static(
				  const LogOptions& logOptions,
			          Real totalWeight,
			          int majority, Real& numSubtreeErrors,
				  const SchemaRC& schema,
				  const Array<Real>& distrArray,
				  LeafDistType leafDistType,
				  Real leafDistParameter, Bool useLoss);

public:
   virtual void OK(int /*level*/) const;
   virtual void set_user_options(const MString& prefix);
  
   SGIDTInducer(const MString& descr, CGraph* aCgraph = NULL);
   SGIDTInducer(const SGIDTInducer& source, CtorDummy);
   virtual ~SGIDTInducer();

   void verify_good_split(SplitAttr& sa, const InstanceList& TS) const;
   virtual void train();
   virtual NodePtr induce_decision_tree(CGraph& aCgraph,
					const Array<int>& tieBreakingOrder,
 					Real& numSubtreeErrors,
 					Real& pessimisticSubtreeErrors,
					int& numLeaves,
					int remainingSiblings = -1) const;
   void top_down_prune(DecisionTree& decisionTree,
		       NodePtr rootNode,
		       const Array<int>& tieBreakingOrder,
		       InstanceList *instanceList,
		       Real& numErrors, Real& pessimisticErrors,
		       int& numLeaves) const;
   virtual void prune_subtree(DecisionTree & decisionTree,
			      const Array<int>& tieBreakingOrder,
			      NodePtr largestChild, Real& numSubtreeErrors,
			      Real& pessimisticSubtreeErrors,
			      int& numLeaves) const;
   virtual NodeCategorizer* best_split(Array<MStringRC>*& catNames) const;
   virtual void best_split_info(SplitAttr*& bestSa,
				Array<SplitAttr>& splitAttrs) const;
   virtual NodeCategorizer* compile_good_splits(Array<SplitAttr>& splitAttrs,
						Real bestScore) const;
   virtual SplitAttr& choose_random_split(Array<SplitAttr>& splitAttrs,
					  Real bestScore) const;
   virtual TDDTInducer* create_subinducer(const MString& dscr,
					  CGraph& aCgraph) const;
   void split_info(int attrNum, SplitAttr& sa,
		   DiscDispatch *dispatcher,
		   PtrArray<RealAndLabelColumn*>* realColumns = NULL) const;
   virtual NodeCategorizer* split_to_cat(SplitAttr&,
					 Array<MStringRC>*& catNames) const;
   void display_options(MLCOStream& stream = Mcout,
			const DisplayPref& dp = DisplayPref::ASCIIDisplay)
	const;
   virtual Inducer* copy() const;
   virtual int class_id() const { return SGI_DT_INDUCER; }
   void set_c45_options();
   void copy_options(const SGIDTInducer& inducer);

   
   void set_multi_thresholding(Bool b)
   { sgiOptions.multiThresholding = b; }
   Bool get_multi_thresholding() const
   { return sgiOptions.multiThresholding; }
   Bool get_always_binary() const
   { return sgiOptions.alwaysBinary; }
   void set_always_binary(Bool binary)
   { sgiOptions.alwaysBinary = binary; }
   int get_disc_threshold() const
   { return sgiOptions.discThreshold; }
   void set_disc_threshold(int disc)
   { sgiOptions.discThreshold = disc; }
   int get_multi_split_max_depth() const
   { return sgiOptions.multiSplitMaxDepth; }
   void set_multi_split_max_depth(int maxDepth)
   { sgiOptions.multiSplitMaxDepth = maxDepth; }
   void set_multi_split_min_weight(Real minWeight)
   { sgiOptions.multiSplitMinWeight = minWeight; }
   Real get_multi_split_min_weight() const
   { return sgiOptions.multiSplitMinWeight; }
   int get_multi_split_max_size() const
   { return sgiOptions.multiSplitMaxSize; }
   void set_multi_split_max_size(int maxSize)
   { sgiOptions.multiSplitMaxSize = maxSize; }
   int get_multi_split_width_change() const
   { return sgiOptions.multiSplitWidthChange; }
   void set_multi_split_width_change(int change)
   { sgiOptions.multiSplitWidthChange = change; }
   WeightType get_multi_split_weight() const
   { return sgiOptions.multiSplitWeight; }
   void set_multi_split_weight(WeightType weight)
   { sgiOptions.multiSplitWeight = weight; }
   void set_multi_bottom_split(Bool split)
   { sgiOptions.multiBottomSplit = split; }
   Bool get_multi_bottom_split() const
   { return sgiOptions.multiBottomSplit; }
   void set_multi_bottom_split_depth(int depth)
   { sgiOptions.multiBottomSplitDepth = depth; }
   int get_multi_bottom_split_depth() const
   { return sgiOptions.multiBottomSplitDepth; }
   DiscDispatch *get_dispatcher();

   void set_partition_nominals(Bool b) { sgiOptions.partitionNominals = b; }
   Bool get_partition_nominals() const
   { return sgiOptions.partitionNominals; } 
   PartitionSplit *get_partition_splitter() const;
   void set_partition_splitter(PartitionSplit *splitter);
   PartitionMethodType get_partition_method() const
   { return sgiOptions.partitionMethod; }
   void set_partition_method(PartitionMethodType method)
   { sgiOptions.partitionMethod = method; }
   void set_split_decision(SplitDecisionType d)
   { sgiOptions.splitDecision = d; }
   SplitDecisionType get_split_decision() const
   { return sgiOptions.splitDecision; }
   void set_fitness_init_ratio(Real val)
   { sgiOptions.fitnessInitialRatio = val; currentFitnessRatio = val; }
   Real get_fitness_init_ratio() const
   { return sgiOptions.fitnessInitialRatio; }
   void set_fitness_ratio(Real val)
   { currentFitnessRatio = val; }
   Real get_fitness_ratio() const
   { return currentFitnessRatio; }      
   void set_fitness_mult_ratio(Real val)
   { sgiOptions.fitnessMultRatio = val; }
   Real get_fitness_mult_ratio() const { return sgiOptions.fitnessMultRatio; }
   void set_random_seed(unsigned int seed);
   unsigned int get_random_seed() const { return sgiOptions.randomSeed; }
   MRandom *get_random_generator() const;
   void set_random_generator(MRandom *generator);
   void set_use_CCP (Bool b)
      { sgiOptions.useCCP = b; }
   Bool get_use_CCP () const
      { return sgiOptions.useCCP; }
   void set_CCP_use_loss(Bool b) { sgiOptions.CCPuseLoss = b; }
   Bool get_CCP_use_loss() const { return sgiOptions.CCPuseLoss; }
   void set_CCP_use_CV (Bool b)
      { sgiOptions.CCPuseCV = b; }
   Bool get_CCP_use_CV () const
      { return sgiOptions.CCPuseCV; }
   void set_CCP_folds (int i)
      { sgiOptions.CCPfolds = i; }
   int get_CCP_folds () const
      { return sgiOptions.CCPfolds; }
   void set_CCP_holdout (Real r)
      { sgiOptions.CCPholdout = r; }
   Real get_CCP_holdout () const
      { return sgiOptions.CCPholdout; }
   void set_use_loss_weight(Bool b) { sgiOptions.useLossWeight = b; }
   Bool get_use_loss_weight() const { return sgiOptions.useLossWeight; }
};

#endif
