// MLC++ - Machine Learning Library in -*- C++ -*-
// See Descrip.txt for terms and conditions relating to use and distribution.

// This is an include file.  For a full description of the class and
// functions, see the file name with a "c" suffix.

#ifndef _Schema_h
#define _Schema_h 1

#include <Attribute.h>
#include <MStringRC.h>
#include <Array2.h>

class BoolArray;
class InstanceOp;

// This class encapsulates the pointer for doubly-linked-list.
class AttrInfoPtr {
public:
   AttrInfoPtr(AttrInfo *ai) : attrInfo(ai) {}
   AttrInfo *attrInfo;
   Bool operator==(const AttrInfoPtr& aip) const
      { return (*aip.attrInfo) == (*attrInfo);}
};


// Instance information
class Schema {
   friend class SchemaRC;
   
   // The schema owns all of the AttrInfo.
   PtrArray<AttrInfo*> attr;
   AttrInfo* labelInfo;
   int refCount;
   PTHR_DECL(pthread_t repOwner;)
   
   // The schema optionally contains an operation to be performed on
   // instances to get them to this state from some original schema
   InstanceOp *instanceOp;

   // Loss function matrix for this schema
   Array2<Real> *lossMatrix;
   
   static Bool transposeLossMatrix;

   NO_COPY_CTOR2(Schema, attr(0));

protected:
   virtual void establish_loss_matrix();
   
public:
   virtual void OK(int level = 0) const;
   // note that there is no default constructor.
   Schema(DblLinkList<AttrInfoPtr>*& attrInfos);
   Schema(DblLinkList<AttrInfoPtr>*& attrInfos, AttrInfo*& lInfo);
   // copy constructor
   Schema(const Schema& source, CtorDummy);
   virtual ~Schema();

   // Not virtual for performance.  See attr_info below.
   const AttrInfo& label_info() const;
   virtual void set_label_info(AttrInfo *& ai);
   Bool is_labelled(Bool fatalOnFalse=FALSE) const;
   int num_attr() const { return attr.size(); }
   // Not virtual for performance reasons.  IBCategorizer::distance
   //   calls this a lot.
   const AttrInfo& attr_info(int num) const { return *attr[num]; }
   // checks whether an AttrInfo is contained in Schema.
   virtual Bool member(const AttrInfo& ainfo) const;

   // equal() compares the labelInfos for the schemas if they exist.
   // Returns FALSE if one has a labelInfo and the other does not.
   virtual Bool equal(const Schema&, Bool fatalOnFalse = FALSE) const;
   virtual Bool compatible_with(const Schema&, Bool fatalOnFalse = FALSE) const;

   // equal_no_label() does not use the labelInfo for either Schema whether
   // or not they exist.
   virtual Bool equal_no_label(const Schema&, Bool fatalOnFalse = FALSE) const;
   virtual Bool compatible_with_no_label(const Schema&,
					 Bool fatalOnFalse = FALSE) const;
   virtual Bool operator==(const Schema& schema) const {
      return equal(schema, FALSE);
   }
   virtual Bool operator!=(const Schema& schema) const {
      return !equal(schema, FALSE);
   }
   virtual void equal_except_del_attr(const Schema& schema,
				      int deletedAttrNum) const;
   virtual const MString& attr_name(int attrNum) const;
   virtual const NominalAttrInfo& nominal_attr_info(int attrNum) const;
   virtual int add_value(int attrNum, const MString& attrVal);
   virtual int num_attr_values(int attrNum) const;
   virtual int num_label_values() const;
   virtual const MString& nominal_to_string(int attrNum,
					    Category cat) const;
   virtual const MString& category_to_label_string(Category cat) const;
   virtual const NominalAttrInfo& nominal_label_info() const;
   virtual Schema* remove_attr(int attrNum) const;
   virtual void display(MLCOStream& stream = Mcout,
			Bool protectChars = TRUE) const;
   virtual void display_names(MLCOStream& stream = Mcout,
			      Bool protectChars = TRUE,
                              const MString& header = EMPTY_STRING) const;
   virtual Schema* project(const BoolArray& attrMask) const;

   virtual const InstanceOp *get_operation(Bool fatalOnNone = TRUE) const;
   virtual void set_operation(InstanceOp *&op);
   virtual void clear_operation();

   virtual Bool get_ignore_mask(BoolArray& mask, int maxNumValues,
				Bool warn) const;
   virtual BoolArray *make_projection_to(const Schema *other) const;
   virtual Array<int>* sort_labels() const;

   virtual int attr_num_from_name(const MString& name,
				  Bool fatalOnNotFound = TRUE) const;
   virtual void set_label_name(const MString& newName);

   virtual Schema *clone_without_label() const;
   virtual Bool is_nominal() const;
   virtual Array<MStringRC> *find_duplicate_attrs() const;
   virtual Bool check_duplicate_attrs(Bool fatalOnTrue = TRUE) const;

   // Loss matrix functions
   virtual const Array2<Real>& get_loss_matrix() const;
   virtual Bool has_loss_matrix() const { return lossMatrix != NULL; }
   virtual void remove_loss_matrix();
   virtual void set_loss_matrix(const Array2<Real>& lossMatrix);
   virtual void set_default_loss_matrix(Real correctLoss, Real incorrectLoss);
   virtual void set_distance_loss_matrix(Real correctLoss, Real distanceLoss,
					 Real unknownLoss);
   virtual void set_undefined_loss_matrix();
   virtual void set_loss_entry(AttrValue_ predicted, AttrValue_ actual,
			       Real loss);
   virtual void verify_loss_matrix() const;

   // The if transpose is set, further setting of the loss matrix will cause
   // it to be set in transpose. This only works for two-class problems.
   static void set_transpose_loss_matrix(Bool t)
   { transposeLossMatrix = t; }
   static Bool get_transpose_loss_matrix()
   { return transposeLossMatrix; }
};

DECLARE_DISPLAY(Schema)

#endif


