// MLC++ - Machine Learning Library in -*- C++ -*-
// See Descrip.txt for terms and conditions relating to use and distribution.
/***************************************************************************
  Description  : The SearchDispatch class stores options common to many
                   different search methods and allows easy selection and
		   setting of these options for building a search.
  Comments     : See SSSearch, HCSearch, BFSearch, SASearch for details
                   on the specific search methods.
  Complexity   : 
  Enhancements :
  History      : Dan Sommerfield                                    10/30/96
                   Initial revision (.h,.c)
***************************************************************************/

#include <basics.h>
#include <SearchDispatch.h>
#include <GetOption.h>

RCSID("MLC++, $RCSfile: SearchDispatch.c,v $ $Revision: 1.6 $")

/***************************************************************************
  Description  : Sets all options to their default values.
  Comments     : 
***************************************************************************/
template<class LocalInfo, class GlobalInfo>
void SearchDispatch<LocalInfo, GlobalInfo>::set_defaults()
{
   searchMethod = DEFAULT_SEARCH_METHOD;
   showTestSetPerf = DEFAULT_SHOW_TEST_SET_PERF;
   evalLimit = DEFAULT_EVAL_LIMIT;
   maxStale = DEFAULT_MAX_STALE;
   epsilon = DEFAULT_EPSILON;
   maxEvals = DEFAULT_MAX_EVALS;
   minExpEvals = DEFAULT_MIN_EXP_EVALS;
   lambda = DEFAULT_LAMBDA;
   sasSeed = DEFAULT_SAS_SEED;
}


/***************************************************************************
  Description  : Sets options based on user input/external options stored
                   in the option server.  Only sets options appropriate
		   for the selected search method.
  Comments     : 
***************************************************************************/
template<class LocalInfo, class GlobalInfo>
void SearchDispatch<LocalInfo, GlobalInfo>
::set_user_options(const MString& prefix)
{
   searchMethod = get_option_enum(prefix + "SEARCH_METHOD",
				  searchMethodEnum,
				  searchMethod,
				  SEARCH_METHOD_HELP, FALSE);
   showTestSetPerf = get_option_enum(prefix + "SHOW_TEST_SET_PERF",
				      showTestSetPerfEnum,
				      showTestSetPerf,
				      SHOW_TEST_SET_PERF_HELP, TRUE);
   evalLimit = get_option_int(prefix + "EVAL_LIMIT",
			      evalLimit,
			      EVAL_LIMIT_HELP, TRUE);

   if(searchMethod == bestFirst ||
      searchMethod == simulatedAnnealing) {
      maxStale = get_option_int(prefix + "MAX_STALE", maxStale,
				MAX_STALE_HELP, FALSE);
      epsilon = get_option_real(prefix + "EPSILON", epsilon,
				EPSILON_HELP, FALSE);
   }

   if(searchMethod == simulatedAnnealing) {
      maxEvals = get_option_int(prefix + "MAX_EVALS", maxEvals,
				MAX_EVALS_HELP, TRUE);
      minExpEvals = get_option_int(prefix + "MIN_EXP_EVALS", minExpEvals,
				   MIN_EXP_EVALS_HELP, TRUE);
      lambda = get_option_real(prefix + "LAMBDA", lambda,
			       LAMBDA_HELP, FALSE);
      sasSeed = get_option_int(prefix + "SAS_SEED", sasSeed,
			       SAS_SEED_HELP, TRUE);
   }
}
				

/***************************************************************************
  Description  : Builds an appropriate subclass of SSSearch based on the
                   options set within this class
  Comments     : 
***************************************************************************/
template<class LocalInfo, class GlobalInfo>
StateSpaceSearch<LocalInfo, GlobalInfo> *
SearchDispatch<LocalInfo, GlobalInfo>::build_search() const
{
   StateSpaceSearch<LocalInfo, GlobalInfo> *search;

   // set search-specific options
   switch(searchMethod) {
      case hillClimbing:
	 search = new HCSearch<LocalInfo, GlobalInfo>;
	 break;

      case bestFirst:
      {
	 search = new BFSearch<LocalInfo, GlobalInfo>;
	 BFSearch<LocalInfo, GlobalInfo> *bfSearch =
	    (BFSearch<LocalInfo, GlobalInfo> *)search;
	 bfSearch->set_max_non_improving_expansions(maxStale);
	 bfSearch->set_min_expansion_improvement(epsilon);
	 break;
      }

      case simulatedAnnealing:
      {
	 search = new SASearch<LocalInfo, GlobalInfo>;
	 SASearch<LocalInfo, GlobalInfo> *saSearch =
	    (SASearch<LocalInfo, GlobalInfo> *)search;
	 saSearch->set_max_non_improving_expansions(maxStale);
	 saSearch->set_min_expansion_improvement(epsilon);
	 saSearch->set_max_evaluations(maxEvals);
	 saSearch->set_min_exp_evaluations(minExpEvals);
	 saSearch->set_initial_lambda(lambda);
	 saSearch->set_sas_seed(sasSeed);
	 break;
      }

      default:
	 err << "SearchDispatch: unsupported type of search" << fatal_error;
	 
   }

   // set general search options
   search->set_show_test_set_perf(showTestSetPerf);
   search->set_eval_limit(evalLimit);
   
   return search;
}


/***************************************************************************
  Description  : Options setting functions; these functions set values for
                   the various available options, validating the input as
		   necessary.
  Comments     : 
***************************************************************************/
template <class LocalInfo, class GlobalInfo>
void SearchDispatch<LocalInfo, GlobalInfo>::set_eval_limit(int limit)
{
   if(limit < 0)
      err << "SearchDispatch::set_eval_limit: EVAL_LIMIT must be >= 0"
	 << fatal_error;
   evalLimit = limit;
}

template <class LocalInfo, class GlobalInfo>
void SearchDispatch<LocalInfo, GlobalInfo>::set_max_stale(int newMaxStale)
{
   if(newMaxStale < 1)
      err << "SearchDispatch::set_max_stale: MAX_STALE must be positive"
	 << fatal_error;
   maxStale = newMaxStale;
}

template <class LocalInfo, class GlobalInfo>
void SearchDispatch<LocalInfo, GlobalInfo>::set_epsilon(Real newEpsilon)
{
   epsilon = newEpsilon;
}

template <class LocalInfo, class GlobalInfo>
void SearchDispatch<LocalInfo, GlobalInfo>::set_max_evals(int evals)
{
   if(evals < 0)
      err << "SearchDispatch::set_max_evals: MAX_EVALS must be >= 0"
	 << fatal_error;
   maxEvals = evals;
}

template <class LocalInfo, class GlobalInfo>
void SearchDispatch<LocalInfo, GlobalInfo>::set_min_exp_evals(int minExp)
{
   if(minExp < 0)
      err << "SearchDispatch::set_min_exp_evals: MIN_EXP_EVALS must be >= 0"
	 << fatal_error;
   minExpEvals = minExp;
}

template <class LocalInfo, class GlobalInfo>
void SearchDispatch<LocalInfo, GlobalInfo>::set_lambda(Real newLambda)
{
   if(newLambda < 0)
      err << "SearchDispatch::set_lambda: LAMBDA must be >= 0"
	 << fatal_error;
   lambda = newLambda;
}

template <class LocalInfo, class GlobalInfo>
void SearchDispatch<LocalInfo, GlobalInfo>::set_sas_seed(int newSeed)
{
   sasSeed = newSeed;
}


