// MLC++ - Machine Learning Library in -*- C++ -*-
// See Descrip.txt for terms and conditions relating to use and distribution.

// This is an include file.  For a full description of the class and
// functions, see the file name with a "c" suffix.

#ifndef _SplitAttr_h
#define _SplitAttr_h 1

#include <MEnum.h>
#include <Array.h>
#include <SplitScore.h>
#include <InstanceList.h>

class SplitAttr : public SplitScore {
public:
enum SplitType {noReasonableSplit, realThresholdSplit,
		 multiRealThresholdSplit, nominalSplit, partitionSplit};
static const MEnum splitTypeEnum;
private:
   NO_COPY_CTOR(SplitAttr);

   // Member data
   int attributeNum;
   SplitType splitType;
   Bool existSplit;
   Bool penalizeByMDL;
   Real mdlPenalty;
   int realValuedSplitIndex;
   union {
      Array<NominalVal> *partition; // used in partitionSplit
      Real theThreshold;
      Array<Real>* thresholds;
      // any other info we need
   } typeInfo;

public:
   SplitAttr();
   SplitAttr(const SplitAttr& source, CtorDummy);
   virtual ~SplitAttr();
   void reset();
   virtual void free_type_info();
   virtual int class_id() const {return CLASS_SPLIT_ATTR; }
   void initialize(const InstanceListArray& instLists, int attributeNumber);
   SplitAttr& operator= (const SplitAttr& source);
   virtual Bool operator== (const SplitScore& rhs) const;
   
   void set_penalize_by_mdl(Bool choice) {penalizeByMDL = choice;}
   void set_split_and_label_dist(Array2<Real>*& sAndLDist)
      {SplitScore::set_split_and_label_dist(sAndLDist);}
   void set_split_and_label_dist(const InstanceList& instList);
   void set_attr_num(int num);
   
   virtual Bool make_nominal_split(const InstanceList& instList,
				   int attributeNumber);
   virtual Bool make_nominal_split();
   virtual Bool make_real_split(RealAndLabelColumn& column, int attrNum,
				Real minSplit, int smoothInst=0,
				Real smoothFactor=0.5);
   
   virtual void save_real_split(Real thresh, int idx, int numDistinctValues);
   
   SplitType split_type() const { return splitType; }
   Bool get_penalize_by_mdl() const {return penalizeByMDL;}
   Real penalty() const;
   Real get_mutual_info(Bool normalize, Bool penalize) const;
   Real get_gain_ratio(Bool penalize) const;
   virtual Real score() const;
   virtual Real score(const Array2<Real>* sAndLDist, const Array<Real>* sDist,
		      const Array<Real>* lDist, Real entropy = UNDEFINED_REAL,
		      Real totalWeight = UNDEFINED_REAL)
      {return SplitScore::score(sAndLDist, sDist, lDist, entropy,
				totalWeight);}
   Bool exist_split() const {return existSplit;}
   Real threshold() const;
   // The following is only used by EntropyDisc
   int real_valued_split_index() const
      {return realValuedSplitIndex;}
   int get_attr_num() const {ASSERT(attributeNum != -1); return attributeNum; }
   virtual void display(MLCOStream& stream = Mcout,
			const DisplayPref& dp =
			DisplayPref::ASCIIDisplay) const;
   virtual void set_thresholds(Array<Real>*& vals);
   virtual void set_nominal_partition(Array<NominalVal>*& partition);
   virtual const Array<Real>& get_thresholds() const;
   virtual const Array<NominalVal>& get_nominal_partition() const;
};

DECLARE_DISPLAY(SplitAttr)

class SplitAttrPtr {
public:
   // Public member data
   SplitAttr *splitAttr;

   // Methods
   SplitAttrPtr() : splitAttr(NULL) { }
   SplitAttrPtr(SplitAttr *ptr) : splitAttr(ptr) { }
   ~SplitAttrPtr() { }

   // This operator will sort from the BEST score to the WORST score (that's
   // the way we almost always want it), so operator< actually compares using
   // the < on the scores.
   // We could do away with this confusion with a sort_descending function on
   // Array, which requires an operator>, for example.
   // To stabilize sort operator < looks at score first then attr number 
   
   Bool operator<(const SplitAttrPtr& rhs) const {
      return ((splitAttr->score() > rhs.splitAttr->score()) ||
	      ((splitAttr->score() == rhs.splitAttr->score()) &&
	       (splitAttr->get_attr_num() < rhs.splitAttr->get_attr_num())));}
	
   Bool operator==(const SplitAttrPtr& rhs) const {
      return ((splitAttr->score() == rhs.splitAttr->score()) &&
	      (splitAttr->get_attr_num() == rhs.splitAttr->get_attr_num()));}
   Bool operator!=(const SplitAttrPtr& rhs) const
      {return ! ((*this)== rhs);}
};   

Bool ok_to_split(int attrNum, const BagCounters& counters, Real minSplit);   

#endif   
