// MLC++ - Machine Learning Library in -*- C++ -*-
// See Descrip.txt for terms and conditions relating to use and distribution.

// This is an include file.  For a full description of the class and
// functions, see the file name with a "c" suffix.

#ifndef _SplitScore_h
#define _SplitScore_h 1
#include <MEnum.h>
#include <DisplayPref.h>
#include <Array2.h>
#include <LogOptions.h>
#include <class_id.h>

#define CLASS_SPLIT_SCORE      SPLIT_SCORE_ID_BASE + 1
#define CLASS_SPLIT_ATTR       SPLIT_SCORE_ID_BASE + 2
#define CLASS_MULTI_SPLIT_ATTR SPLIT_SCORE_ID_BASE + 3

// The majority of this class's data is lazily evaluated.  What that means
//   here is that if an object is copied, most of the data is not copied
//   to the recipient, but flags are utilized to identify invalid data.
// The practical consequence of this is threefold:
//   copies are somewhat faster,
//   == is not bitwise, it only compares valid data, and
//   const get_xxx() methods are only logically const.
// Upon copying, only that data that can not be rebuilt from the
//   splitScoreCriterion array is copied.
// The cache is defined as all other data items (except validCache).
// Upon copying, the validCache flag of the recipient is set FALSE.
// To access an item from the cache, valid_cache() *MUST* first be called.
//   This sets the various cache items to invalid values, if validCache is
//   FALSE, and then resets validCache.  Then the desired item may be checked
//   for its individual invalid value, and, if invalid, it may then be
//   recalculated.  If it depends on other items in the cache, they must be
//   recalculated first.
// The individual invalid values come in two flavors:
//   UNDEFINED_INT is used for theEntropy, theCondEntropy, and theSplitEntropy;
//   and UNDEFINED_REAL is used for theMutualInfo and theGainRatio.
// Note that these values must also be used for initialization.

class SplitScore {
   LOG_OPTIONS;
   NO_COPY_CTOR(SplitScore);
public:
// mutualInfo is entropy - cond_entropy, sometimes called information gain.
// normalizedMutualInfo divides the mutualInfo by the number of splits
//   values to make sure it's between 0 and 1 and to bias against
//   multi-way splits
// gainRatio is Quinlan's gain ratio.
// mutualInfoRatio is mutualInfo divided by entropy.  It is again bounded
//   between 0 and 1.
// externalScore indicates that the value returned by invocation of score()
//   will have been set from outside the class.  Any bounds checking for
//   an externalScore must be done from outside this class.

   enum SplitScoreCriterion {mutualInfo, normalizedMutualInfo, gainRatio,
			     mutualInfoRatio, externalScore};

   // As copies are shallow, don't need a destructor.
   // Need to make public because we occasionally get a warning in
   // SplitScore.c about access errors if its not public.
   struct CacheStruct {
      // Public member data (of CacheStruct)
      Real totalWeight;
      Real entropy;
      Real mutualInfo;
      Real condEntropy;
      Real gainRatio;
      Real splitEntropy;

      Array<Real>* splitDist;
      Array<Real>* labelDist;

      // Methods (of CacheStruct)
      CacheStruct();
      void copy(const CacheStruct& source); // shallow copy
      CacheStruct(const CacheStruct& source) // shallow copy
         {copy(source);}
      CacheStruct& operator= (const CacheStruct& source) // shallow copy
         {copy(source); return *this;}
   };

   // Public data (also see private data)
   static const SplitScoreCriterion defaultSplitScoreCriterion;

   static const MEnum splitScoreCriterionEnum; 

   // The following is used in TDDTInducer.
   static const MString splitScoreCriterionHelp;

private:
   // Member data (also see public data)
   SplitScoreCriterion splitScoreCriterion;
   Bool validCache;
   CacheStruct cache;
   Array2<Real>* splitAndLabelDist;
   Real theExternalScore;

   // Private methods
   void copy_dist(const SplitScore& source);
   // No copy_split_dist() or copy_label_dist() as they are lazily evaluated
   void copy_split_and_label_dist(const SplitScore& source);
   Bool split_and_label_dist_is_equal(const Array2<Real>* source) const;
   // The following method is logically const, but may cause the
   //   cache to be updated.
   Bool valid_cache() const;
public:
   virtual void OK(int level = 0) const;
   SplitScore();
   SplitScore(const SplitScore& source, CtorDummy);
   virtual ~SplitScore();
   void reset();
   virtual int class_id() const {return CLASS_SPLIT_SCORE; }
   SplitScore& operator= (const SplitScore& rhs);
   virtual Bool operator== (const SplitScore& rhs) const;
   Bool operator != (const SplitScore& rhs) const;
   // Do you have the splitAndLabelDistribution
   Bool has_distribution(Bool fatalOnFalse = TRUE) const;
   virtual void set_split_score_criterion(SplitScoreCriterion choice)
      {splitScoreCriterion = choice; }
   void set_external_score(Real extScore);
   // There is a crucial difference in side-effects among the
   //   following setXXdist() functions.  The set_split_and_label_dist
   //   function frees the cache.  The set_label_dist and set_split_dist
   //   functions do not.  Freeing the cache, among other things, empties
   //   the splitDist and labelDist arrays.  The reasoning is this:
   //   If the splitAndLabelDist needs to be changed, then the
   //   cache can no longer be valid.  If one of the derived
   //   arrays (splitDist or labelDist) is being changed, then the
   //   user is assumed to be setting it to correspond to an
   //   already-correct splitAndLabelDist array.  So don't use
   //   set_split_and_label_dist() after using either
   //   set_label_dist() or set_split_dist() unless you want to
   //   lose the array(s) set in the earlier usage(s).

   // Of the next 3 functions (set_XX_dist()), the final two are optional.
   //   Since both labelDist and splitDist are derived from
   //   splitAndLabelDist, setting splitAndLabelDist provides all
   //   necessary information.  If the caller happens to have either or
   //   both of the derived arrays, they may be provided to SplitScore
   //   so as to save SplitScore the effort of deriving them.
   // The following gets ownership of sAndLDist.
   void set_split_and_label_dist(Array2<Real>*& sAndLDist);
   // The following gets ownership of lDist.
   void set_label_dist(Array<Real>*& lDist);
   // The following gets ownership of sDist.
   void set_split_dist(Array<Real>*& sDist);
   // The following methods are logically const, but may cause the
   //   cache to be updated.
   Real total_weight() const;
   int num_splits() const;
   Real normalize_by_num_splits(Real orgScore) const;
   Real get_mutual_info(Bool normalize) const;
   Real get_gain_ratio() const;
   Real get_mutual_info_ratio() const;
   Real get_cond_entropy() const;
   Real get_split_entropy() const;
   Real get_entropy() const;
   virtual Real score() const;
   virtual Real score(const Array2<Real>* sAndLDist,
		      const Array<Real>* sDist,
		      const Array<Real>* lDist,
		      Real entropy = UNDEFINED_INT, //@@ UNDEFINED_REAL??
		      Real totalWeight = UNDEFINED_REAL);
   const Array2<Real>& get_split_and_label_dist() const;
   const Array<Real>& get_label_dist() const;
   const Array<Real>& get_split_dist() const;
   // End of public logically const methods.
   SplitScoreCriterion get_split_score_criterion() const
      {return splitScoreCriterion;}
   Real has_external_score() const {return theExternalScore != UNDEFINED_REAL;}
   Real get_external_score() const;
   Array2<Real>* release_split_and_label_dist();
   Array<Real>* release_label_dist();
   Array<Real>* release_split_dist();

   virtual void display(MLCOStream& stream = Mcout,
			const DisplayPref& dp = defaultDisplayPref) const;
};

DECLARE_DISPLAY(SplitScore)

#endif   
