// MLC++ - Machine Learning Library in -*- C++ -*-
// See Descrip.txt for terms and conditions relating to use and distribution.
/***************************************************************************
  Description  : StringMap implements a general mapping/lookup class for
                   strings.  The strings may be mapped to any data type
		   Elem.
		 The mapping is accomplished using both a hash table and an
		   array.  The hash table is provided for speedy searches,
		   and the array is provided for iteration.
  Assumptions  : 
  Comments     : 
  Complexity   : 
  Enhancements : 
  History      : Dan Sommerfield                                     2/01/96
                   Initial revision (.h,.c)
***************************************************************************/

#include <basics.h>
#include <string.h>
#include <stdlib.h>
#include <StringMap.h>


/******************************************************************************
  Description  : Display function for the StringHashData key-value pair
                   class.  Note that this function requires Elem to
		   support a display operation.
  Comments     : Display is required by UniversalHashTable.
******************************************************************************/
template <class Elem>
void StringHashData<Elem>::display(MLCOStream& stream) const
{
   stream << key << ": " << value;
}

template <class Elem>
DEF_DISPLAY(StringHashData<Elem>)


/******************************************************************************
  Description  : Integrity constraints for the table.  Makes sure numbers
                   of items in the table and list match, and in slow mode
		   will check each item in the list to make sure its in the
		   table.
  Comments     : 
******************************************************************************/
template <class Elem>
void StringMap<Elem>::OK() const
{
   if(table.num_items() != names.size())
      err << "StringMap::OK: number of items in hash table (" <<
	 table.num_items() << ") does not match number of items in "
	 "list (" << names.size() << ")" << fatal_error;

   // insure that all items in list are actually in the table
   DBGSLOW(
      for(int i=0; i<names.size(); i++)
         (void)get(names.index(i));  // will abort if not found
	 );
}

/******************************************************************************
  Description  : Constructor.  The default constructor allows an optional
                   parameter for an estimated size of the hash table
  Comments     : 
******************************************************************************/
template <class Elem>
StringMap<Elem>::StringMap(int sizeHint)
   : table(sizeHint),
     names(0)
{
   if(sizeHint < 256)
      err << "StringMap::StringMap: provided size hint of " << sizeHint
	 << " is too small.  It must be at least 256." << fatal_error;
}


/******************************************************************************
  Description  : Adds a mapping to the table.  Aborts if the mapping
                   already exists.
  Comments     : 
******************************************************************************/
template <class Elem>
void StringMap<Elem>::add(const MString& name, const Elem& value)
{
   StringHashData<Elem> *keyVal = table.find(name);
   if(keyVal)
      err << "StringMap::add: \"" << name << "\" is already in the "
	 "table" << fatal_error;
   names.index(names.size()) = name;
   table.insert(StringHashData<Elem>(name, value));
   OK();
}

/******************************************************************************
  Description  : Sets a mapping in the hash table, aborting if the mapping
                   is missing.
  Comments     : 
******************************************************************************/
template <class Elem>
void StringMap<Elem>::set(const MString& name, const Elem& value)
{
   // make sure we already set the value
   StringHashData<Elem> *keyVal = table.find(name);
   if(keyVal == NULL)
      err << "StringMap::set: attempting to set a mapping for the "
	 "name \"" << name << "\" which is not in the table" <<
	 fatal_error;

   keyVal->set_value(value);
   OK();
}

/******************************************************************************
  Description  : Sets the mapping for an item, adding the mapping if the
                   item is not currently in the table.
  Comments     : It is safer to use add or set if you know in advance
                   whether or not the item will be in the table.
******************************************************************************/
template <class Elem>
void StringMap<Elem>::map(const MString& name, const Elem& value)
{
   StringHashData<Elem> *keyVal = table.find(name);
   if(keyVal)
      keyVal->set_value(value);
   else {
      names.index(names.size()) = name;
      table.insert(StringHashData<Elem>(name, value));
   }
   OK();
}


/******************************************************************************
  Description  : Retrieves a mapping given the name.  This function aborts
                   if the mapping is missing from the table.
  Comments     : 
******************************************************************************/
template <class Elem>
const Elem& StringMap<Elem>::get(const MString& name) const
{
   // lookup the mapping
   const StringHashData<Elem> *keyVal = table.find(name);
   if(keyVal == NULL)
      err << "StringMap::get: the name \"" << name << "\" was not found "
	 "in the table" << fatal_error;
   return keyVal->get_value();
}


/******************************************************************************
  Description  : Retrieves a mapping given the name and a reference in which
                   to store the value.  Returns TRUE if the mapping exists,
		   false if it does not.  No value is placed in the value
		   reference if the mapping does not exist.
  Comments     : 
******************************************************************************/
template <class Elem>
Bool StringMap<Elem>::get(const MString& name, Elem& value) const
{
   // lookup the mapping
   const StringHashData<Elem> *keyVal = table.find(name);
   if(keyVal == NULL)
      return FALSE;

   // get_value() returns a reference, so we need to force the copy
   // constructor here so we may set the value reference we're given.
   Elem aval = keyVal->get_value();
   value = aval;
   return TRUE;
}

/******************************************************************************
  Description  : Displays the table, in the order given by the list.
  Comments     : 
******************************************************************************/
template <class Elem>
void StringMap<Elem>::display(MLCOStream& stream) const
{
   OK();
   if(names.size() == 0)
      stream << "Empty table" << endl;
   else {
      for(int i=0; i<names.size(); i++) {
	 stream << names.index(i) << ": " << get(names.index(i)) << endl;
      }
   }
}

template <class Elem>
DEF_DISPLAY(StringMap<Elem>)










