// MLC++ - Machine Learning Library in -*- C++ -*-
// See Descrip.txt for terms and conditions relating to use and distribution.

// This is an include file.  For a full description of the class and
// functions, see the file name with a "c" suffix.

#ifndef _TDDTInducer_h
#define _TDDTInducer_h 1

#include <Inducer.h>
#include <SplitScore.h>
#include <CGraph.h>

class DTCategorizer;
class DecisionTree;
class CatGraph;
class LeafCategorizer;

class TDDTInducer : public Inducer {
   NO_DEFAULT_OPS(TDDTInducer);
public:
   enum LeafDistType { allOrNothing, frequencyCounts,
                       laplaceCorrection, evidenceProjection };

   enum PruningMethod { none, confidence, penalty, linear, KLdistance,
                        lossConfidence, lossLaplace };
   static MEnum pruningMethodEnum;
   enum EvalMetric { error, MSE, logLoss };
   static MEnum evalMetricEnum;
   
private:

   // Member data
   // options
   struct TDDTOptions{
      // Public member data (for TDDTOptions)
      int maxLevel;                 // max level to grow.
      Real lowerBoundMinSplitWeight;// The lower bound for the minimum weight
				    //   of instances in a node.
      Real upperBoundMinSplitWeight;// The upper bound for the same.
      Real minSplitWeightPercent;   // The percent (p) used to calculate the
				    // min weight of instances in a node (m).
				    // m = p * num instances / num categories
      Bool nominalLBoundOnly;       // The values for the above three options
                                    //   are always used with real-valued
                                    //   attributes to determine if it's ok
                                    //   to split; for nominal-valued
                                    //   attributes they will only be used
                                    //   for setting min. instances in a node
                                    //   when this option is FALSE.
      Bool debug;
      Bool unknownEdges;            // Have an edge with "unknown" from every
				    // node 
      SplitScore::SplitScoreCriterion splitScoreCriterion;
      Bool emptyNodeParentDist;     // Should an empty node have the parent's
                                    //   distribution?
      Bool parentTieBreaking;       // Should a node inherit the parent's
                                    //   tie-breaking class?
      PruningMethod pruningMethod;  // How should the tree be pruned?  If not
                                    //   NONE and pruning_factor is 0, then a
                                    //   node will be made a leaf when its
                                    // (potential) children do not improve
                                    // the error count.
      Bool pruningBranchReplacement;
                                    // Should pruning allow replacing a node
                                    //   with its largest subtree?
      Bool adjustThresholds;        // Should threshold be adjusted to
                                    // equal instance values?
      Real pruningFactor;           // How much to prune.  High->more pruning.
      Bool contMDLAdjust;           // Should the Minimum Description Length
                                    //   Adjustment for continuous attributes
                                    //   be applied to mutual info?
      int smoothInst;               // Number of thresholds on either side
                                    // to use for smoothing; 0 for no smoothing
      Real smoothFactor;            // Exponential factor for smoothing

      LeafDistType leafDistType;    // Type of distribution to build at
                                    // leaves.
      Real MEstimateFactor;         // M-estimate factor for laplace
      Real evidenceFactor;          // Evidence correction factor

      EvalMetric evaluationMetric;  // how the induced tree will be measured
   } tddtOptions;

   CGraph *cgraph; // Graph to build this in, or NULL if to create one.
   DTCategorizer *decisionTreeCat;
   int level;

   Real totalInstWeight; // total instance weight for decision tree (during
                         //   recursive calls this doesn't change).

   Bool haveContinuousAttributes;    // Should we transpose at each node?

protected:
   // Protected methods
   static Real num_cat_errors(const Categorizer& cat,
			      const Category predictClass,
			      Real totalWeight);
   
   // functions called within induce_decision_tree
   virtual void induce_tree_from_split(DecisionTree& decisionTree,
				       NodeCategorizer& splitCat,
				       Array<MStringRC> *catNames,
				       const Array<int>& tieBreakingOrder,
				       Real& numSubtreeErrors,
				       Real& pessimisticSubtreeErrors,
				       int& numLeaves,
				       int remainingSiblings) const;
   virtual void set_total_inst_weight(Real wt) { totalInstWeight = wt; }
   virtual Real get_total_inst_weight() const { return totalInstWeight; }
   void connect(CatGraph& catGraph, NodePtr from, NodePtr to,
		Category edgeVal, const MStringRC& edgeName) const;
   virtual Real pessimistic_subtree_errors(const LogOptions& logOptions,
				   DecisionTree& dt, NodePtr subtree,
				   const InstanceList& il, Real zValue,
				   const Array<int>& tieBreakingOrder) const;
   virtual Real pessimistic_subtree_loss(const LogOptions& logOptions,
					 DecisionTree& dt, NodePtr subtree,
					 const InstanceList& il, Real zValue,
					 const Array<int>& tieBreakingOrder)
      const;
   Real laplace_subtree_loss(const LogOptions& logOptions,
			     DecisionTree& dt, NodePtr subtree,
			     const InstanceList& il,
			     const Array<int>& tieBreakingOrder) const;

   void build_categorizer(DecisionTree*& dTree);
   void accumulate_tree_stats() const;
public:

   virtual void OK(int level = 0) const;
   // Given a cgraph, the tree will be formed there.
   TDDTInducer(const MString& dscr, CGraph* aCgraph = NULL);
   TDDTInducer(const TDDTInducer& source, CtorDummy, CGraph* aCgraph = NULL);
   virtual ~TDDTInducer();

   virtual int train_no_stats();        // build structure
   virtual void train();                // build structure, record size
   virtual Bool was_trained(Bool fatalOnFalse = TRUE) const;
   virtual const Categorizer& get_categorizer() const;
   virtual Categorizer *release_categorizer();
   virtual const DTCategorizer& get_dt_categorizer() const;
   virtual DTCategorizer& get_dt_categorizer();
   
   // This creates a name for the sub-inducer.  
   virtual MString name_sub_inducer(MString catDescr, Category catNum) const;
   // create_leaf_cat creates a leaf categorizer when splitting is done.
   // If the distribution is not given, we compute it from the instance list
   virtual LeafCategorizer* create_leaf_categorizer(Real totalWeight,
	  const Array<int>& tieBreakingOrder, Real& numErrors,
	  Real& pessimisticErrors, const Array<Real>* distrArray = NULL) const;
   virtual void prune_subtree(DecisionTree & decisionTree,
			      const Array<int>& tieBreakingOrder,
			      NodePtr largestChild, Real& numSubtreeErrors,
			      Real& pessimisticSubtreeErrors,
			      int& numLeaves) const;
   virtual Real calc_KL_distance (DecisionTree& decisionTree) const;
   static Real kullback_leibler_distance(const Array<Real>& p,
   					 const Array<Real>& q);

   // Actual Inducer.  It builds the tree in CGraph and returns
   //   a pointer to the root.
   virtual NodePtr induce_decision_tree(CGraph&,
					const Array<int>& tieBreakingOrder,
 					Real& numSubtreeErrors,
 					Real& pessimisticSubtreeErrors,
					int& numLeaves,
					int remainingSiblings = -1) const;

   // Best_split finds the best split in the node and returns a
   // categorizer implementing it.  It allocates and returns catNames
   // containing the names of the resulting categories.
   virtual NodeCategorizer *best_split(Array<MStringRC>*& catNames) const = 0;
   // create_subinducer creates the Inducer for calling recursively.
   // Note that since this is an abstract class, it can't create a copy
   //   of itself.
   // Important: you MUST copy all settable options to the created inducer.  A
   //   good way to do this is to update copy_options by adding new copies and
   //   calling TDDTInducer::copy.
   virtual TDDTInducer* create_subinducer(const MString& dscr,
                                          CGraph& aCgraph) const = 0;
   virtual int num_nontrivial_nodes() const;
   virtual int num_nontrivial_leaves() const;
   virtual int num_nodes() const;
   virtual int num_leaves() const;

   // options.  Don't forget to update copy_options.
   void set_c45_options();
   void set_debug(Bool val) { tddtOptions.debug = val; }
   Bool get_debug() const { return tddtOptions.debug; }
   int get_max_level() const {return tddtOptions.maxLevel;}
   void set_max_level(int level)
      { tddtOptions.maxLevel = level;}
   void set_lower_bound_min_split_weight(Real val)
      { tddtOptions.lowerBoundMinSplitWeight = val; }
   Real get_lower_bound_min_split_weight() const
      { return tddtOptions.lowerBoundMinSplitWeight; }
   void set_upper_bound_min_split_weight(Real val)
      { tddtOptions.upperBoundMinSplitWeight = val; }
   Real get_upper_bound_min_split_weight() const
      { return tddtOptions.upperBoundMinSplitWeight; }
   void set_min_split_weight_percent(Real val)
      { tddtOptions.minSplitWeightPercent = val; }
   Real get_min_split_weight_percent() const
      { return tddtOptions.minSplitWeightPercent; }
   void set_nominal_lbound_only(Bool val)
      { tddtOptions.nominalLBoundOnly = val; }
   Bool get_nominal_lbound_only() const
      { return tddtOptions.nominalLBoundOnly; }
   void set_unknown_edges(Bool val) { tddtOptions.unknownEdges = val; }
   Bool get_unknown_edges() const { return tddtOptions.unknownEdges; }
   SplitScore::SplitScoreCriterion get_split_score_criterion() const
      {return tddtOptions.splitScoreCriterion; }
   void set_split_score_criterion(const SplitScore::SplitScoreCriterion val)
     {tddtOptions.splitScoreCriterion = val; }
   void set_empty_node_parent_dist(Bool b)
      {tddtOptions.emptyNodeParentDist = b; }
   Bool get_empty_node_parent_dist() const
      {return tddtOptions.emptyNodeParentDist; }
   void set_parent_tie_breaking(Bool b)
      {tddtOptions.parentTieBreaking = b; }
   Bool get_parent_tie_breaking() const
      {return tddtOptions.parentTieBreaking; }
   void set_pruning_method(PruningMethod pM)
      {tddtOptions.pruningMethod = pM; }
   PruningMethod get_pruning_method() const
      {return tddtOptions.pruningMethod; }
   void set_pruning_branch_replacement(Bool b)
      {tddtOptions.pruningBranchReplacement = b; }
   Bool get_pruning_branch_replacement() const
      {return tddtOptions.pruningBranchReplacement; }
   void set_adjust_thresholds(Bool b)
      {tddtOptions.adjustThresholds = b; }
   Bool get_adjust_thresholds() const
      {return tddtOptions.adjustThresholds; }
   void set_pruning_factor(Real val)
      { tddtOptions.pruningFactor = val; }
   Real get_pruning_factor() const
      { return tddtOptions.pruningFactor; }
   int get_smooth_inst() const { return tddtOptions.smoothInst; }
   void set_smooth_inst(int inst) { tddtOptions.smoothInst = inst; }
   Real get_smooth_factor() const { return tddtOptions.smoothFactor; }
   void set_smooth_factor(Real factor) { tddtOptions.smoothFactor = factor; }
  
   void set_cont_mdl_adjust(Bool val)
      { tddtOptions.contMDLAdjust = val; }
   Bool get_cont_mdl_adjust() const
      { return tddtOptions.contMDLAdjust; }

   void set_leaf_dist_type(LeafDistType type)
      { tddtOptions.leafDistType = type; }
   LeafDistType get_leaf_dist_type() const
      { return tddtOptions.leafDistType; }

   void set_m_estimate_factor(Real factor)
      { tddtOptions.MEstimateFactor = factor; }
   Real get_m_estimate_factor() const { return tddtOptions.MEstimateFactor; }

   void set_evidence_factor(Real factor)
      { tddtOptions.evidenceFactor = factor; }
   Real get_evidence_factor() const { return tddtOptions.evidenceFactor; }

   Real get_leaf_dist_parameter() const;
   /* @@ We get an assembler warning if this is left inline
   { if (tddtOptions.leafDistType == laplaceCorrection)
      return tddtOptions.MEstimateFactor;
   else if (tddtOptions.leafDistType == evidenceProjection)
      return tddtOptions.evidenceFactor;
   else return 0; }
   */
   void set_evaluation_metric(EvalMetric metric)
   { tddtOptions.evaluationMetric = metric; }
   EvalMetric get_evaluation_metric() const
   { return tddtOptions.evaluationMetric; }

   Bool get_have_continuous_attributes() const
   { return haveContinuousAttributes; }
   void set_have_continuous_attributes(Bool val)
   { haveContinuousAttributes = val; }

   // @@ This shouldn't work because TDDTOptions is private.
   // @@ move to private section.
   TDDTOptions get_options() const { return tddtOptions; }
   void set_options(TDDTOptions val) { tddtOptions = val; }
   virtual void set_user_options(const MString& prefix);
   void copy_options(const TDDTInducer& inducer);
   
   int get_level() const {return level;}
   void set_level(int lvl) {level = lvl;}
   virtual void dribble_level(int level, const MString& str,
			      int remaining) const;
   virtual void display_options(MLCOStream& stream = Mcout,
				const DisplayPref& dp =
				DisplayPref::ASCIIDisplay) const;

   // override normalize_weights to change MIN_SPLIT option
   virtual void normalize_weights(Real normFactor);
};

#endif
