// MLC++ - Machine Learning Library in -*- C++ -*-
// See Descrip.txt for terms and conditions relating to use and distribution.

// This is an include file.  For a full description of the class and
// functions, see the file name with a "c" suffix.

#ifndef _entropy_h
#define _entropy_h 1

#include <Attribute.h>
#include <Array2.h>

// some constants for splitAndLabelDist and splitDist counts
const int UNKNOWN_NODE = UNKNOWN_CATEGORY_VAL;
const int LEFT_NODE = FIRST_CATEGORY_VAL;
const int RIGHT_NODE = FIRST_CATEGORY_VAL+1;

class LogOptions;
class InstanceList;
class SplitScore;
class RealAndLabelColumn;

// Used in make_score_array()
struct ThresholdInfo {
   NO_COPY_CTOR(ThresholdInfo);

   // Public member data
   int index;
   Real score;
   Real weight;

   // Methods
   ThresholdInfo()
      : index(UNDEFINED_INT), score(UNDEFINED_REAL),
	weight(UNDEFINED_REAL) {}
   ThresholdInfo& operator=(const ThresholdInfo& rhs);
};

// Note that this should theoretically be declared as a private class
// to entropy.c, but because we generate an array of this type, the
// template mechanism fails.
// operator<, operator= only compare the real value.
// origin and pos are used only in find_best_threshold() and get_score_array().
class AttrAndOrigin {
   NO_COPY_CTOR(AttrAndOrigin);
public:
   // Public member data
   int pos;  // pos is the instance list number where the instance goes to.
   int origin;  // origin is where the instance comes from.
   StoredReal attrVal;
   AttrValue_ label;

   // Methods
   AttrAndOrigin()
      : pos(UNDEFINED_INT), origin(UNDEFINED_INT), attrVal(-1), label() {}
   AttrAndOrigin& operator=(const AttrAndOrigin& source);
   // Comparison is only concerned with the Stored Real attrVal
   Bool operator==(const AttrAndOrigin& rhs) const
      {return (attrVal == rhs.attrVal);}
   Bool operator<(const AttrAndOrigin& rhs) const
      {return (attrVal < rhs.attrVal);}
   Bool operator!=(const AttrAndOrigin& rhs) const
      {return ! ((*this)== rhs);}
};


// Entropy H(Y) for the label Y
// These are in order of efficiency.  If you know the distribution, give it
//   and not the list.  If you know the labelCount, don't make us count.
Real entropy(const InstanceList&);
Real entropy(const Array<Real>& labelCount);
Real entropy(const Array<Real>& labelCount, Real totalWeight);

// Entropy weighted by the parent.  Used in LazyDT.
// Confidence interval gives a pessimistic estimate shifted towards
//   p = 1/num-classes.
Real weighted_entropy(const Array<Real>& labelCount, 
		      const Array<Real>& parentLabelCount, 
		      Real totalWeight, Real parentTotalWeight,
		      Real confIntervalZ = 0);
Real weighted_entropy(const InstanceList& instList,
		      const InstanceList& parentInstList,
		      Real confIntervalZ = 0);

// Gini weighted by the parent.  Used in LazyDT.
// Confidence interval gives a pessimistic estimate shifted towards
//   p = 1/num-classes.
Real weighted_gini(const Array<Real>& labelCount, 
		   const Array<Real>& parentLabelCount, 
		   Real totalWeight,
		   Real parentTotalWeight,
		   Real confIntervalZ = 0);
Real weighted_gini(const InstanceList& instList,
		   const InstanceList& parentInstList,
		   Real confIntervalZ = 0);


// The unaccounted instances allow the list of nodes to be partial, i.e.,
//    not to contain all instances.  The split will be created so that
//    the unaccounted instances are in an extra split with the same label,
//    so that the entropy will be decreased correctly as if they were
//    in a pure node.
Array2<Real>* build_split_and_label_dist(const Array<const InstanceList*>
					&currentLevel, int attrNum,
					Real numUnaccountedInst = 0);
Array<Real>* build_nominal_attr_split_dist(const Array<const InstanceList*>
					  &currentLevel, int attrNum,
					  Real numUnaccountedInst = 0);
// Conditional entropy H(Y|X) for the label Y
Real cond_entropy(const Array2<Real>& splitAndLabelDist,
		  const Array<Real>& splitDist, Real totalWeight);
Real cond_entropy(const InstanceList&, int attrNumX);

Real j_measure(const Array2<Real>& splitAndLabelDist,
	       const Array<Real>& splitDist, const Array<Real>& labelCounts,
	       NominalVal x,  Real totalWeight);
Real j_measure(const InstanceList& instList, int attrNumX, NominalVal x);

Real two_way_cond_entropy(const Array2<Real>& splitAndLabelDist,
			  const Array<Real>& splitDist,
                          Real totalWeight, NominalVal val);
Real two_way_cond_entropy(const InstanceList&, int attrNum, NominalVal val);

// Mutual information I(Y;X) = H(Y) - H(Y|X) for label Y, attribute X
// Mutual information is called "gain" by some researchers like Quinlan.
Real mutual_info(Real ent,
		 const Array2<Real>& splitAndLabelDist,
		 const Array<Real>& splitDist, Real totalWeight);
Real mutual_info(const InstanceList& instList, Real ent, int attrNumX);
Real mutual_info(const InstanceList& instList, int attrNumX);

// Automatic determining of the lower bound on minSplit
Real auto_lbound_min_split(Real totalWeight);

// Caculate the minSplit.
Real min_split(Real totalWeight,
	       int numTotalCategories, Real lowerBoundMinSplit,
	       Real upperBoundMinSplit, Real minSplitPercent);
Real min_split(const InstanceList &instList,
	       Real lowerBoundMinSplit, Real upperBoundMinSplit,
	       Real minSplitPercent, Bool ignoreNumCat = FALSE);
// @@Bool check_min_split(Array<Real>& splitDist, Real minSplit, int instListNum);
PtrArray<RealAndLabelColumn*>*
   build_real_and_label_columns(const InstanceList& instList,
				int attrNum = UNDEFINED_INT);

RealAndLabelColumn*
   build_real_and_label_column(const InstanceList& instList, int attrNum);

Real find_best_score(Real totalKnownWeight,
		     DynamicArray<ThresholdInfo>& scores, Real minSplit,
		     int& bestSplitIndex);
void find_best_threshold(RealAndLabelColumn& realAndLabelColumn,
		         Real minSplit, SplitScore& split,
		         Real& bestThreshold, int& bestSplitIndex,
		         int& numDistinctValues, int smoothInst=0,
		         Real smoothFactor=0.5);
void find_best_threshold(Array<RealAndLabelColumn*>& columns,
		         Real minSplit, SplitScore& split,
		         Real& bestThreshold, int& bestSplitIndex,
		         int& numDistinctValues, int smoothInst=0,
		         Real smoothFactor=0.5);
void find_best_threshold_main_body(Array<RealAndLabelColumn*>& columns,
				   Real minSplit, SplitScore& split,
				   Real& bestThreshold, int& bestSplitIndex,
				   int& numDistinctValues, int smoothInst=0,
				   Real smoothFactor=0.5);

Array2<Real> *get_score_array(RealAndLabelColumn& realAndLabelColumn,
			      SplitScore& split, Real minSplit,
			      DynamicArray<ThresholdInfo> &outScores,
			      int &numDistinctValues, int smoothInst,
			      Real smoothFactor);
Array2<Real> *get_score_array(Array<RealAndLabelColumn*>& columns,
			      DynamicArray<AttrAndOrigin>& attrOriginArray,
			      SplitScore& split, Real minSplit,
			      DynamicArray<ThresholdInfo> &outScores,
			      int &numDistinctValues, int smoothInst,
			      Real smoothFactor);
void fill_scores(RealAndLabelColumn& realAndLabelColumn,
		 SplitScore& split, Real minSplit, Real theEntropy,
		 DynamicArray<ThresholdInfo>& outScores,
		 int& numDistinctValues, Array2<Real>& splitAndLabelDist,
		 Array<Real>& splitDist);
void fill_scores(Array<RealAndLabelColumn*>& columns,
		 Array<AttrAndOrigin>& attrOriginArray,
		 SplitScore& split, Real minSplit, Real theEntropy,
		 DynamicArray<ThresholdInfo>& outScores,
		 int& numDistinctValues, Array2<Real>& splitAndLabelDist,
		 Array<Real>& splitDist);
void smooth_scores(DynamicArray<ThresholdInfo> &scores,
		   int smoothInst, Real smoothFactor);


// Split Info as defined by Quinlan
Real split_info(const LogOptions&,
		 const InstanceList&, int attrNumX);
Real split_info(const LogOptions&, const Array2<Real>& splitAndLabelDist,
		const Array<Real>& splitDist, const Array<Real>& lc,
		Real totalWeight);

// Gain Ratio as defined by Quinlan
Real gain_ratio(const LogOptions& logOptions, const InstanceList& instList,
		int attrNumX, Real mutualInfo);
Real gain_ratio(const LogOptions& logOptions, const Array<Real>& splitDist,
		Real totalWeight, Real mutualInfo);
Real gain_ratio(Real mutualInfo, Real SplitEntropy);



#endif
